#include "iplookup.h"

addr_table_t **addresses;
trie_t *trie;
ip_addr_t RLE_len;
ip_addr_t tot_leaves = 0;
rle_t **RLE_seq;
void *col;
char flag, comp_col = 0;  
col_index_t *col_index = NULL;
col_index_t row_len[2];
ip_addr_t tot_rows[2];
cell_t **row_index;
ip_addr_t b_size = 0, b_tmp = 0;

int main (int argc, char *argv[]) {
  struct net_table *in_table;
  struct net_table *cpt_table;
  ip_addr_t ntimes = 1;
  char *next_hops[MAX_NEXT_HOPS];
  long tempo;
  long look_time[2];
  quadr_t **quadr_vect;
  ip_addr_t tot_route;
  flag = eval_params (argc, argv, &ntimes);
  /*  if ntimes is 0  there isn't an argoument for address file*/
  if (ntimes == 0) 
    in_table = load_table (argv[argc - 1], next_hops, &tot_route, flag & 1);
  else
    in_table = load_table (argv[argc - 2], next_hops, &tot_route, flag & 1);
  if (in_table == NULL) return -1;
  cpt_table = copy (in_table);
  if (cpt_table == NULL) return -1;
  /*  input table loaded, here start to build structure  */
  /*  Create table front-end  */
  Mark_Time ();  
  if ((flag & 16)) {
    quadr_vect = create_quads_S (in_table, &tot_route);
    if (quadr_vect == NULL) return -1;
    /*  In_table was freed and now set pointer to NULL  */
    in_table = NULL;
    RLE_seq = create_RLE_sequence (quadr_vect, &tot_route);
    if (RLE_seq == NULL) return -1;
    /*  copy the exact length of RLE_seq vector  */
    RLE_len = tot_route;
    /*  Create trie for update operation */
    trie = create_trie (cpt_table, &tot_leaves);
  }
  else {
    /*  Create trie for update operation */
    trie = create_trie (in_table, &tot_leaves);
    /*  In_table was freed and now set pointer to NULL  */
    in_table = NULL;  
    RLE_len = tot_leaves;
    RLE_seq = update_RLE_sequence (trie, &RLE_len);
    if (RLE_seq == NULL) return -1;
  }
  /*  create table back-end  */
  /*  Create colun index  */
  col_index = (col_index_t *) malloc (exp (32-K_SIZE) * sizeof (col_index_t));
  if (col_index == NULL) {
    printf ("Malloc failure creating colunn index...can`t go on\n");
    return -1;
  }
  row_index = rebuild_table(RLE_seq,flag,col_index,RLE_len,tot_rows,row_len);
  if (row_index == NULL) return -1;
  /*  RLE_seq was freed and now set pointer to NULL  */
  RLE_seq = NULL;
  /*  Colunn index creation  */
  col = (void *) col_index;
  /*   force normal colun index  */
  if (!(flag & 64)) {
    col = build_col_index ((col_index_t *) col, flag, row_len, 
			   &b_size, &b_tmp, &comp_col);
  }
  tempo = Mark_Time ();
  printf ("Elapsed time %ld\n", tempo);
  printf ("Done\n");
  /*  Write cdg.log file if requested */
  if (flag & 1) {
    write_log_file (tot_rows, row_len, b_size, b_tmp, flag, comp_col, tempo);
  }
  /*  If iteration is unspecified was set to 1 from eval_params ()
      if iteration is 0 skip lookup  */
  if (ntimes != 0) {
    addresses = load_address (argv[argc - 1], &tot_route, flag & 8, next_hops);
    if (addresses == NULL) return -1;
    look_time[0] = no_lookup_time (addresses, tot_route, ntimes);
    switch (comp_col) {
    case 0:
      printf ("Normal lookup procedure\n"); 
      look_time[1] = make_lookups (row_index, (col_index_t *) col, addresses,
				  tot_route, ntimes);
      break;
    case 1:
      printf ("Short type lookup procedure\n");
      look_time[1] = make_lookups_sc (row_index, (col_short_t *) col,
				      addresses, tot_route, ntimes);
      break;
    case 2: 
      printf ("bucket lookup procedure\n");
      look_time[1] = make_bucket_lookups (row_index, (bucket_t *) col,
					  addresses, tot_route, ntimes,b_size);
      break;
    }
    printf ("-- Lookup time --\n");
    printf ("Entire lookup procedure %ld\n",look_time[1]);
    printf ("No lookup operations %ld\n",look_time[0]);
    printf ("Real lookup time %ld\n",look_time[1] - look_time[0]);
    /*  Add lookup information to cdg.log file if requested */
    if (flag & 1) {
      write_log_lookup (comp_col, look_time);
    }
    write_results (addresses, tot_route, next_hops, flag & 8);
    unload_address (addresses, tot_route);
    printf ("Done\n");
  }

  unload_next_hops (next_hops);
  return 0;
}

void **insert_and_rebuild (ip_addr_t ind) {
  struct net_table elem; 
  void **ret = malloc (2 * sizeof (cell_t *));
  if (ret == NULL) exit (-2);
  elem.ip = addresses[ind]->first_pair;
  elem.netmask = addresses[ind]->last_pair;
  elem.next_hop = addresses[ind]->next_hop;
  RLE_len = trie_insert_network (&elem, trie);
  /*  -1 if route not exist in trie  */
  if (RLE_len != -1) {
    tot_leaves += RLE_len;
    rebuild ();
    ret[0] = row_index;
    ret[1] = col;
    return ret;
  }
  return NULL;
}

void **delete_and_rebuild (ip_addr_t ind) {
  void **ret = malloc (2 * sizeof (cell_t *));
  if (ret == NULL) exit (-2);
  RLE_len = trie_remove_network (addresses[ind]->first_pair,
				 addresses[ind]->last_pair, trie);
  /*  -1 if route not exist in trie  */
  if (RLE_len != -1) {
    tot_leaves -= RLE_len;
    rebuild ();
    ret[0] = row_index;
    ret[1] = col;
    return ret;
  }
  return NULL;
}

void rebuild (void) {
  /*  tot_leaves is the length of RLE vector  */
  RLE_len = tot_leaves;
  /*  Update the RLE vector  */
  RLE_seq = update_RLE_sequence (trie, &RLE_len);
  if (RLE_seq == NULL) exit (-1);
  /*  rebuild structure  */
  row_index = rebuild_table(RLE_seq,flag,col_index,RLE_len,tot_rows,row_len);
  if (row_index == NULL) exit (-1);
  RLE_seq = NULL;
  col = (void *) col_index;
  /*  in normal lookup procedure col index was already modified  */
  if (comp_col == 0) return;
  /*   force normal colun index  */
  if (!(flag & 64)) {
    col = rebuild_col_index ((col_index_t *) col, flag, row_len, 
			   &b_size, &b_tmp, comp_col);
  }
}


    



