#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/time.h>

typedef unsigned char cell_t;     /*  next hop  */
typedef unsigned char ip_t;       /*  IP element from 0 to 255  */
typedef unsigned int ip_addr_t;   /*  IPv4 entire address  */
typedef unsigned short int col_index_t; /*  A type for colunn index  */
typedef unsigned char col_short_t;  /*  Short data value for colun index  */
typedef short int bucket_t;       /*  for bucket colun index  */

typedef unsigned short int usi_t; /*  Short unsigned int  */

#define MAX_IP_LUNG_AS_STRING 16  /*  strlen ("XXX.XXX.XXX.XXX");  */
#define MAX_RIGA_READ 256
#define MAX_NEXT_HOPS 256   /*  Maximum number of next hops in table  */
#define MAX_SHORT_COL 256   /*  Maxinum value storable in col_short_t   */
#define K_SIZE 16
#define LAST_L_VALUE 3  /*  illegal value for the last S triple  */
#define	OVER_BUFFER 1   /*  Usually this is a good value!  */
#define DONTCARE MAX_NEXT_HOPS - 1  /*  match everything */

#define LOOKUP 0  /*  Lookup operation requested  */
#define INSERT 1  /*  Insert operation requested  */
#define DELETE 2  /*  Delete operation requested  */

#define first_k(x) (x >> (32 - K_SIZE))
#define exp(x) (1 << (x))
#define bit_k(x,k) ((x & (1 << (k))) >> (k))

struct net_table {
  ip_addr_t ip;  
  ip_t netmask;
  cell_t next_hop;
  struct net_table *next;
};

typedef struct {
  ip_addr_t ip_0;         /*   p0 or p1   */
  ip_t netmask;           /*   net mask   */  
  ip_t tipo;              /*    0 or 1    */
  cell_t next_hop;        /*   Next hop   */
  void *next;             /* used as list */
  ip_addr_t l_val;        /*  value of l  */
} quadr_t;

typedef struct {
  cell_t next_hop;       /*  Next hop  */
  ip_addr_t run_length;  /*  length */
} rle_t;

typedef struct {
  ip_addr_t num;         /*  cluster number (rebuild right order after sort) */
  ip_addr_t vlen;
  cell_t *next_hop;
  ip_addr_t *run_length;
} cluster_t;

/*  The size of IP halves is unknown because depend from K_SIZE costant  */
typedef struct {
  ip_addr_t first_pair;  /*   first part of ip address  */
#if K_SIZE - 16          /*  if K_SIZE != 16  */
  ip_addr_t last_pair;   /*  second part of ip address  */
#else                    /*  if K_SIZE == 16  */
  usi_t last_pair;       /*  second part of ip address  */
#endif
  cell_t next_hop;
  char op;
  void *next;
} addr_list_t;

/*  this type havn't the pointer next. That need for optimize cache use in
    lookup operation  */
typedef struct {
  ip_addr_t first_pair;  /*   first part of ip address  */
#if K_SIZE - 16          /*  if K_SIZE != 16  */
  ip_addr_t last_pair;   /*  second part of ip address  */
#else                    /*  if K_SIZE == 16  */
  usi_t last_pair;       /*  second part of ip address  */
#endif
  cell_t next_hop;
  char op;
} addr_table_t;


typedef struct {
  cell_t next_hop;
  ip_t mark; 
  void *child[2];
} trie_t;

/*  createRLE.c  */
rle_t **create_RLE_sequence (quadr_t **vect, ip_addr_t *tot_triple);
cluster_t **create_cluster (rle_t **rle_seq, ip_addr_t *size, ip_addr_t *sort, ip_addr_t *tab_size);
void applay_changes (ip_addr_t *sort, usi_t *subs);
char is_equal (cluster_t *el1, cluster_t *el2);
int comp_clust2 (const void *elem1, const void *elem2);
int comp_clust3 (const void *elem1, const void *elem2);
ip_addr_t max_RLE_cluster_length (rle_t **rle_seq, ip_addr_t size);

/*  createipt.c  */
void create_colun_index (col_index_t *col_index, cluster_t **RLE, ip_addr_t RLE_size, ip_addr_t exp_col);
char createrow (cell_t *row, ip_addr_t *metric, col_index_t size, cluster_t *RLE, col_index_t row_len);
cell_t **create_ip_table (cluster_t **cluster, ip_addr_t row_len, ip_addr_t *sort, ip_addr_t *metric, ip_addr_t met_len, ip_addr_t tab_len);
ip_addr_t *create_metrics (col_index_t *col_index);

/*  quadfront.c  */
quadr_t **create_quads_S (struct net_table *in_table, ip_addr_t *tot_route);
int ins_quad (quadr_t *el, ip_addr_t ip, ip_t net, ip_t t, cell_t nh);
ip_addr_t ins_holes_quads (quadr_t *list, ip_addr_t *err);
int compar (const void *elem1, const void *elem2);
ip_addr_t ip_1 (ip_addr_t address, ip_t netmask);

/*  loadroute.c  */
int ins_in_next_hop_array (char **next_hops, char *next_hop);
ip_t used_bits (ip_t elem);
ip_t calcola_netmask (ip_t ip1, ip_t ip2, ip_t ip3, ip_t ip4);
ip_addr_t correct_nerwork (ip_addr_t ip, ip_t netmask);
void unload_next_hops (char **next_hops);
struct net_table *load_table (char *nome, char **next_hops,ip_addr_t *rou_c, char flag);
struct net_table *copy (struct net_table *in_table);

/*  supportfn.c  */
long int Mark_Time (void);
char eval_params (int argc, char *argv[], ip_addr_t *ntimes);
void echo_use (void);
void help (void);
void write_log_file (ip_addr_t *tot_rows, col_index_t *row_len, ip_addr_t b_size, ip_addr_t b_tmp, char flag, char comp_col, long tempo);
void write_log_lookup (char comp_col, long *look_time);
void write_results (addr_table_t **array, ip_addr_t len, char **next_hop, char flag);
long no_lookup_time (addr_table_t **array, ip_addr_t len, ip_addr_t ntimes);

/*  t_backend.c  */
cell_t **rebuild_table (rle_t **RLE_seq, char flag, col_index_t *col_index, ip_addr_t tot_route, ip_addr_t *tot_rows, col_index_t *row_len);
void *build_col_index (col_index_t *col_index, char flag, col_index_t *row_len, ip_addr_t *b_size, ip_addr_t *b_tmp, char *comp_col);
void *rebuild_col_index (col_index_t *col_index, char flag, col_index_t *row_len, ip_addr_t *b_size, ip_addr_t *b_tmp, char comp_col);

/*  cdglookup.c  */
int main (int argc, char *argv[]);
void **insert_and_rebuild (ip_addr_t ind);
void **delete_and_rebuild (ip_addr_t ind);
void rebuild (void);

/*  mk_lookup.c  */
long make_lookups (cell_t **row, col_index_t *col, addr_table_t **array, ip_addr_t len, ip_addr_t ntimes);
long make_lookups_sc (cell_t **row, col_short_t *col, addr_table_t **array, ip_addr_t len, ip_addr_t ntimes);
long make_bucket_lookups (cell_t **row, bucket_t *col, addr_table_t **array, ip_addr_t len, ip_addr_t ntimes, ip_addr_t size);

/*  load_addr.c  */
addr_table_t **load_address (char *nome, ip_addr_t *len, char flag, char **nh);
void unload_address (addr_table_t **array, ip_addr_t len);
int load_data (addr_list_t *self, char *riga, char flag, char **nh);

/*  splitRLE.c  */
cluster_t ***create_split_cluster (rle_t **rle_seq, ip_addr_t *size, ip_addr_t *sort, ip_addr_t *tab_size);
cell_t **create_block (cluster_t ***cluster,ip_addr_t rows, ip_addr_t *metric, col_index_t *met_len);
cell_t **shuffle (cell_t **block, ip_addr_t *sort);
char compress_structure (cell_t **block, ip_addr_t *sort, ip_addr_t bl_size);
ip_addr_t erase_rows (cell_t **block, ip_addr_t *sort, ip_addr_t size, col_index_t len);
char are_equals (cell_t *el1, cell_t *el2, col_index_t len);
int comp_blk2 (const void *elem1, const void *elem2);
int comp_blk (const void *elem1, const void *elem2);

/*  mutch.c  */
cell_t **create_match_set (cell_t **block, ip_addr_t *size_r, col_index_t *size_c, ip_addr_t len);
col_index_t **create_bipartite_graph (cell_t **nodes, ip_addr_t *size_r, col_index_t *size_c);
col_index_t hamming_distance (cell_t *str1, cell_t *str2, ip_addr_t len);
col_index_t *find_perfect_matching (col_index_t **graph, col_index_t *len);
void make_permutations (cell_t **block, ip_addr_t rows, col_index_t cols, col_index_t max, col_index_t *col_index, col_index_t *subs);
void destroy_graph (col_index_t **graph, col_index_t len);
void destroy_match_set (cell_t **nodes);

/*   col_index.c  */
col_short_t *compress_col_index_type (col_index_t *col_index);
ip_addr_t compute_bucket_size (col_index_t *col_index);
ip_addr_t tot_bucket_fault (col_index_t *col_index, ip_addr_t bucket);
bucket_t *create_bucket_col_index (col_index_t *col_index);

/*  cdgupdate.c  */
trie_t *create_trie (struct net_table *in_table, ip_addr_t *tot_nodes);
trie_t *create_root (void);
int trie_insert_network (struct net_table *elem, trie_t *root);
void visit_b_tree (trie_t *root, rle_t **RLE, ip_addr_t *RLE_ind, ip_addr_t ip, ip_t netmask);
rle_t **build_RLE_sequence (rle_t **RLE_seq, ip_addr_t *RLE_len);
rle_t **build_empty_RLE_sequence (ip_addr_t len);
int trie_remove_network (ip_addr_t ip, ip_t netmask, trie_t *root);
ip_addr_t merge_leaves (trie_t *root, ip_addr_t delete);
void cover_holes (trie_t *root);
rle_t **update_RLE_sequence (trie_t *root, ip_addr_t *RLE_len);
