#include "iplookup.h"

addr_table_t **load_address (char *nome, ip_addr_t *len, char flag, char **nh){
  FILE *pf;
  int error;  /*  Error inside input routing table  */
  char riga[MAX_RIGA_READ];
  ip_addr_t route_count, err_route_count, i;
  addr_list_t *init, *self, *prec;
  addr_table_t **array;
  if ((pf = fopen (nome, "r")) == NULL) {
    printf ("Unable to open address file\n");
    return NULL;
  }
  /*  header (will be destroyed at end)  */
  init = (addr_list_t *) malloc (sizeof (addr_list_t));
  prec = init;
  route_count = 0;
  err_route_count = 0;
  /*  Load entry form file  */
  while (fgets (riga, MAX_RIGA_READ, pf) != NULL) {
    error = 0;  /*  Not errors found in this entry of input table  */
    route_count ++;
    self = (addr_list_t *) malloc (sizeof (addr_list_t));
    if (self == NULL) {
      printf ("Malloc faliure... can't go on\n");
      /*  Free already allocated resources  */
      prec->next = NULL;
      while (init != NULL) {
	prec = init;
	init = init->next;
	free (init);
      }
      return NULL;
    }
    switch (riga[0]) {
    case 105: 
    case 73:
      self->op = INSERT;
      error = load_data (self, riga + 2, 0, nh);  /*  + 2 skip "I "  */
      break;
    case 100:
    case 68:
      self->op = DELETE;
      error = load_data (self, riga + 2, 0, NULL);  /*  + 2 skip "D "  */
      break;
    default:
      self->op = LOOKUP;
      error = load_data (self, riga, flag, NULL);
      break;
    }
    if (error == 0) {	
      prec->next = self;
      prec = prec->next;
    }
    else {
      free (self);
      err_route_count ++;
    }
  }
  /*  close list and destroy useless header  */
  prec->next = NULL;
  prec = init;
  init = init->next;
  free (prec);
  /*  set the number of corrected elements loaded  */
  *len = route_count - err_route_count;
  printf ("%u addresses loaded with %d errors\n",route_count, err_route_count);
  fclose (pf);
  /*  Store list into array  */
  array = (addr_table_t **) malloc (*len * sizeof (addr_table_t *));
  if (array == NULL) {
    printf ("Malloc failure storing addresses into array\n");
    return NULL;
  }
  /* the structure must to be copied in shorter structure for minimize bad
     effects of it in using cache during lookups  */
  self = init;
  for (i = 0; i < *len; i ++) {
    array[i] = (addr_table_t *) malloc (sizeof (addr_table_t));
    if (array[i] == NULL) {
      printf ("Malloc failure storing addresses into array\n");
      return NULL;
    }
    array[i]->first_pair = self->first_pair;
    array[i]->last_pair = self->last_pair;
    array[i]->op = self->op;
    array[i]->next_hop = self->next_hop;
    init = self;
    self = self->next;
    free (init);  /* free old structure  */
  }
  return array;
}

void unload_address (addr_table_t **array, ip_addr_t len) {
  ip_addr_t i;
  for (i = 0; i < len; i ++) {
    free (array[i]);
  }
  free (array);
}

int load_data (addr_list_t *self, char *riga, char flag, char **nh) {
  int ip[4], netmask;  /* need to be int  */      
  char buff[MAX_RIGA_READ];
  switch (self->op) {
  case LOOKUP:
    sscanf (riga,"%d.%d.%d.%d",&ip[0],&ip[1],&ip[2],&ip[3]);
    break;
  case INSERT:
    sscanf(riga,"%d.%d.%d.%d/%d %s",&ip[0],&ip[1],&ip[2],&ip[3],&netmask,buff);
    break;
  case DELETE:
    sscanf (riga,"%d.%d.%d.%d/%d",&ip[0],&ip[1],&ip[2],&ip[3],&netmask);
    break;
  }
  sscanf (riga,"%d.%d.%d.%d",&ip[0],&ip[1],&ip[2],&ip[3]);
  /*  Verify data integrity and copy into list  */
  if (ip[0] < 0 || ip[0] > 255 || ip[1] < 0 || ip[1] > 255) {
    printf ("Address error %d.%d.%d.%d, route ignored\n",ip[0],ip[1],ip[2],ip[3]);
    return 1;
  }
  else {
    if (ip[2] < 0 || ip[2] > 255 || ip[3] < 0 || ip[3] > 255) {
      printf ("Address error %d.%d.%d.%d, route ignored\n",ip[0],ip[1],ip[2],ip[3]);
    return 1;
    }
  }
  /*  Lookup operation  */
  if (self->op == LOOKUP) {
    /*  Load 16 or 17 bit in first_pair  */
    if (!flag)  /*  16 bit  */
      self->first_pair = (ip[0] << 8) + ip[1];
    else        /*  17 bit  */
      self->first_pair = (((ip[0] << 8) + ip[1]) << 1) + (ip[2] >> 7) ;
    self->last_pair  = (ip[2] << 8) + ip[3];
    return 0;
  }
  /*  other cases */
  self->first_pair = (ip[0] << 24) + (ip[1] << 16) + (ip[2] << 8) + ip[3];
  /*  Veryfy netmassk integrity  */
  if (netmask < 1 || netmask > 32) {
    printf ("Netmask error %d.%d.%d.%d/%d, route ignored\n",ip[0],ip[1],ip[2],ip[3],netmask);
    return 1;
  }
  self->last_pair = netmask;
  /*  Compute minimal netmask and correct if not enough  */
  netmask = calcola_netmask (ip[0], ip[1], ip[2], ip[3]);
  if (netmask > self->last_pair) {
    printf ("Network error %d.%d.%d.%d/%d...",ip[0],ip[1],ip[2],ip[3],self->last_pair);
    self->first_pair = correct_nerwork (self->first_pair, self->last_pair);
    ip[0] = self->first_pair >> 24;
    ip[1] = (self->first_pair << 8) >> 24;
    ip[2] = (self->first_pair << 16) >> 24;
    ip[3] = self->first_pair % 256;
    printf ("corrected in %d.%d.%d.%d/%d\n",ip[0],ip[1],ip[2],ip[3],self->last_pair);
  }
  if (self->op == DELETE) return 0;
  /*  Insert next hop in his array and code in cell_t  */
  self->next_hop = (cell_t) ins_in_next_hop_array (nh, buff);
  return 0;
}
