#include "iplookup.h"

cluster_t ***create_split_cluster (rle_t **rle_seq, ip_addr_t *size, ip_addr_t *sort, ip_addr_t *tab_size) {
  cluster_t ***RLE;
  cluster_t ***RLE_sort[2];
  ip_addr_t max, cl_len[2], cluster[2], clust_pos[2];
  ip_addr_t i, j, k, vlen;
  ip_addr_t exp_val = exp (32 - K_SIZE);
  ip_addr_t exp_half = exp (32 - K_SIZE - 1);
  ip_addr_t *run_len;
  int who = 0; /* may be only 0 or 1, indicate the index of used half table */
  usi_t *subs;
  cell_t *next_hop;
  /* It allocate the 2 parts of the cluster multistructure  */
  RLE = (cluster_t ***) malloc (2 * sizeof (cluster_t **));
  if (RLE == NULL) {
    printf ("Malloc failure creating RLE main structure\n");
    return NULL;
  }
  for (i = 0; i < 2; i ++) {
    RLE[i] = (cluster_t **) malloc (exp_val * sizeof (cluster_t *));
    if (RLE[i] == NULL) {
      printf ("Malloc failure creating RLE main structure\n");
      return NULL;
    }
  }
  /*  calcolate max RLE size for cluster  */
  max = max_RLE_cluster_length (rle_seq, *size);
  /*  Allocate max len buffers for len and next hop  */
  run_len = (ip_addr_t *) malloc (max * sizeof (ip_addr_t));
  if (run_len == NULL) {
    printf ("Malloc failure creating RLE buffer\n");
    return NULL;
  }
  next_hop = (cell_t *) malloc (max * sizeof (cell_t));
  if (next_hop == NULL) {
    printf ("Malloc failure creating RLE buffer\n");
    return NULL;
  }
  for (i = 0; i < 2; i ++) {
    cluster[i] = 0;
    clust_pos[i] = 0;
    cl_len[i] = 0;
  }
  vlen = 0;
  for (i = 0; i < *size; i ++) {
    if (rle_seq[i]->run_length < exp_half - cl_len[who]) {
      cl_len[who] += rle_seq[i]->run_length;
      /*  Buffer overflow  */
      if (vlen == max) {
	printf ("Buffer too small...\n");
	max += OVER_BUFFER;
	run_len = (ip_addr_t *) realloc ((ip_addr_t *) run_len, 
					 max * sizeof (ip_addr_t));
	if (run_len == NULL) {
	  printf ("Malloc failure creating RLE buffer\n");
	  return NULL;
	}
	next_hop = (cell_t *) realloc ((cell_t *) next_hop,
				       max * sizeof (cell_t));
	if (next_hop == NULL) {
	  printf ("Malloc failure creating RLE buffer\n");
	  return NULL;
	}
	printf ("corrected\n");
      }
      run_len[vlen] = rle_seq[i]->run_length;
      next_hop[vlen] = rle_seq[i]->next_hop;
      vlen ++;
      free (rle_seq[i]);
    }
    else {
      RLE[who][cluster[who]] = (cluster_t *) malloc (sizeof (cluster_t));
      if (RLE[who][cluster[who]] == NULL) {
	printf ("Malloc failure creating RLE cluster\n");
	return NULL;
      }
      RLE[who][cluster[who]]->num = cluster[who];
      rle_seq[i]-> run_length -= (exp_half - cl_len[who]);
      /*  Buffer overflow  */
      if (vlen == max) {
	printf ("Buffer too small...");
	max += OVER_BUFFER;
	run_len = (ip_addr_t *) realloc ((ip_addr_t *) run_len, 
					 max * sizeof (ip_addr_t));
	if (run_len == NULL) {
	  printf ("Malloc failure creating RLE buffer\n");
	  return NULL;
	}
	next_hop = (cell_t *) realloc ((cell_t *) next_hop,
				       max * sizeof (cell_t));
	if (next_hop == NULL) {
	  printf ("Malloc failure creating RLE buffer\n");
	  return NULL;
	}
	printf ("corrected\n");
      }
      run_len[vlen] = exp_half - cl_len[who];
      next_hop[vlen] = rle_seq[i]->next_hop;
      vlen ++;
      RLE[who][cluster[who]]->vlen = vlen;
      RLE[who][cluster[who]]->next_hop = (cell_t *) malloc (vlen * sizeof (cell_t));
      if (RLE[who][cluster[who]]->next_hop == NULL) {
	printf ("Malloc failure creating RLE cluster\n");
	return NULL;
      }
      RLE[who][cluster[who]]->run_length=(ip_addr_t*) malloc (vlen * sizeof (ip_addr_t));
      if (RLE[who][cluster[who]]->run_length == NULL) {
	printf ("Malloc failure creating RLE cluster\n");
	return NULL;
      }
      for (j = 0; j < vlen; j ++) {
	RLE[who][cluster[who]]->next_hop[j] = next_hop[j];
	RLE[who][cluster[who]]->run_length[j] = run_len[j];
      }
      /*  inizialize counters and free sequence  */
      cl_len[who] = 0;
      vlen = 0;
      /*  verify is not first cluster  */
      if (cluster[who] > 0) {
	/*  Compare with previus cluster, if equal it is destroyed  */
	if (!is_equal (RLE[who][cluster[who]], RLE[who][cluster[who] - 1])) {
	  free (RLE[who][cluster[who]]->next_hop);
	  free (RLE[who][cluster[who]]->run_length);
	  free (RLE[who][cluster[who]]);
	  RLE[who][cluster[who]] = NULL;
	  cluster[who] --;
	}
      }
      /*  take note of order  */
      sort[(who * exp_val) + clust_pos[who]] = cluster[who];
      clust_pos[who] ++;
      cluster[who] ++;
      /* switch who  */
      if (who == 0) who ++; else who --;
      /*  If element is not fully used will be processed again  */
      if (rle_seq[i]->run_length != 0) i --;
      else free (rle_seq[i]);  /*  Free used element  */
    }
  }
  free (rle_seq);
  printf ("Cluster total %u x %u\n",cluster[0], cluster[1]);
  printf ("Sorting...");
  fflush (stdout);
  for (j = 0; j < 2; j ++) {
    /* Create triple pointer vector to make sort untouching structure */
    RLE_sort[j] = (cluster_t ***) malloc (cluster[j] * sizeof (cluster_t **));
    for (i = 0; i < cluster[j]; i ++) RLE_sort[j][i] = &RLE[j][i];
    qsort ((void*) RLE_sort[j], cluster[j], sizeof (cluster_t**), comp_clust3);
  }
  printf ("done\n");
  printf ("Erasing useles clusters...");fflush (stdout);
  for (k = 0; k < 2; k ++) {
    /*  Create subst vector and initialize this  */
    subs = (usi_t *) malloc (cluster[k] * sizeof (usi_t));
    for (i = 0; i < cluster[k]; i ++) subs[i] = i;  
    j = 0;
    cl_len[k] = 0;
    /*  compare from second to last cluster  */
    for (i = 1; i < cluster[k]; i ++) {
      /*  Cluster are equals  */
      if (is_equal (*RLE_sort[k][j], *RLE_sort[k][i]) == 0) {
	ip_addr_t ni = (*RLE_sort[k][i])->num;
	ip_addr_t nj = (*RLE_sort[k][j])->num;
	free ((*RLE_sort[k][i])->next_hop);
	free ((*RLE_sort[k][i])->run_length);
	free ((*RLE_sort[k][i]));
	*RLE_sort[k][i] = NULL;
	cl_len[k] ++;
	/*  Take note of change in sort vector  */
	/*
	  assert(ni==RLE_sort[i]-RLE[who]);
	  assert(nj==RLE_sort[j]-RLE[who]);
	*/
	subs[ni] = nj;
      }
      else j = i;  /*  go to next element  */
    }
    /*  apply changes to sort vector  */
    applay_changes (sort + (k * exp_val), subs);
    /*  free resources for updating sort vector  */
    free (RLE_sort[k]);
    free (subs);
  }
  printf ("done\n");
  /*  Recalibrate shortest stricture to biggest  */
  if (cluster[0] > cluster[1]) {
    /*  size will contain the len of RLE structure  */
    *size = cluster[0];
    for (k = cluster[1]; k < cluster[0]; k ++) RLE[1][k] = NULL;
  }
  else {
    /*  size will contain the len of RLE structure  */
    *size = cluster[1];
    for (k = cluster[0]; k < cluster[1]; k ++) RLE[0][k] = NULL;
  }
  tab_size[0] = cluster[0] - cl_len[0];
  tab_size[1] = cluster[1] - cl_len[1];
  return RLE;
}

cell_t **create_block (cluster_t ***cluster,ip_addr_t rows,
		      ip_addr_t *metric, col_index_t *met_len)
{
  ip_addr_t max_size;
  ip_addr_t i, k, offset, metoff;
  cell_t **block;
  printf ("Creating blocks");
  /* Allocate the 2 blocks in the same vector */
  block = (cell_t **) malloc (2 * rows * sizeof (cell_t *));
  if (block == NULL) {
    printf ("fail\nMalloc failure creating block structure...can't go on\n");
    return NULL;
  }
  max_size = (met_len[0] > met_len[1]) ? met_len[0] : met_len[1];
  for (k = 0; k < 2; k ++) {
    printf ("..");
    offset = k * rows;
    metoff = k * met_len[0];
    for (i = 0; i < rows; i ++) {
      if (cluster[k][i] != NULL) {
	/*  Abort if fail  */
	block[offset + i] = malloc (max_size * sizeof (cell_t));
	if (block[offset + i] == NULL) {
	  printf ("fail\nMalloc failure allocating element...can't go on\n");
	  return NULL;
	}
	if (createrow (block[offset + i], metric + metoff, 
		       met_len[k], cluster[k][i], max_size)) return NULL;
	free (cluster[k][i]->run_length);
	free (cluster[k][i]->next_hop);
	free (cluster[k][i]);
      }
      else block[offset + i] = NULL;
    }
    free (cluster[k]);  
  }
  /*  Free temporary structures  */
  free (cluster);
  printf ("done\n");
  return block;
}

cell_t **shuffle (cell_t **block, ip_addr_t *sort) {
  cell_t **row_index;
  ip_addr_t i;
  ip_addr_t exp_row = exp (32 - K_SIZE + 1);
  ip_addr_t exp_half = exp (32 - K_SIZE);
  row_index = (cell_t **) malloc (exp_row * sizeof (cell_t *));
  if (row_index == NULL) {
    printf ("Malloc failure creating colun index...can't go on\n");
    return NULL;
  }
  /*  Creating row index makeing shuffle  */
  for (i = 0; i < exp_half; i ++) 
    row_index[2 * i] = block[sort[i]];
  for (; i < exp_row; i ++) 
    row_index[2 * (i - exp_half) + 1] = block[sort[i]];
  /*  free useless resources  */
  free (sort);
  free (block);
  return row_index;
}

char compress_structure (cell_t **block, ip_addr_t *sort, ip_addr_t bl_size) {
  ip_addr_t i, j;
  usi_t *subs;
  printf ("Compressing fast...");
  /*  Create subst vector and initialize this  */
  subs = (usi_t *) malloc (bl_size * sizeof (usi_t));
  if (subs == NULL) {
    printf ("fail\nMalloc failure allocating subs vector\n");
    return -1;
  }
  for (i = 0; i < bl_size; i ++) subs[i] = i;  
  for (i = 0; block[i] != NULL; i ++);
  for (j = bl_size - 1; block[j] == NULL; j --);
  while (i < j) {
    block[i] = block[j];
    block[j] = NULL;
    subs[j] = i;
    while (block[i] != NULL) i ++;
    while (block[j] == NULL) j --;
  }
  for (i = 0; i < exp (32 - K_SIZE + 1); i ++) {
    sort[i] = subs[sort[i]];
  }
  printf ("done\n");
  return 0;
}

ip_addr_t erase_rows (cell_t **block, ip_addr_t *sort, ip_addr_t size, col_index_t len) {
  cell_t *tmp;
  cell_t ***blk_sort;
  usi_t *subs;
  ip_addr_t i, j, erased;
  char erase;
  blk_sort = (cell_t ***) malloc (size * sizeof (cell_t **));
  if (blk_sort == NULL) {
    printf ("Malloc failure creating ausiliary structure...can't go on\n");
    return ~0;  /* surely unused value  111111....  */
  }
  for (i = 0; i < size; i ++) blk_sort[i] = &block[i];
  /*  Initialize sort function len  */
  comp_blk2 ((const void *) &len, NULL);
  /*  Make sort  */
  printf ("Sorting...");
  qsort ((void*) blk_sort, size, sizeof (cell_t**), comp_blk);
  printf ("done\n");
  subs = (usi_t *) malloc (size * sizeof (usi_t));
  if (subs == NULL) {
    printf ("Malloc failure creating subs vector...can't go on\n");
    return ~0;  /* surely unused value  111111....  */
  }
  for (i = 0; i < size; i ++) subs[i] = i;  
  j = erased = 0;
  /*  compare from second to last cluster  */
  for (i = 1; i < size; i ++) {
    /*  Cluster are equals  */
    if ((erase = are_equals (*blk_sort[j], *blk_sort[i], len)) != 0) {
      ip_addr_t ni = blk_sort[i]-block;
      ip_addr_t nj = blk_sort[j]-block;
      erased ++;
      /* j is DONTCARE. elements will swap without loss of information */
      if (erase == 1) {
	tmp = *blk_sort[j];
	*blk_sort[j] = *blk_sort[i];
	*blk_sort[i] = tmp;
      }
      free (*blk_sort[i]);
      *blk_sort[i] = NULL;
      subs[ni] = nj;
    }
    else j = i;  /*  go to next element  */
  }
  /*  Applay changes  */
  for (i = 0; i < exp (32 - K_SIZE + 1); i ++) {
    sort[i] = subs[sort[i]];
  }
  return erased;  /*  may be 0 ... I'm unlucky  */
}

char are_equals (cell_t *el1, cell_t *el2, col_index_t len) {
  col_index_t i;
  for (i = 0; i < len; i ++)
    if (el1[i] != el2[i]) 
      if (el1[i] != DONTCARE && el2[i] != DONTCARE)
	return 0;
  if (el1[len - 1] == DONTCARE) return 1;
  return 2;
}

int comp_blk2 (const void *elem1, const void *elem2) {
  col_index_t i;
  static col_index_t *point = NULL;
  static col_index_t len;
  const cell_t **el1, **el2;
  if (point == NULL) {
    point = (col_index_t *) elem1;
    len = *point;
    return 0;
  }
  el1 = (const cell_t **) elem1;
  el2 = (const cell_t **) elem2;  
  for (i = 0; i < len; i ++) {
    if (el1[0][i] != el2[0][i]) return el1[0][i] - el2[0][i]; 
  }
  return 0;
}

int comp_blk (const void *elem1, const void *elem2) {
  const cluster_t ***el1, ***el2;
  el1 = (const cluster_t ***) elem1;
  el2 = (const cluster_t ***) elem2;  
  return comp_blk2(*el1,*el2);
}

