#define DEBUG

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/times.h>
#include <unistd.h>

#define MAXTOEXECUTE 1024
#define MAXNEXTHOP 240         /* 40  */
#define MAXSTRINGLEN 5000
#define MAXIPTABLE 1000000
#define NEXTHOPTYPE unsigned char
#define NEXTHOPCOMBINATIONTYPE unsigned short
#define MAXUPDATEINONEBGP 200
#define MAXBGP 5000
#define FREEBREAKPOINT 10


#define CELL unsigned char
#define MAXBREAKPOINTS 1600    /* 600  */
#define NOROUTE 0
#define MAXROW 15000

typedef enum {
  NOTHING,
  LOAD,
  TEST,
  UPDATE
} command_type;

typedef enum {
  WITHDRAW,
  ANNOUNCE
} update_type;

typedef struct {
  command_type command;
  char *inputfilename;
  char *outputfilename;
} t_toexecute;

typedef struct {
  unsigned short ipfrom;
  unsigned char netmask;
} t_singleroute_key;

typedef struct {
  unsigned short ipto;
  NEXTHOPTYPE nexthop;
} t_singleroute_data;


typedef struct {
  unsigned int low_ip;
  NEXTHOPCOMBINATIONTYPE destination;
} t_singlebreakpoint;

typedef struct {
  unsigned short breaklen;
  t_singlebreakpoint singlebreak[MAXBREAKPOINTS];
} t_breakpoint;

typedef struct t_destinationtreetag {
  struct t_destinationtreetag *left;
  struct t_destinationtreetag *right;
  CELL *row;
} t_destinationtree;

typedef struct {
  update_type command;
  NEXTHOPTYPE nexthop;
  unsigned short n_key;
  unsigned short high_ip[MAXUPDATEINONEBGP];
  t_singleroute_key key[MAXUPDATEINONEBGP];
} t_bgpupdate;

CELL *rowindex[65536];
CELL *nullrow;

unsigned int n_totalbreakpoint;
t_breakpoint breakpoint;
unsigned int mergedbreakpoint[MAXBREAKPOINTS];
unsigned int newbreakpointvalue[MAXBREAKPOINTS];
unsigned short newbreakpointposition[MAXBREAKPOINTS];
t_toexecute toexecute[MAXTOEXECUTE];
unsigned int nexthophash[(NEXTHOPTYPE)-1];
NEXTHOPTYPE nexthopcombinationhash[(NEXTHOPCOMBINATIONTYPE)-1][MAXNEXTHOP];
CELL destinationrow[MAXBREAKPOINTS];
unsigned short ndestinationtable=0;
t_destinationtree *destinationtree=NULL;
unsigned short iptable[MAXIPTABLE][2];
CELL ipresult[MAXIPTABLE];
t_bgpupdate bgpupdate[MAXBGP];
unsigned short colindex[65536];
unsigned int n_nexthop=0;
unsigned int n_nexthopcombination=0;
unsigned int n_collision=0;
unsigned int n_notfound;
unsigned int n_route=0;
unsigned int n_mergedbreakpoint=0;
unsigned int n_newbreakpoint=0;
unsigned int n_iptable=0;
unsigned int n_bgpupdate=0;
unsigned int n_totalupdatecommand=0;
unsigned char Rand8[256];
unsigned char IPTOMASK[256];
unsigned int HIGHNETMASKWIDTH[33] = { 65536,32768,16384,8192,4096,2048,1024,512,256,128,64,32,16,8,4,2,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1 };
unsigned int LOWNETMASKWIDTH[33] = { 65535,65535,65535,65535,65535,65535,65535,65535,65535,65535,65535,65535,65535,65535,65535,65535,65535,32767,16383,8191,4095,2047,1023,511,255,127,63,31,15,7,3,1,0 };

struct rusage TimeStart_createtable,TimeStart_updatetable,TimeStart_testtable;
unsigned int TimeElapsedcreatetable,TimeElapsed_updatetable,TimeElapsed_testtable;

struct rusage TimeStart_testiptable,TimeStart_voidtestiptable,TimeStart_makeupdate;
unsigned int n_makeupdate;
unsigned  int TimeElapsed_testiptable,TimeElapsed_voidtestiptable,TimeElapsed_makeupdate;

#ifdef DEBUG
unsigned int maxbreakpoint;
struct rusage TimeStart_insertrbt,TimeStart_nexthophash,TimeStart_nexthopcombinationhash,TimeStart_createsinglebreakpoint,TimeStart_createsinglebreakpoint2,TimeStart_mergebreakpoint,TimeStart_createdestinationrow,TimeStart_searchdestinationtree,TimeStart_createcolindex;
struct rusage TimeEnd;
unsigned int n_insertrbt,n_nexthophash,n_nexthopcombinationhash,n_createsinglebreakpoint,n_createsinglebreakpoint2,n_mergebreakpoint,n_createdestinationrow,n_searchdestinationtree;
unsigned int TimeElapsed_insertrbt,TimeElapsed_nexthophash,TimeElapsed_nexthopcombinationhash,TimeElapsed_createsinglebreakpoint,TimeElapsed_createsinglebreakpoint2,TimeElapsed_mergebreakpoint,TimeElapsed_createdestinationrow,TimeElapsed_searchdestinationtree,TimeElapsed_createcolindex;
#endif

void StartTime(struct rusage *start)
{
  getrusage(RUSAGE_SELF,start);
}

unsigned int TimeElapsed(struct rusage start)
{
  struct rusage end;
  getrusage(RUSAGE_SELF,&end);
  return((end.ru_utime.tv_sec*1000000+end.ru_utime.tv_usec)-(start.ru_utime.tv_sec*1000000+start.ru_utime.tv_usec));
}



/*******************************************************************/
/* red-black tree */

#include <stdarg.h>

/* implementation dependend declarations */
typedef enum {
    STATUS_OK,
    STATUS_MEM_EXHAUSTED,
    STATUS_DUPLICATE_KEY,
    STATUS_KEY_NOT_FOUND
} statusEnum;

typedef t_singleroute_key keyType;            /* type of key */

/* user data stored in tree */
typedef t_singleroute_data recType;

#define compLT(a,b) ((a.netmask < b.netmask) || ((a.netmask == b.netmask) && (a.ipfrom < b.ipfrom)))
#define compEQ(a,b) ((a.ipfrom == b.ipfrom) && (a.netmask == b.netmask))

/* implementation independent declarations */
/* Red-Black tree description */
typedef enum { BLACK, RED } nodeColor;

typedef struct nodeTag {
    struct nodeTag *left;       /* left child */
    struct nodeTag *right;      /* right child */
    struct nodeTag *parent;     /* parent */
    nodeColor color;            /* node color (BLACK, RED) */
    keyType key;                /* key used for searching */
    recType rec;                /* user data */
} nodeType;

#define NIL &sentinel           /* all leafs are sentinels */
static nodeType sentinel = { NIL, NIL, 0, BLACK, 0};

/* last node found, optimizes find/delete operations */
static nodeType *lastFind;

static nodeType *root = NIL;    /* root of Red-Black tree */

static void rotateLeft(nodeType *x) {

   /**************************
    *  rotate node x to left *
    **************************/

    nodeType *y = x->right;

    /* establish x->right link */
    x->right = y->left;
    if (y->left != NIL) y->left->parent = x;

    /* establish y->parent link */
    if (y != NIL) y->parent = x->parent;
    if (x->parent) {
        if (x == x->parent->left)
            x->parent->left = y;
        else
            x->parent->right = y;
    } else {
        root = y;
    }

    /* link x and y */
    y->left = x;
    if (x != NIL) x->parent = y;
}

static void rotateRight(nodeType *x) {

   /****************************
    *  rotate node x to right  *
    ****************************/

    nodeType *y = x->left;

    /* establish x->left link */
    x->left = y->right;
    if (y->right != NIL) y->right->parent = x;

    /* establish y->parent link */
    if (y != NIL) y->parent = x->parent;
    if (x->parent) {
        if (x == x->parent->right)
            x->parent->right = y;
        else
            x->parent->left = y;
    } else {
        root = y;
    }

    /* link x and y */
    y->right = x;
    if (x != NIL) x->parent = y;
}

static void insertFixup(nodeType *x) {

   /*************************************
    *  maintain Red-Black tree balance  *
    *  after inserting node x           *
    *************************************/

    /* check Red-Black properties */
    while (x != root && x->parent->color == RED) {
        /* we have a violation */
        if (x->parent == x->parent->parent->left) {
            nodeType *y = x->parent->parent->right;
            if (y->color == RED) {

                /* uncle is RED */
                x->parent->color = BLACK;
                y->color = BLACK;
                x->parent->parent->color = RED;
                x = x->parent->parent;
            } else {

                /* uncle is BLACK */
                if (x == x->parent->right) {
                    /* make x a left child */
                    x = x->parent;
                    rotateLeft(x);
                }

                /* recolor and rotate */
                x->parent->color = BLACK;
                x->parent->parent->color = RED;
                rotateRight(x->parent->parent);
            }
        } else {

            /* mirror image of above code */
            nodeType *y = x->parent->parent->left;
            if (y->color == RED) {

                /* uncle is RED */
                x->parent->color = BLACK;
                y->color = BLACK;
                x->parent->parent->color = RED;
                x = x->parent->parent;
            } else {

                /* uncle is BLACK */
                if (x == x->parent->left) {
                    x = x->parent;
                    rotateRight(x);
                }
                x->parent->color = BLACK;
                x->parent->parent->color = RED;
                rotateLeft(x->parent->parent);
            }
        }
    }
    root->color = BLACK;
}

statusEnum insert(keyType key, recType *rec) {
    nodeType *current, *parent, *x;

#ifdef DEBUG
    n_insertrbt++;
    StartTime(&TimeStart_insertrbt);
#endif

   /***********************************************
    *  allocate node for data and insert in tree  *
    ***********************************************/

    /* find future parent */
    current = root;
    parent = 0;
    while (current != NIL) {
        if (compEQ(key, current->key))
            return STATUS_DUPLICATE_KEY;
        parent = current;
        current = compLT(key, current->key) ?
            current->left : current->right;
    }

    /* setup new node */
    if ((x = malloc (sizeof(*x))) == 0)
        return STATUS_MEM_EXHAUSTED;
    x->parent = parent;
    x->left = NIL;
    x->right = NIL;
    x->color = RED;
    x->key = key;
    x->rec = *rec;

    /* insert node in tree */
    if(parent) {
        if(compLT(key, parent->key))
            parent->left = x;
        else
            parent->right = x;
    } else {
        root = x;
    }

    insertFixup(x);
    lastFind = NULL;

#ifdef DEBUG
    TimeElapsed_insertrbt+=TimeElapsed(TimeStart_insertrbt);
#endif
    return STATUS_OK;
}

static void deleteFixup(nodeType *x) {

   /*************************************
    *  maintain Red-Black tree balance  *
    *  after deleting node x            *
    *************************************/

    while (x != root && x->color == BLACK) {
        if (x == x->parent->left) {
            nodeType *w = x->parent->right;
            if (w->color == RED) {
                w->color = BLACK;
                x->parent->color = RED;
                rotateLeft (x->parent);
                w = x->parent->right;
            }
            if (w->left->color == BLACK && w->right->color == BLACK) {
                w->color = RED;
                x = x->parent;
            } else {
                if (w->right->color == BLACK) {
                    w->left->color = BLACK;
                    w->color = RED;
                    rotateRight (w);
                    w = x->parent->right;
                }
                w->color = x->parent->color;
                x->parent->color = BLACK;
                w->right->color = BLACK;
                rotateLeft (x->parent);
                x = root;
            }
        } else {
            nodeType *w = x->parent->left;
            if (w->color == RED) {
                w->color = BLACK;
                x->parent->color = RED;
                rotateRight (x->parent);
                w = x->parent->left;
            }
            if (w->right->color == BLACK && w->left->color == BLACK) {
                w->color = RED;
                x = x->parent;
            } else {
                if (w->left->color == BLACK) {
                    w->right->color = BLACK;
                    w->color = RED;
                    rotateLeft (w);
                    w = x->parent->left;
                }
                w->color = x->parent->color;
                x->parent->color = BLACK;
                w->left->color = BLACK;
                rotateRight (x->parent);
                x = root;
            }
        }
    }
    x->color = BLACK;
}

statusEnum delete(keyType key) {
    nodeType *x, *y, *z;

   /*****************************
    *  delete node z from tree  *
    *****************************/

    /* find node in tree */
    if (lastFind && compEQ(lastFind->key, key))
        /* if we just found node, use pointer */
        z = lastFind;
    else {
        z = root;
        while(z != NIL) {
            if(compEQ(key, z->key))
                break;
            else
                z = compLT(key, z->key) ? z->left : z->right;
        }
        if (z == NIL) return STATUS_KEY_NOT_FOUND;
    }

    if (z->left == NIL || z->right == NIL) {
        /* y has a NIL node as a child */
        y = z;
    } else {
        /* find tree successor with a NIL node as a child */
        y = z->right;
        while (y->left != NIL) y = y->left;
    }

    /* x is y's only child */
    if (y->left != NIL)
        x = y->left;
    else
        x = y->right;

    /* remove y from the parent chain */
    x->parent = y->parent;
    if (y->parent)
        if (y == y->parent->left)
            y->parent->left = x;
        else
            y->parent->right = x;
    else
        root = x;

    if (y != z) {
        z->key = y->key;
        z->rec = y->rec;
    }


    if (y->color == BLACK)
        deleteFixup (x);

    free (y);
    lastFind = NULL;

    return STATUS_OK;
}

statusEnum find(keyType key, recType *rec) {

   /*******************************
    *  find node containing data  *
    *******************************/

    nodeType *current = root;
    while(current != NIL) {
        if(compEQ(key, current->key)) {
            *rec = current->rec;
            lastFind = current;
            return STATUS_OK;
        } else {
            current = compLT (key, current->key) ?
                current->left : current->right;
        }
    }
    return STATUS_KEY_NOT_FOUND;
}

/*******************************************************************/

nodeType *routetable[65536];

void init(int argc,char *argv[])
     /* Initialize variable and decode command line arguments filling toexecute array */
{

  int i,j;

  j=0;
  i=1;

  while (i<argc)
    {
      if (strcasecmp(argv[i],"load")==0)
        toexecute[j].command=1;
      else if (strcasecmp(argv[i],"test")==0)
        toexecute[j].command=2;
      else if (strcasecmp(argv[i],"update")==0)
        toexecute[j].command=3;
      else if (strcasecmp(argv[i],"next")==0)
        j++;
      else if (strcasecmp(argv[i],"input")==0)
        toexecute[j].inputfilename=argv[++i];
      else if (strcasecmp(argv[i],"output")==0)
        toexecute[j].outputfilename=argv[++i];
      i++;
    }
  if (toexecute[0].command==0) {
    fprintf(stderr,"iplookup 1.0.0 - (c) 1999 - L. Dardini, R. Grossi, P. Crescenzi\n");
    fprintf(stderr,"Usage:\niplookup [ [load] [test] [update] [next] ] [input filename] [output filename]\n");
    exit(-1);
  }

  /* Initialize IPTOMASK for rapid ip to netmask conversion */

  j=1;
  for (i=8;i>=0;i--)
    {
      IPTOMASK[256-j]=i;
      j*=2;
    }

  /* Initialize routetable with NIL */

  for (i=0;i<65536;i++)
    routetable[i]=NIL;

  /* Initialize nexthop hash table */

  for (i=0;i<(NEXTHOPTYPE)-1;i++)
    nexthophash[i]=0;

  /* Initialize nexthop combination hash table */

  for (i=0;i<(NEXTHOPCOMBINATIONTYPE)-1;i++)
    for (j=0;j<MAXNEXTHOP;j++)
      nexthopcombinationhash[i][j]=0;

  /* Initialize random values for hash function */

  for (i=0;i<256;i++)
    Rand8[i]=(unsigned char)(255.0*rand()/(RAND_MAX+1.0));



  /* Initialize null destination row for unroutable address */

  nullrow=calloc(MAXBREAKPOINTS,sizeof(CELL));

#ifdef DEBUG
  TimeElapsed_insertrbt=0;
  TimeElapsed_nexthophash=0;
  TimeElapsed_nexthopcombinationhash=0;
  TimeElapsed_createsinglebreakpoint=0;
  TimeElapsed_mergebreakpoint=0;
  TimeElapsed_searchdestinationtree=0;
  TimeElapsed_createcolindex=0;
  TimeElapsed_testiptable=0;
  TimeElapsed_voidtestiptable=0;
  n_insertrbt=0;
  n_nexthophash=0;
  n_nexthopcombinationhash=0;
  n_createsinglebreakpoint=0;
  n_mergebreakpoint=0;
  n_notfound=0;
  n_searchdestinationtree=0;
#endif
  mergedbreakpoint[0]=0;
  mergedbreakpoint[1]=65536;
  n_mergedbreakpoint=2;
  n_route=0;
}

NEXTHOPTYPE insertnexthophash(unsigned int ipnh[4])
{
  NEXTHOPTYPE hashvalue;
  unsigned int ipn;
  int toreturn;

#ifdef DEBUG
  n_nexthophash++;
  StartTime(&TimeStart_nexthophash);
#endif

  toreturn=-1;
  ipn=(ipnh[0]<<24)+(ipnh[1]<<16)+(ipnh[2]<<8)+ipnh[3];
  hashvalue=ipnh[3];
  while (toreturn<0)
    {
      if (nexthophash[hashvalue]==ipn)
        toreturn=hashvalue;
      else if(nexthophash[hashvalue]==0)
        {
          nexthophash[hashvalue]=ipn;
          toreturn=hashvalue;
          n_nexthop++;
        }
      else
        {
          hashvalue++;
#ifdef DEBUG
          printf("Nexthophash: collision\n");
#endif
        }
    }
#ifdef DEBUG
  TimeElapsed_nexthophash+=TimeElapsed(TimeStart_nexthophash);
#endif

  return(toreturn);
}

NEXTHOPCOMBINATIONTYPE insertnexthopcombinationhash(NEXTHOPTYPE nexthop[MAXNEXTHOP],unsigned short n_nexthop)
{
  int i;
  NEXTHOPCOMBINATIONTYPE hashvalue;
  int toreturn;
  unsigned char h1,h2;


#ifdef DEBUG
  n_nexthopcombinationhash++;
  StartTime(&TimeStart_nexthopcombinationhash);
#endif

  toreturn=-1;
  nexthop[n_nexthop]=0;
  nexthop[n_nexthop+1]=0;

  h1=nexthop[0];
  h2=nexthop[1];
  i=2;
  while (nexthop[i]!=0)
    {
      h1=Rand8[h1 ^ nexthop[i]];
      h2=Rand8[h2 ^ nexthop[i]];
      i++;
    }
  hashvalue=((NEXTHOPCOMBINATIONTYPE)h1 << 8) | (NEXTHOPCOMBINATIONTYPE)h2;
  while (toreturn<0)
    {
      if (memcmp(nexthopcombinationhash[hashvalue],nexthop,n_nexthop)==0)
        toreturn=hashvalue;
      else if (nexthopcombinationhash[hashvalue][0]==0)
        {
          memcpy(nexthopcombinationhash[hashvalue],nexthop,n_nexthop);
          toreturn=hashvalue;
          n_nexthopcombination++;
        }
      else
        {
          hashvalue++;
          if (hashvalue==NOROUTE)
            hashvalue++;
#ifdef DEBUG
          n_collision++;
#endif
        }
    }

#ifdef DEBUG
  TimeElapsed_nexthopcombinationhash+=TimeElapsed(TimeStart_nexthopcombinationhash);
#endif

  return(toreturn);
}

void loadroutetable(char *inputfilename,char *outputfilename)
     /* load a routing table */
{

  FILE *filein;
  FILE *fileout;
  char *string;
  unsigned int ipn[8];
  unsigned int ipnh[4];
  char *nexthops;
  char *singlenexthop;
  t_singleroute_key key;
  t_singleroute_data data;
  unsigned short high_ip;
  unsigned short low_ip;
  unsigned char netmask;
  int i;
  statusEnum status;
  int result;
  int i_nexthop;
  unsigned int oldhigh_ip=0;
  unsigned int maxnhigh_ip=0;
  unsigned int nhigh_ip=0;
  unsigned int avgnhigh_ip=0;
  unsigned int distincthigh_ip=0;
  unsigned int ipnumber;
  unsigned int netmasknumber;
  string=malloc(MAXSTRINGLEN*sizeof(char));
  nexthops=malloc(MAXSTRINGLEN*sizeof(char));
  singlenexthop=malloc(MAXSTRINGLEN*sizeof(char));

#ifdef DEBUG
  printf("Loading route table...\n");
#endif
  filein=fopen(inputfilename,"r");
  while (fgets(string,MAXSTRINGLEN-1,filein)!=NULL) {
    nexthops[0]=0;
    singlenexthop[0]=0;
    if (sscanf(string,"%d.%d.%d.%d %d.%d.%d.%d %[1234567890. ]",&ipn[0],&ipn[1],&ipn[2],&ipn[3],&ipn[4],&ipn[5],&ipn[6],&ipn[7],nexthops)>0)
        {
          n_route++;
          ipnumber=(ipn[0]<<24)+(ipn[1]<<16)+(ipn[2]<<8)+ipn[3];
          netmasknumber=(ipn[4]<<24)+(ipn[5]<<16)+(ipn[6]<<8)+ipn[7];
          if ((ipnumber & netmasknumber) != ipnumber) {
            printf("Error in ip route: %d.%d.%d.%d excedes netmask %d.%d.%d.%d\n",ipn[0],ipn[1],ipn[2],ipn[3],ipn[4],ipn[5],ipn[6],ipn[7]);
            ipn[0]=ipn[0] & ipn[4];
            ipn[1]=ipn[1] & ipn[5];
            ipn[2]=ipn[2] & ipn[6];
            ipn[3]=ipn[3] & ipn[7];
            ipnumber=(ipn[0]<<24)+(ipn[1]<<16)+(ipn[2]<<8)+ipn[3];
            netmasknumber=(ipn[4]<<24)+(ipn[5]<<16)+(ipn[6]<<8)+ipn[7];
            printf("Corrected in %d.%d.%d.%d netmask %d.%d.%d.%d\n",ipn[0],ipn[1],ipn[2],ipn[3],ipn[4],ipn[5],ipn[6],ipn[7]);
          }
          sscanf(nexthops,"%[1234567890.] %[1234567890. ]",singlenexthop,nexthops);
          if (sscanf(singlenexthop,"%d.%d.%d.%d",&ipnh[0],&ipnh[1],&ipnh[2],&ipnh[3])==4)
            data.nexthop=insertnexthophash(ipnh);
          else
            data.nexthop=NOROUTE;

          high_ip=ipn[0]*256+ipn[1];
          low_ip=ipn[2]*256+ipn[3];

          if (high_ip==oldhigh_ip)
            nhigh_ip++;
          else {
            oldhigh_ip=high_ip;
            avgnhigh_ip+=nhigh_ip;
            nhigh_ip=1;
            distincthigh_ip++;
          }
          if (maxnhigh_ip<nhigh_ip)
            maxnhigh_ip=nhigh_ip;

          netmask=IPTOMASK[ipn[4]]+IPTOMASK[ipn[5]]+IPTOMASK[ipn[6]]+IPTOMASK[ipn[7]];
          for (i=0;i<HIGHNETMASKWIDTH[netmask];i++)
            {
              key.ipfrom=low_ip;
              key.netmask=netmask;
              data.ipto=low_ip+LOWNETMASKWIDTH[netmask];
              root=routetable[high_ip+i];
              status=insert(key,&data);
              if (status!=STATUS_OK) {
                printf("Error on insert... exiting\n");
                exit(-1);
                }
              routetable[high_ip+i]=root;
            }
        }
  }
  fclose(filein);

#ifdef DEBUG
  printf("Routing table loaded: found %d distinct nexthop in %d route, %d distinct combination with %d collision in nexthopcombinationhash\n",n_nexthop,n_route,n_nexthopcombination,n_collision);
  printf("Elapsed time in 'insertrbt'                    %8d us with %8d calls\n",TimeElapsed_insertrbt,n_insertrbt);
  printf("                'insertnexthophash'            %8d us      %8d      \n",TimeElapsed_nexthophash,n_nexthophash);
  printf("                'insertnextcombinationhophash' %8d us      %8d      \n",TimeElapsed_nexthopcombinationhash,n_nexthopcombinationhash);
  printf("Found %d max elements for a single 16 bit route with an average of %d (%d / %d)\n",maxnhigh_ip,avgnhigh_ip/distincthigh_ip,avgnhigh_ip,distincthigh_ip);
#endif

}

void visittree(nodeType *node)
{
  unsigned short start,end,middle,i;
  unsigned int iptoplusone;

  if (node==NIL)
    return;

  visittree(node->left);

  iptoplusone=node->rec.ipto+1;
  start=0;
  end=breakpoint.breaklen-1;

  while ((start+1)<end)
    {
      middle=(start+end)/2;
      if (breakpoint.singlebreak[middle].low_ip>node->key.ipfrom)
        end=middle;
      else
        start=middle;
    }
    if (breakpoint.singlebreak[start].destination!=node->rec.nexthop)
    {
      if ((breakpoint.singlebreak[start].low_ip<node->key.ipfrom) && (breakpoint.singlebreak[start+1].low_ip>iptoplusone))
        {
          for (i=(breakpoint.breaklen+=2)-1;i>start+2;i--)
            {
              breakpoint.singlebreak[i].low_ip=breakpoint.singlebreak[i-2].low_ip;
              breakpoint.singlebreak[i].destination=breakpoint.singlebreak[i-2].destination;
            }
          breakpoint.singlebreak[start+1].low_ip=node->key.ipfrom;
          breakpoint.singlebreak[start+1].destination=node->rec.nexthop;
          breakpoint.singlebreak[start+2].low_ip=iptoplusone;
          breakpoint.singlebreak[start+2].destination=breakpoint.singlebreak[start].destination;
        }
      else
        if (breakpoint.singlebreak[start].low_ip<node->key.ipfrom)
          {
              if (((start+2)<breakpoint.breaklen) & (breakpoint.singlebreak[start+1].destination==node->rec.nexthop))
                {
                  breakpoint.singlebreak[start+1].low_ip=node->key.ipfrom;
                }
              else
                {
                  for (i=breakpoint.breaklen++;i>start;i--)
                    {
                      breakpoint.singlebreak[i].low_ip=breakpoint.singlebreak[i-1].low_ip;
                      breakpoint.singlebreak[i].destination=breakpoint.singlebreak[i-1].destination;
                    }
                }
            breakpoint.singlebreak[start+1].low_ip=node->key.ipfrom;
            breakpoint.singlebreak[start+1].destination=node->rec.nexthop;
          }
        else
          if (breakpoint.singlebreak[end].low_ip>iptoplusone)
            {
                if ((start>0) && (breakpoint.singlebreak[start-1].destination==node->rec.nexthop))
                {
                  breakpoint.singlebreak[start].low_ip=iptoplusone;
                }
              else
                {
                  for (i=breakpoint.breaklen++;i>end;i--)
                    {
                      breakpoint.singlebreak[i].low_ip=breakpoint.singlebreak[i-1].low_ip;
                      breakpoint.singlebreak[i].destination=breakpoint.singlebreak[i-1].destination;
                    }
                  breakpoint.singlebreak[start+1].low_ip=iptoplusone;
                  breakpoint.singlebreak[start+1].destination=breakpoint.singlebreak[start].destination;
                  breakpoint.singlebreak[start].destination=node->rec.nexthop;
                }
            }
      else
      breakpoint.singlebreak[start].destination=node->rec.nexthop;
    }
  visittree(node->right);
}

void createsinglebreakpoint(unsigned short high_ip)
{

#ifdef DEBUG
  n_createsinglebreakpoint++;
  StartTime(&TimeStart_createsinglebreakpoint);
#endif

  breakpoint.breaklen=2;
  breakpoint.singlebreak[0].low_ip=0;
  breakpoint.singlebreak[0].destination=NOROUTE;
  breakpoint.singlebreak[1].low_ip=65536;
  breakpoint.singlebreak[1].destination=NOROUTE;
  visittree(routetable[high_ip]);

#ifdef DEBUG
  TimeElapsed_createsinglebreakpoint+=TimeElapsed(TimeStart_createsinglebreakpoint);
#endif
}


void mergebreakpoint()
{
  unsigned int start,end,middle;
  unsigned char found=0;
  unsigned int i,j;

#ifdef DEBUG
  n_mergebreakpoint++;
  StartTime(&TimeStart_mergebreakpoint);
#endif

 n_newbreakpoint=0;
 for (i=0;i<breakpoint.breaklen;i++)
   {
     start=0;
     end=n_mergedbreakpoint-1;
     found=0;
     do
       {
          middle=(end+start)/2;
          if (mergedbreakpoint[middle]>breakpoint.singlebreak[i].low_ip)
            end=middle-1;
          else if (mergedbreakpoint[middle]<breakpoint.singlebreak[i].low_ip)
            start=middle+1;
          else
            found=1;
       }
     while ((found==0) && (start<=end));
     if (found==0)
       {
         newbreakpointvalue[n_newbreakpoint]=breakpoint.singlebreak[i].low_ip;
         newbreakpointposition[n_newbreakpoint++]=start;
         for (j=n_mergedbreakpoint++;j>start;j--)
           mergedbreakpoint[j]=mergedbreakpoint[j-1];
         mergedbreakpoint[start]=breakpoint.singlebreak[i].low_ip;
       }
   }

#ifdef DEBUG
 TimeElapsed_mergebreakpoint+=TimeElapsed(TimeStart_mergebreakpoint);
#endif
}

void createbreakpoint()
{

  int j;

#ifdef DEBUG
  printf("Creating breakpoint...\n");
#endif

  for (j=0;j<65536;j++)
    {
      if (routetable[j]!=NIL)
        {
          createsinglebreakpoint(j);
          mergebreakpoint();
        }
#ifdef DEBUG
      if (breakpoint.breaklen>maxbreakpoint)
        maxbreakpoint=breakpoint.breaklen;
#endif
    }

  n_totalbreakpoint=n_mergedbreakpoint+FREEBREAKPOINT;

#ifdef DEBUG
  printf("Found a maximum of %d breakpoints in a 16 bit route, a total of %d merged breakpoint\n",maxbreakpoint,n_mergedbreakpoint);
  printf("Elapsed time in 'createsinglebreakpoint'       %8d us with %8d calls\n",TimeElapsed_createsinglebreakpoint,n_createsinglebreakpoint);
  printf("                'mergebreakpoint'              %8d us with %8d calls\n",TimeElapsed_mergebreakpoint,n_mergebreakpoint);
#endif

}

void createdestinationrow()
{

  unsigned int i,j;
  unsigned int start,end,middle,found;
  unsigned int ileft;

#ifdef DEBUG
  n_createdestinationrow++;
  StartTime(&TimeStart_createdestinationrow);
#endif


  start=0;
  ileft=0;
  for (i=1;i<breakpoint.breaklen;i++)
    {
      end=n_mergedbreakpoint-1;
      found=0;

      do
        {
          middle=(end+start)/2;
          if (mergedbreakpoint[middle]>breakpoint.singlebreak[i].low_ip)
            end=middle-1;
          else if (mergedbreakpoint[middle]<breakpoint.singlebreak[i].low_ip)
            start=middle+1;
          else
            found=1;
        }
      while (found==0);
      for (j=ileft;j<middle;j++)
        destinationrow[j]=breakpoint.singlebreak[i-1].destination;
      start=middle+1;
      ileft=middle;
    }

#ifdef DEBUG
  TimeElapsed_createdestinationrow+=TimeElapsed(TimeStart_createdestinationrow);
#endif

}

int memacmp(unsigned char *s1,unsigned char *s2,unsigned int len)
{
  int i;

  i=0;
  do
    {
      i++;
    }
  while ((i<(len-1)) && (s1[i]==s2[i]));
  if (s1[i]<s2[i])
    return(-1);
  else if (s1[i]==s2[i])
    return(0);
  else
    return(1);
}



CELL *searchdestinationtree(t_destinationtree **node)


{
  int result,i;

#ifdef DEBUG
  n_searchdestinationtree++;
  StartTime(&TimeStart_searchdestinationtree);
#endif

  if ((*node)==NULL)
    {
      if (((*node)=malloc(sizeof(t_destinationtree))) == NULL )
        {
          printf("Memory allocation error....\n");
          exit(-1);
        }
      (*node)->left=NULL;
      (*node)->right=NULL;
      (*node)->row=malloc(sizeof(CELL)*n_totalbreakpoint);
      memcpy((*node)->row,destinationrow,sizeof(CELL)*n_mergedbreakpoint);

#ifdef DEBUG
  TimeElapsed_searchdestinationtree+=TimeElapsed(TimeStart_searchdestinationtree);
#endif

      ndestinationtable++;
      return((*node)->row);
    }

  result=memcmp((unsigned char*)destinationrow,(unsigned char*)((*node)->row),sizeof(CELL)*n_mergedbreakpoint);
  if (result>0)
      return(searchdestinationtree(&(*node)->left));
  else if (result<0)
      return(searchdestinationtree(&(*node)->right));
  else
    {
#ifdef DEBUG
  TimeElapsed_searchdestinationtree+=TimeElapsed(TimeStart_searchdestinationtree);
#endif
      return((*node)->row);
    }
}

void createrowindex()
{
  unsigned int i;

#ifdef DEBUG
  TimeElapsed_createsinglebreakpoint2=TimeElapsed_createsinglebreakpoint;
  n_createsinglebreakpoint2=n_createsinglebreakpoint;
  TimeElapsed_createsinglebreakpoint=0;
  n_createsinglebreakpoint=0;
  printf("Creating rowindex...\n");
#endif


  for (i=0;i<65536;i++)
    {
      if (routetable[i]!=NIL) {
        createsinglebreakpoint(i);
        createdestinationrow();
        rowindex[i]=searchdestinationtree(&destinationtree);
      }
      else
        rowindex[i]=nullrow;
    }
#ifdef DEBUG
  printf("Found %d different rows\n",ndestinationtable);
  printf("Elapsed time in 'createsinglebreakpoint'       %8d us with %8d calls\n",TimeElapsed_createsinglebreakpoint,n_createsinglebreakpoint);
  printf("                'createdestinationrow'         %8d us with %8d calls\n",TimeElapsed_createdestinationrow,n_createdestinationrow);
  printf("                'searchdestinationtree'        %8d us with %8d calls\n",TimeElapsed_searchdestinationtree,n_searchdestinationtree);
#endif
}

void createcolindex()
{
  int i,j;

#ifdef DEBUG
  StartTime(&TimeStart_createcolindex);
#endif

  for (i=0;i<n_mergedbreakpoint;i++)
    for (j=mergedbreakpoint[i];j<mergedbreakpoint[i+1];j++)
      colindex[j]=i;
#ifdef DEBUG
  TimeElapsed_createcolindex+=TimeElapsed(TimeStart_createcolindex);
#endif


}

void loadiptable(char *inputfilename,char *outputfilename)
{

  FILE *filein;
  char string[MAXSTRINGLEN];
  unsigned int ipn[4];

  filein=fopen(inputfilename,"r");

  while (fgets(string,MAXSTRINGLEN-1,filein)!=NULL)
    if (sscanf(string,"%d.%d.%d.%d",&ipn[0],&ipn[1],&ipn[2],&ipn[3])>0)
      {
        iptable[n_iptable][0]=(ipn[0]<<8)+ipn[1];
        iptable[n_iptable++][1]=(ipn[2]<<8)+ipn[3];
      }
  fclose(filein);
#ifdef DEBUG
  printf("Loaded %d ip address\n",n_iptable);
#endif
}

void testiptable()
{
  int i;


  StartTime(&TimeStart_voidtestiptable);

  for (i=0;i<n_iptable;i++)
      ipresult[i]=0;

  TimeElapsed_voidtestiptable+=TimeElapsed(TimeStart_voidtestiptable);

  StartTime(&TimeStart_testiptable);

  for (i=0;i<n_iptable;i++) {
      ipresult[i]=*(rowindex[iptable[i][0]]+colindex[iptable[i][1]]);
  }
  TimeElapsed_testiptable+=TimeElapsed(TimeStart_testiptable);
#ifdef DEBUG
  printf("Elapsed time in 'testiptable'                  %8d us\n",TimeElapsed_testiptable);
  printf("Elapsed time in true searching                 %8d us\n",TimeElapsed_testiptable-TimeElapsed_voidtestiptable);
  printf("Single IP lookup speed (for %8d searchs)  %8d ns\n",n_iptable,(TimeElapsed_testiptable-TimeElapsed_voidtestiptable)*1000/n_iptable);
  printf("Number of IP lookup in a second                %8d\n",(unsigned int)(((double)(n_iptable)/(TimeElapsed_testiptable-TimeElapsed_voidtestiptable)*1000000.0)));

#endif


}

void printresult(char *outputfilename)
{

  int i;
  FILE *fileout;

  fileout=fopen(outputfilename,"w");

  for (i=0;i<n_iptable;i++)
    if (ipresult[i]!=NOROUTE)
  fprintf(fileout,"%d.%d.%d.%d -> %d.%d.%d.%d\n",iptable[i][0] >> 8,iptable[i][0] & 255,iptable[i][1] >> 8,iptable[i][1] & 255,nexthophash[ipresult[i]]>>24,(nexthophash[ipresult[i]]>>16) & 255, (nexthophash[ipresult[i]]>>8)&255,nexthophash[ipresult[i]] & 255);
    else
      fprintf(fileout,"%d.%d.%d.%d ->\n",iptable[i][0] >> 8,iptable[i][0] & 255,iptable[i][1] >> 8,iptable[i][1] & 255);
  fclose(fileout);
}

void loadupdatetable(char *inputfilename,char *outputfilename)
{

  FILE *filein;
  char string[MAXSTRINGLEN];
  unsigned int ipn[4];
  unsigned int ipnm[4];
  int result;
  n_bgpupdate=0;

  filein=fopen(inputfilename,"r");
  if (filein!=NULL)
    {
      while (fgets(string,MAXSTRINGLEN-1,filein)!=NULL) {
        if (sscanf(string,"%d %d.%d.%d.%d %[0123456789. ]",&bgpupdate[n_bgpupdate].command,&ipn[0],&ipn[1],&ipn[2],&ipn[3],string)>0)
          {
            bgpupdate[n_bgpupdate].n_key=0;
            bgpupdate[n_bgpupdate].nexthop=insertnexthophash(ipn);
          }
        do
          {
            result=sscanf(string,"%d.%d.%d.%d %d.%d.%d.%d %[01234567890. ]",&ipn[0],&ipn[1],&ipn[2],&ipn[3],&ipnm[0],&ipnm[1],&ipnm[2],&ipnm[3],string);
            bgpupdate[n_bgpupdate].high_ip[bgpupdate[n_bgpupdate].n_key]=(ipn[0]<<8)+ipn[1];
            bgpupdate[n_bgpupdate].key[bgpupdate[n_bgpupdate].n_key].ipfrom=(ipn[2]<<8)+ipn[3];
            bgpupdate[n_bgpupdate].key[bgpupdate[n_bgpupdate].n_key].netmask=IPTOMASK[ipnm[0]]+IPTOMASK[ipnm[1]]+IPTOMASK[ipnm[2]]+IPTOMASK[ipnm[3]];
            bgpupdate[n_bgpupdate].n_key++;
            n_totalupdatecommand++;
          }
        while (result>8);
        n_bgpupdate++;
      }
      fclose(filein);
#ifdef DEBUG
      printf("Loaded %d bgp update with a total of %d command\n",n_bgpupdate,n_totalupdatecommand);
#endif
    }
}

void updatedestinationtree(t_destinationtree *node)
{
  unsigned int i;

  if (node!=NULL)
    {
      for (i=0;i<n_newbreakpoint;i++)
      memmove(&(node->row[newbreakpointposition[i]]),&(node->row[newbreakpointposition[i]-1]),sizeof(CELL)*n_mergedbreakpoint-i);
      updatedestinationtree(node->left);
      updatedestinationtree(node->right);
    }

}


void makeupdate()
{
  int i,j;
  statusEnum result;
  unsigned int new=0,modified=0;

  t_singleroute_data data;

  n_notfound=0;
  n_makeupdate=0;

  StartTime(&TimeStart_makeupdate);

  for (i=0;i<n_bgpupdate;i++)
    {
      //      printf("@@@@@@@@@@@@@@@@@ %d @@@@@@@@@@@@@@@@@@@@@\n",i);
      switch (bgpupdate[i].command)
        {
        case WITHDRAW:
          for (j=0;j<bgpupdate[i].n_key;j++)
            {
              root=routetable[bgpupdate[i].high_ip[j]];
              result=delete(bgpupdate[i].key[j]);
              if (result==STATUS_KEY_NOT_FOUND)
                n_notfound++;
              else
                {
                  routetable[bgpupdate[i].high_ip[j]]=root;
                  if (bgpupdate[i].high_ip[j+1]!=bgpupdate[i].high_ip[j])
                    {
                      n_makeupdate++;
                      createsinglebreakpoint(bgpupdate[i].high_ip[j]);
                      mergebreakpoint();
                      createdestinationrow();
                      if (n_newbreakpoint>0)
                        {
                          if (n_mergedbreakpoint+n_newbreakpoint<=n_totalbreakpoint)
                            {
                              updatedestinationtree(destinationtree);
                              createcolindex();
                            }
                          else
                            printf("Unhandled case\n");
                        }
                      rowindex[bgpupdate[i].high_ip[j]]=searchdestinationtree(&destinationtree);
                    }
                }
              routetable[bgpupdate[i].high_ip[j]]=root;
            }
          break;

        case ANNOUNCE:

          for (j=0;j<bgpupdate[i].n_key;j++)
            {

              root=routetable[bgpupdate[i].high_ip[j]];
               result=find(bgpupdate[i].key[j],&data);

              if (result==STATUS_OK)
                {
                  data.nexthop=bgpupdate[i].nexthop;
                  delete(bgpupdate[i].key[j]);
                  result=insert(bgpupdate[i].key[j],&data);
                  if (result!=STATUS_OK) {
                    printf("Error on insert... exiting\n");
                    exit(-1);
                  }
                  modified++;
                  routetable[bgpupdate[i].high_ip[j]]=root;
                }

              else
                {
                  data.ipto=bgpupdate[i].key[j].ipfrom+LOWNETMASKWIDTH[bgpupdate[i].key[j].netmask];
                  data.nexthop=bgpupdate[i].nexthop;
                  result=insert(bgpupdate[i].key[j],&data);
                  if (result!=STATUS_OK) {
                    printf("Error on insert... exiting\n");
                    exit(-1);
                  }
                new++;
                }
              routetable[bgpupdate[i].high_ip[j]]=root;
              if (bgpupdate[i].high_ip[j+1]!=bgpupdate[i].high_ip[j])
                {
                  n_makeupdate++;
                  createsinglebreakpoint(bgpupdate[i].high_ip[j]);
                  mergebreakpoint();
                  createdestinationrow();
                  if (n_newbreakpoint>0)
                    {
                      if (n_mergedbreakpoint+n_newbreakpoint<=n_totalbreakpoint)
                        updatedestinationtree(destinationtree);
                      else
                            printf("Unhandled case\n");
                    }
                  rowindex[bgpupdate[i].high_ip[j]]=searchdestinationtree(&destinationtree);
                }
            }
             break;
        }
    }

  TimeElapsed_makeupdate=TimeElapsed(TimeStart_makeupdate);

#ifdef DEBUG
  printf("Processed %d bgpupdate with %d key not found and %d table update\n",n_bgpupdate,n_notfound,n_makeupdate);
  printf("Elapsed time in 'makeupdate'                   %8d us\n",TimeElapsed_makeupdate);
  printf("Number of bgp update in a second               %8d\n",n_bgpupdate*1000/TimeElapsed_makeupdate*1000);
  printf("Elapsed time for a single bgp update           %8d us\n",TimeElapsed_makeupdate/n_bgpupdate);
  printf("Number of table update in a second             %8d\n",n_makeupdate*1000/TimeElapsed_makeupdate*1000);
  printf("Elapsed time for a single table update         %8d us\n",TimeElapsed_makeupdate/n_makeupdate);

#endif

}


int main(int argc,char* argv[])
{

  int i=0;

  init(argc,argv);
  while (toexecute[i].command!=0)
    {
      switch(toexecute[i].command)
        {
        case LOAD:loadroutetable(toexecute[i].inputfilename,toexecute[i].outputfilename);
#ifndef DEBUG
          StartTime(&TimeStart_createtable);
#endif
          createbreakpoint();
          createrowindex();
          createcolindex();

#ifdef DEBUG
          printf("Total elapsed time for creating tables         %8d us\n",TimeElapsed_insertrbt+TimeElapsed_nexthophash+TimeElapsed_nexthopcombinationhash+TimeElapsed_createsinglebreakpoint+TimeElapsed_createsinglebreakpoint2+TimeElapsed_mergebreakpoint+TimeElapsed_createdestinationrow+TimeElapsed_searchdestinationtree+TimeElapsed_createcolindex);
#endif
          i++;
          break;
        case TEST:loadiptable(toexecute[i].inputfilename,toexecute[i].outputfilename);
          testiptable();
          if (toexecute[i].outputfilename!=NULL)
            printresult(toexecute[i].outputfilename);
#ifndef DEBUG
          printf("%s %d %d %d %d %d ",toexecute[0].inputfilename,n_route,n_mergedbreakpoint,ndestinationtable,sizeof(CELL),TimeElapsed(TimeStart_createtable));
          printf("%s %d %d %d %d\n",toexecute[i].inputfilename,n_bgpupdate,n_notfound,n_makeupdate,TimeElapsed_makeupdate);
#endif
          i++;
          break;
        case UPDATE:loadupdatetable(toexecute[i].inputfilename,toexecute[i].outputfilename);
          makeupdate();
#ifndef DEBUG
          printf("%s %d %d %d %d %d ",toexecute[0].inputfilename,n_route,n_mergedbreakpoint,ndestinationtable,sizeof(CELL),TimeElapsed(TimeStart_createtable));
          printf("%s %d %d %d %d\n",toexecute[i].inputfilename,n_bgpupdate,n_notfound,n_makeupdate,TimeElapsed_makeupdate);
#endif
          i++;
          break;
        }
    }
  return(0);
}