#include "iplookup.h"

cell_t **create_match_set (cell_t **block, ip_addr_t *size_r, col_index_t *size_c, ip_addr_t len) {
  cell_t **nodes;
  ip_addr_t i, j, k, h;
  ip_addr_t gr_len;
  printf ("Creating matching set...");fflush (stdout);
  gr_len = (size_c[0] > size_c[1]) ? size_c[0] : size_c[1];
  nodes = (cell_t **) malloc (2 * sizeof (cell_t *));
  if (nodes == NULL) return NULL;
  nodes[0] = (cell_t *) malloc (gr_len * size_r[0] * sizeof (cell_t));
  if (nodes[0] == NULL) {
    free (nodes);
    return NULL;
  }
  nodes[1] = (cell_t *) malloc (gr_len * size_r[1] * sizeof (cell_t));
  if (nodes[1] == NULL) {
    free (nodes);
    free (nodes[0]);    
    return NULL;
  }
  for (h = 0; h < 2; h ++) {
    j = 0;
    for (i = (h * len); j < size_r[h]; i ++)
      if (block[i] != NULL) {
	for (k = 0; k < gr_len; k ++)
	  nodes[h][(k * size_r[h]) + j] = block[i][k];
	j ++;
      }
  }
  printf ("done\n");
  return nodes;
}

col_index_t **create_bipartite_graph (cell_t **nodes, ip_addr_t *size_r, col_index_t *size_c) {
  col_index_t i, j;
  col_index_t **graph;
  ip_addr_t len, gr_len;
  cell_t *str_i, *str_j;
  printf ("Creating bipartite graph...");fflush (stdout);
  len = (size_r[0] > size_r[1]) ? size_r[0] : size_r[1];
  /*  I create a bilanced graph and nodes are in a square matrix  */
  gr_len = (size_c[0] > size_c[1]) ? size_c[0] : size_c[1];
  graph = (col_index_t **) malloc (gr_len * sizeof (col_index_t *));
  if (!graph) return NULL;
  for (i = 0; i < gr_len; i ++) {
    graph[i] = (col_index_t *) malloc (gr_len * sizeof (col_index_t));
    if (!graph[i]) {
      /*  free resurces  */
      for (j = 0; j < i; j ++) free (graph[j]);
      free (graph);
      return NULL;
    }
  }
  str_i = nodes[0] - size_r[0];
  for (i = 0; i < gr_len; i ++) {
    str_i += size_r[0];
    str_j = nodes[1] - size_r[1];
    for (j = 0; j < gr_len; j ++) {
      str_j += size_r[1];
      graph[i][j] = hamming_distance (str_i, str_j, len);
    }
  }
  printf ("done\n");
  return graph;
}

col_index_t hamming_distance (cell_t *str1, cell_t *str2, ip_addr_t len) {
  col_index_t dist = 0;
  ip_addr_t i;
  /*  if first caracter is DONTCARE, every other caracter, by costruction, 
      is DONTCARE too  */
  if (str1[0] == DONTCARE || str2[0] == DONTCARE) return len;
  for (i = 0; i < len; i ++)
    if (str1[i] == str2[i]) dist ++;
  return dist;
}

col_index_t *find_perfect_matching (col_index_t **graph, col_index_t *len) {
  col_index_t i, j, ni, nj, max, glen;
  col_index_t *subs;
  char again;
  printf ("Looking for perfect matching...");fflush (stdout);
  /*  Need to avoid pedentic warning  */
  ni = nj = 0;
  glen = (len[0] > len[1]) ? len[0] : len[1];
  subs = (col_index_t *) malloc (glen * sizeof (col_index_t));
  if (!subs) return NULL;
  do {
    max = again = 0;
    for (i = 0; i < glen; i ++) {
      if (graph[i] != NULL) {
	for (j = 0; j < glen; j ++) {
	  if (graph[i][j] > max) {
	    max = graph[i][j];
	    ni = i;
	    nj = j;
	  }
	}
      }
    }
    /*  subs[shorter size side index] = longer size side index */
    if (len[0] < len[1])
      subs[ni] = nj;
    else
      subs[nj] = ni;
    /*  free and reset useless nodes  */
    free (graph[ni]);
    graph[ni] = NULL;
    for (i = 0; i < glen; i ++)
      if (graph[i] != NULL) {
	graph[i][nj] = 0;
	again = 1;
      }
  } while (again);
  printf ("done\n");
  return subs;
}



void make_permutations (cell_t **block, ip_addr_t rows, col_index_t cols, col_index_t max, col_index_t *col_index, col_index_t *subs) {
  cell_t *tmp, *buff;
  ip_addr_t i, j, k;
  printf ("Peremutating...");fflush (stdout);
  buff = (cell_t *) malloc (max * sizeof (cell_t));
  if (!buff) {
    printf ("Malloc failure, permutation will be ignored...use default\n");
    return;
  }
  j = 0;
  for (i = 0; j < rows; i ++) {
    if (block[i] != NULL) {
      for (k = 0; k < max; k ++) buff[subs[k]] = block[i][k];
      /*  go to next step  */
      j ++;
      tmp = block[i];
      block[i] = buff;
      buff = tmp;
    }
  }
  /*  free  */
  free (buff);
  /*  Adjust colun index  */
  for (i = 0; i < exp (32 - K_SIZE - 1); i ++)
    col_index[i] = subs[col_index[i]];
  printf ("done\n");
}

void destroy_graph (col_index_t **graph, col_index_t len) {
  col_index_t i;
  for (i = 0; i < len; i ++) 
    if (!graph[i]) free (graph[i]);
  free (graph);
}

void destroy_match_set (cell_t **nodes) {
  free (nodes[0]);
  free (nodes[1]);
  free (nodes);
}
