#include "iplookup.h"

long make_lookups (cell_t **row, col_index_t *col, addr_table_t **array, ip_addr_t len, ip_addr_t ntimes) {
  register ip_addr_t i,j;
  Mark_Time ();
  /*  repeat the same operation n times  */
  for (i = 0; i < ntimes; i ++) {
    /*  Each loop is a lookup  */
    for (j = 0; j < len; j ++) {
      switch (array[j]->op) {
      case LOOKUP:
	/*  It's all for this !!!  */
	array[j]->next_hop=row[array[j]->first_pair][col[array[j]->last_pair]];
	break;
      case INSERT:
	{
	  void **up;
	  up = insert_and_rebuild (j);
	  if (up != NULL) {
	    row = up[0];
	    col = up[1];
	  }
	}
	break;
      case DELETE:
	{
	  void **up;
	  up = delete_and_rebuild (j);
	  if (up != NULL) {
	    row = up[0];
	    col = up[1];
	  }
	}
	break;
      } 
    }
  }
  return Mark_Time ();
}

/*  like above make_lookups () but col_index is declared as short type  */
long make_lookups_sc (cell_t **row, col_short_t *col, addr_table_t **array, ip_addr_t len, ip_addr_t ntimes) {
  register ip_addr_t i,j;
  Mark_Time ();
  /*  repeat the same operation n times  */
  for (i = 0; i < ntimes; i ++) {
    /*  Each loop is a lookup  */
    for (j = 0; j < len; j ++) {
      switch (array[j]->op) {
      case LOOKUP:
	/*  It's all for this !!!  */
	array[j]->next_hop=row[array[j]->first_pair][col[array[j]->last_pair]];
	break;
      case INSERT:
	{
	  void **up;
	  up = insert_and_rebuild (j);
	  if (up != NULL) {
	    row = up[0];
	    col = up[1];
	  }
	}
	break;
      case DELETE:
	{
	  void **up;
	  up = delete_and_rebuild (j);
	  if (up != NULL) {
	    row = up[0];
	    col = up[1];
	  }
	}
	break;
      } 
    }
  }
  return Mark_Time ();
}

/*  Bucket lookup procedure  */
long make_bucket_lookups (cell_t **row, bucket_t *col, addr_table_t **array, ip_addr_t len, ip_addr_t ntimes, ip_addr_t size) {
  register ip_addr_t i,j;
  register bucket_t offset;
  register unsigned char shift;
  register col_index_t s_size = size;
  offset = (exp (K_SIZE) / size) - size;
  shift = -1;
  while (size != 0) {
    shift ++;
    size >>= 1;
  }
  Mark_Time ();
  /*  repeat the same operation n times  */
  for (i = 0; i < ntimes; i ++) {
    /*  Each loop is a lookup  */
    for (j = 0; j < len; j ++) {
      switch (array[j]->op) {
      case LOOKUP:
	{
	  /*  It's all for this !!!  */
	  register bucket_t last = col[array[j]->last_pair >> shift];
	  if (last >= 0) 
	    array[j]->next_hop = row[array[j]->first_pair][last];
	  else 
	    array[j]->next_hop = row[array[j]->first_pair][col[offset +((-last) << shift) + (array[j]->last_pair % s_size)]];
	}
	break;
      case INSERT:
	{
	  void **up;
	  up = insert_and_rebuild (j);
	  if (up != NULL) {
	    row = up[0];
	    col = up[1];
	  }
	}
	break;
      case DELETE:
	{
	  void **up;
	  up = delete_and_rebuild (j);
	  if (up != NULL) {
	    row = up[0];
	    col = up[1];
	  }
	}
	break;
      }
    }
  }
  return Mark_Time ();
}

