#include "iplookup.h"

int main (int argc, char *argv[]) {
  addr_table_t **addresses;
  trie_t *trie;
  ip_addr_t tot_leaves = 0;
  void *col;
  char flag, comp_col = 0;  
  col_index_t *col_index = NULL;
  col_index_t row_len[2];
  ip_addr_t tot_rows[2];
  ip_addr_t RLE_len[2];  
  rle_t **RLE_seq[2];  
  cell_t **row_index, **row_ind[2]; 
  ip_addr_t b_size = 0, b_tmp = 0;
  struct net_table *in_table;
  ip_addr_t ntimes = 1;
  char *next_hops[MAX_NEXT_HOPS];
  long tempo;
  long look_time[2];
  ip_addr_t tot_route;
  ip_addr_t i;
#if FINE_TIME
  double cpu_speed;
  my_time_t start_time ,stop_time;
  double total_time;
  printf ("Computing CPU speed for time measurement...");
  fflush (stdout);
  cpu_speed = detect_cpu_clock ();
  printf ("done\n");
#endif
  flag = eval_params (argc, argv, &ntimes);
  /*  if ntimes is 0  there isn't an argoument for address file*/
  if (ntimes == 0) 
    in_table = load_table (argv[argc - 1], next_hops, &tot_route, flag & 1);
  else
    in_table = load_table (argv[argc - 2], next_hops, &tot_route, flag & 1);
  if (in_table == NULL) return -1;
#if FINE_TIME
  rdtscll (start_time);
#else
  Mark_Time ();  
#endif
  /*  Create trie for update operation */
  trie = create_trie (in_table, &tot_leaves);
  /*  In_table was freed and now set pointer to NULL  */
  in_table = NULL;  
  RLE_len[0] = RLE_len[1] = tot_leaves;
  RLE_seq[0] = update_RLE_sequence (trie, &RLE_len[0], 0); /*  sup  */
  if (RLE_seq[0] == NULL) return -1;
  RLE_seq[1] = update_RLE_sequence (trie, &RLE_len[1], 1); /*  inf  */
  if (RLE_seq[1] == NULL) return -1;
  /*  create table back-end  */
  /*  Create colun index  */
  col_index = (col_index_t *) malloc (exp (32-K_SIZE) * sizeof (col_index_t));
  if (col_index == NULL) {
    printf ("Malloc failure creating colunn index...can`t go on\n");
    return -1;
  }
  row_ind[0] = rebuild_table (RLE_seq[0], flag, col_index,
			      RLE_len[0], &tot_rows[0], &row_len[0]);
  if (row_ind[0] == NULL) return -1;
  row_ind[1] = rebuild_table (RLE_seq[1], flag, col_index,
			      RLE_len[1], &tot_rows[1], &row_len[1]);
  if (row_ind[1] == NULL) return -1;
  /*  RLE_seq was freed and now set pointer to NULL  */
  RLE_seq[0] = RLE_seq[1] = NULL;
  /*  Build real row index  */
  row_index = (cell_t **) malloc (exp (K_SIZE + 1) * sizeof (cell_t *));
  if (row_index == NULL) return -1;
  for (i = 0; i < exp (K_SIZE); i ++) row_index[i] = row_ind[0][i];
  for (; i < exp (K_SIZE + 1); i ++) row_index[i] = row_ind[1][i-exp (K_SIZE)];

  /*  Colunn index creation  */
  col = (void *) col_index;
  /*   force normal colun index  */
  if (!(flag & 64)) {
    col = build_col_index ((col_index_t *) col, flag, &row_len[1],
                           &b_size, &b_tmp, &comp_col);
  }
#if FINE_TIME
  rdtscll (stop_time);
  tempo = stop_time - start_time;
  total_time = (1000000 / cpu_speed) * tempo;
  printf ("Elapsed time %f\n", total_time);
#else
  tempo = Mark_Time ();
  printf ("Elapsed time %ld\n", tempo);
#endif
  printf ("Done\n");
  /*  Write cdg.log file if requested */
  if (flag & 1) {
    write_log_file (tot_rows, row_len, b_size, b_tmp, flag, comp_col, tempo);
  }
  /*  If iteration is unspecified was set to 1 from eval_params ()
      if iteration is 0 skip lookup  */
  if (ntimes != 0) {
    addresses = load_address (argv[argc - 1], &tot_route, flag & 8, next_hops);
    if (addresses == NULL) return -1;
    look_time[0] = no_lookup_time (addresses, tot_route, ntimes);
    switch (comp_col) {
    case 0:
      printf ("Normal lookup procedure\n"); 
      look_time[1] = make_lookups (row_index, (col_index_t *) col, addresses,
				  tot_route, ntimes);
      break;
    case 1:
      printf ("Short type lookup procedure\n");
      look_time[1] = make_lookups_sc (row_index, (col_short_t *) col,
				      addresses, tot_route, ntimes);
      break;
    case 2: 
      printf ("bucket lookup procedure\n");
      look_time[1] = make_bucket_lookups (row_index, (bucket_t *) col,
					  addresses, tot_route, ntimes,b_size);
      break;
    }
    printf ("-- Lookup time --\n");
#if FINE_TIME
    total_time = (1000000 / cpu_speed) * look_time[1];
    printf ("Entire lookup procedure %f\n", total_time);
    total_time = (1000000 / cpu_speed) * look_time[0];
    printf ("No lookup operations %f\n", total_time);
    total_time = (1000000 / cpu_speed) * (look_time[1] -look_time[0]);
    printf ("Real lookup time %f\n", total_time);
#else
    printf ("Entire lookup procedure %ld\n",look_time[1]);
    printf ("No lookup operations %ld\n",look_time[0]);
    printf ("Real lookup time %ld\n",look_time[1] - look_time[0]);
#endif
    /*  Add lookup information to cdg.log file if requested */
    if (flag & 1) {
#if FINE_TIME
      write_log_lookup (comp_col, look_time, cpu_speed);
#else
      write_log_lookup (comp_col, look_time);
#endif
    }
    write_results (addresses, tot_route, next_hops);
    unload_address (addresses, tot_route);
    printf ("Done\n");
  }

  unload_next_hops (next_hops);
  return 0;
}
