#include "iplookup.h"


void create_colun_index (col_index_t *col_index, cluster_t **RLE, 
			  ip_addr_t RLE_size, ip_addr_t exp_col) {
  /*  Colun is exp (32 - K_SIZE) row is exp (K_SIZE)  */
  ip_addr_t i, j, offset;
  printf ("Creating colunn index...");
  /*  clear col_index to be marked  */
  for (i = 0; i < exp_col; i ++) col_index[i] = 0;
  /*  mark first element where value will change  */
  for (i = 0; i < RLE_size; i ++) {
    offset = 0;
    if (RLE[i] != NULL) {
      for (j = 0; j < RLE[i]->vlen; j ++) {
	/*  last mark is out of memory limit and is not inserted  */
	if (RLE[i]->run_length[j] < exp_col) {
	  /*  Mark  */
	  offset += RLE[i]->run_length[j];
	  col_index[offset] = 1;
	}
      }
    }
  }
  /*  Create real colun index vector  */
  for (i = 1; i < exp_col; i ++) col_index[i] += col_index[i - 1];
  printf ("done\n");
  return;
}

/* for each i such that 0 < i < vlen sum(RLE->run_length[i]) = exp_val  */
char createrow (cell_t *row, ip_addr_t *metric, col_index_t size, cluster_t *RLE, col_index_t row_len) {
  ip_addr_t met_i, row_j;
  row_j = 0;
  /*  loop trough metric vector  */
  for (met_i = 0; met_i < size; met_i ++) {
    if (RLE->run_length[row_j] > metric[met_i]) {
      RLE->run_length[row_j] -= metric[met_i];
      row[met_i] = RLE->next_hop[row_j];
    }
    else {
      if (RLE->run_length[row_j] == metric[met_i]) {
	RLE->run_length[row_j] = 0;   /*  useless statement */
	row[met_i] = RLE->next_hop[row_j];
	row_j ++;
      }
      else {
	/* If this happen, congratulation! You found a bug in this software! */
	printf ("\nfail\nMetric error...can't go on\n");
	return -1;
      }
    }
  }
  return 0;
}

cell_t **create_ip_table (cluster_t **cluster, ip_addr_t row_len, ip_addr_t *sort, ip_addr_t *metric, ip_addr_t met_len, ip_addr_t tab_len) {
  ip_addr_t i, j;
  ip_addr_t exp_row = exp (K_SIZE);
  cell_t **tmp, **row_index;
  cell_t *whole;
  printf ("Creating IP table...");
  tmp = (cell_t **) malloc (row_len * sizeof (cell_t *));
  if (tmp == NULL) {
    printf ("fail\nMalloc failure creating support structure...can't go on\n");
    return NULL;
  }
  row_index = (cell_t **) malloc (exp_row * sizeof (cell_t *));
  if (row_index == NULL) {
    printf ("fail\nMalloc failure creating row index...can't go on\n");
    return NULL;
  }
  whole = (cell_t *) malloc (met_len * tab_len * sizeof (cell_t));
  if (whole == NULL) {
    printf ("fail\nMalloc failure creating whole structure...can't go on\n");
    return NULL;
  }
  j = 0;
  
  for (i = 0; i < row_len; i ++) {
    if (cluster[i] != NULL) {
      tmp[i] = whole + (j * met_len * sizeof (cell_t));
      if (createrow (tmp[i], metric, met_len, cluster[i], 0)) return NULL;
      if (tmp[i] == NULL) return NULL;
      free (cluster[i]->run_length);
      free (cluster[i]->next_hop);
      free (cluster[i]);
      j ++;
    }
  }
  /*  Creating row index maintaining data order  */
  for (i = 0; i < exp_row; i ++) {
    row_index[i] = tmp[sort[i]];
  }
  /*  Free temporary structures  */
  free (tmp);
  free (sort);
  free (cluster);
  printf ("done\n");
  return row_index;
}

ip_addr_t *create_metrics (col_index_t *col_index) {
  ip_addr_t exp_col = exp (K_SIZE);
  ip_addr_t *metric;
  ip_addr_t i, j;
  col_index_t len;
  printf ("Creating metric vector...");
  i = exp_col >> 1;
  /*  intervals from 0 to n - 1 are n */  
  len = col_index[exp_col - 1] + 1;  
  /*  Compute exact len for splitted  lable  */
  if (!col_index[i] && col_index[i - 1]) 
    len += col_index[i - 1] + 1;
  metric = (ip_addr_t *) malloc (len * sizeof (ip_addr_t));
  if (metric == NULL) {
    printf ("fail\nMalloc fail allocating metric vector...can't go on\n");
    return NULL;
  }
  j = 0;
  len = 1;   /*  include myself  */
  for (i = 1; i < exp_col; i ++) {
    if (col_index[i] != col_index[i - 1]) {
      metric[j] = len;
      j ++;
      len = 1;    /*  include myself  */
    }
    else len ++;
  }
  metric[j] = len; /*  Insert last element  */
  printf ("done\n");
  return metric;
}
