#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/time.h>
#include <unistd.h>

typedef unsigned char cell_t;     /*  next hop  */
typedef unsigned char ip_t;       /*  IP element from 0 to 255  */
typedef unsigned int ip_addr_t;   /*  IPv4 entire address  */
typedef unsigned short int col_index_t; /*  A type for colunn index  */
typedef unsigned char col_short_t;  /*  Short data value for colun index  */
typedef short int bucket_t;       /*  for bucket colun index  */

typedef unsigned short int usi_t; /*  Short unsigned int  */

#if FINE_TIME
typedef long long int my_time_t;
#endif

#define MAX_IP_LUNG_AS_STRING 16  /*  strlen ("XXX.XXX.XXX.XXX");  */
#define MAX_RIGA_READ 256
#define MAX_NEXT_HOPS 256   /*  Maximum number of next hops in table  */
#define MAX_SHORT_COL 256   /*  Maxinum value storable in col_short_t   */
#define K_SIZE 16
#define LAST_L_VALUE 3  /*  illegal value for the last S triple  */
#define	OVER_BUFFER 1   /*  Usually this is a good value!  */
#define FOLLOW_UP MAX_NEXT_HOPS - 1  /*  match everything */

#define first_k(x) (x >> (32 - K_SIZE))
#define exp(x) (1 << (x))
#define bit_k(x,k) ((x & (1 << (k))) >> (k))

#if FINE_TIME
#define rdtscll(val) \
     __asm__ __volatile__("rdtsc" : "=A" (val))
#endif
 
struct net_table {
  ip_addr_t ip;  
  ip_t netmask;
  cell_t next_hop;
  struct net_table *next;
};

typedef struct {
  cell_t next_hop;       /*  Next hop  */
  ip_addr_t run_length;  /*  length */
} rle_t;

typedef struct {
  ip_addr_t num;         /*  cluster number (rebuild right order after sort) */
  ip_addr_t vlen;
  cell_t *next_hop;
  ip_addr_t *run_length;
} cluster_t;

/*  The size of IP halves is unknown because depend from K_SIZE costant  */
typedef struct {
  usi_t first_pair;      /*   first part of ip address  */
  usi_t last_pair;       /*  second part of ip address  */
  cell_t next_hop;
  void *next;
} addr_list_t;

/*  this type havn't the pointer next. That need for optimize cache use in
    lookup operation  */
typedef struct {
  usi_t first_pair;      /*   first part of ip address  */
  usi_t last_pair;       /*  second part of ip address  */
  cell_t next_hop;
} addr_table_t;


typedef struct {
  cell_t next_hop;
  ip_t mark; 
  void *child[2];
} trie_t;

/*  createRLE.c  */
cluster_t **create_cluster (rle_t **rle_seq, ip_addr_t *size, ip_addr_t *sort, ip_addr_t *tab_size);
void applay_changes (ip_addr_t *sort, usi_t *subs);
char is_equal (cluster_t *el1, cluster_t *el2);
int comp_clust2 (const void *elem1, const void *elem2);
int comp_clust3 (const void *elem1, const void *elem2);
ip_addr_t max_RLE_cluster_length (rle_t **rle_seq, ip_addr_t size);

/*  createipt.c  */
void create_colun_index (col_index_t *col_index, cluster_t **RLE, ip_addr_t RLE_size, ip_addr_t exp_col);
char createrow (cell_t *row, ip_addr_t *metric, col_index_t size, cluster_t *RLE, col_index_t row_len);
cell_t **create_ip_table (cluster_t **cluster, ip_addr_t row_len, ip_addr_t *sort, ip_addr_t *metric, ip_addr_t met_len, ip_addr_t tab_len);
ip_addr_t *create_metrics (col_index_t *col_index);

/*  loadroute.c  */
int ins_in_next_hop_array (char **next_hops, char *next_hop);
ip_t used_bits (ip_t elem);
ip_t calcola_netmask (ip_t ip1, ip_t ip2, ip_t ip3, ip_t ip4);
ip_addr_t correct_nerwork (ip_addr_t ip, ip_t netmask);
void unload_next_hops (char **next_hops);
struct net_table *load_table (char *nome, char **next_hops,ip_addr_t *rou_c, char flag);

/*  supportfn.c  */
#if FINE_TIME
double detect_cpu_clock (void);
#endif
long int Mark_Time (void);
char eval_params (int argc, char *argv[], ip_addr_t *ntimes);
void echo_use (void);
void help (void);
void write_log_file (ip_addr_t *tot_rows, col_index_t *row_len, ip_addr_t b_size, ip_addr_t b_tmp, char flag, char comp_col, long tempo);
#if FINE_TIME
void write_log_lookup (char comp_col, long *look_time, double cpu_speed);
#else
void write_log_lookup (char comp_col, long *look_time);
#endif
void write_results (addr_table_t **array, ip_addr_t len, char **next_hop);
long no_lookup_time (addr_table_t **array, ip_addr_t len, ip_addr_t ntimes);

/*  t_backend.c  */
cell_t **rebuild_table (rle_t **RLE_seq, char flag, col_index_t *col_index, ip_addr_t tot_route, ip_addr_t *tot_rows, col_index_t *row_len);
void *build_col_index (col_index_t *col_index, char flag, col_index_t *row_len, ip_addr_t *b_size, ip_addr_t *b_tmp, char *comp_col);

/*  cdglookup.c  */
int main (int argc, char *argv[]);

/*  mk_lookup.c  */
long make_lookups (cell_t **row, col_index_t *col, addr_table_t **array, ip_addr_t len, ip_addr_t ntimes);
long make_lookups_sc (cell_t **row, col_short_t *col, addr_table_t **array, ip_addr_t len, ip_addr_t ntimes);
long make_bucket_lookups (cell_t **row, bucket_t *col, addr_table_t **array, ip_addr_t len, ip_addr_t ntimes, ip_addr_t size);

/*  load_addr.c  */
addr_table_t **load_address (char *nome, ip_addr_t *len, char flag, char **nh);
void unload_address (addr_table_t **array, ip_addr_t len);
int load_data (addr_list_t *self, char *riga);

/*   col_index.c  */
col_short_t *compress_col_index_type (col_index_t *col_index);
ip_addr_t compute_bucket_size (col_index_t *col_index);
ip_addr_t tot_bucket_fault (col_index_t *col_index, ip_addr_t bucket);
bucket_t *create_bucket_col_index (col_index_t *col_index);

/*  cdgupdate.c  */
trie_t *create_trie (struct net_table *in_table, ip_addr_t *tot_nodes);
trie_t *create_root (void);
int trie_insert_network (struct net_table *elem, trie_t *root);
rle_t **build_RLE_sequence (rle_t **RLE_seq, ip_addr_t *RLE_len);
rle_t **build_empty_RLE_sequence (ip_addr_t len);
int trie_remove_network (ip_addr_t ip, ip_t netmask, trie_t *root);
ip_addr_t merge_leaves (trie_t *root, ip_addr_t delete);
void cover_holes (trie_t *root);
rle_t **update_RLE_sequence (trie_t *root, ip_addr_t *RLE_len, char flag);
void visit_b_tree_sup (trie_t *root, rle_t **RLE, ip_addr_t *RLE_ind, ip_addr_t ip, ip_t netmask);
void visit_b_tree_inf (trie_t *root, rle_t **RLE, ip_addr_t *RLE_ind, ip_addr_t ip, ip_t netmask);
void visit_b_tree (trie_t *root, rle_t **RLE, ip_addr_t *RLE_ind, ip_addr_t ip, ip_t netmask);
