#include "iplookup.h"

long make_lookups (cell_t **row, col_index_t *col, addr_table_t **array, ip_addr_t len, ip_addr_t ntimes) {
  register ip_addr_t i,j;
  register cell_t **rw = row + exp (K_SIZE);
#if FINE_TIME
  my_time_t start_time ,stop_time;
  rdtscll (start_time);
#else
  Mark_Time ();
#endif
  /*  repeat the same operation n times  */
  for (i = 0; i < ntimes; i ++) {
    /*  Each loop is a lookup  */
    for (j = 0; j < len; j ++) {
      /*  It's all for this !!!  */
      array[j]->next_hop = row[array[j]->first_pair][array[j]->last_pair >> 8];
      if (array[j]->next_hop == FOLLOW_UP) {
	array[j]->next_hop=rw[array[j]->first_pair][col[array[j]->last_pair]];
      }
    }
  }
#if FINE_TIME
  rdtscll (stop_time);
  return stop_time - start_time;
#else
  return Mark_Time ();
#endif
}

/*  like above make_lookups () but col_index is declared as short type  */
long make_lookups_sc (cell_t **row, col_short_t *col, addr_table_t **array, ip_addr_t len, ip_addr_t ntimes) {
  register ip_addr_t i,j;
  register cell_t **rw = row + exp (K_SIZE);
#if FINE_TIME
  my_time_t start_time ,stop_time;
  rdtscll (start_time);
#else
  Mark_Time ();
#endif
  /*  repeat the same operation n times  */
  for (i = 0; i < ntimes; i ++) {
    /*  Each loop is a lookup  */
    for (j = 0; j < len; j ++) {
      /*  It's all for this !!!  */
      array[j]->next_hop = row[array[j]->first_pair][array[j]->last_pair >> 8];
      if (array[j]->next_hop == FOLLOW_UP) {
        array[j]->next_hop=rw[array[j]->first_pair][col[array[j]->last_pair]];
      }
    }
  }
#if FINE_TIME
  rdtscll (stop_time);
  return stop_time - start_time;
#else
  return Mark_Time ();
#endif
}

/*  Bucket lookup procedure  */
long make_bucket_lookups (cell_t **row, bucket_t *col, addr_table_t **array, ip_addr_t len, ip_addr_t ntimes, ip_addr_t size) {
  register ip_addr_t i,j;
  register bucket_t offset;
  register unsigned char shift;
  register col_index_t s_size = size;
  register cell_t **rows = row + exp (K_SIZE);
#if FINE_TIME
  my_time_t start_time ,stop_time;
#endif
  offset = (exp (K_SIZE) / size) - size;
  shift = -1;
  while (size != 0) {
    shift ++;
    size >>= 1;
  }
#if FINE_TIME
  rdtscll (start_time);
#else
  Mark_Time ();
#endif
  /*  repeat the same operation n times  */
  for (i = 0; i < ntimes; i ++) {
    /*  Each loop is a lookup  */
    for (j = 0; j < len; j ++) {
      /*  It's all for this !!!  */
      array[j]->next_hop = row[array[j]->first_pair][array[j]->last_pair >> 8];
      if (array[j]->next_hop == FOLLOW_UP) {
	register bucket_t last = col[array[j]->last_pair >> shift];
	if (last >= 0) 
	  array[j]->next_hop = rows[array[j]->first_pair][last];
	else 
	  array[j]->next_hop = rows[array[j]->first_pair][col[offset +((-last) << shift) + (array[j]->last_pair % s_size)]];
      }
    }
  }
#if FINE_TIME
  rdtscll (stop_time);
  return stop_time - start_time;
#else
  return Mark_Time ();
#endif
}

