#include <stdlib.h>
#include <stdio.h>
#include "vars.h"

int strcasecmp(const char *, const char *);   /* prototype */

int exp2(int exp) {
  /* 
   * This function return 2^exp with exp in the range [0,16]
   * else display on stderr an error and continue returning zero
   */
  switch(exp) {
  case 0:
    return(1);
  case 1:
    return(2);
  case 2:
    return(4);
  case 3:
    return(8);
  case 4:
    return(16);
  case 5:
    return(32);
  case 6:
    return(64);
  case 7:
    return(128);
  case 8:
    return(256);
  case 9:
    return(512);
  case 10:
    return(1024);
  case 11:
    return(2048);
  case 12:
    return(4096);
  case 13:
    return(8192);
  case 14:
    return(16384);
  case 15:
    return(32768);
  case 16:
    return(65536);
  default:
    fprintf(stderr,"Syntax Error in using exp2(%d)\n",exp);
    return(0);
  }
}

void StartTime()
{
  /*
   * Load in variable TimeStart the actual time
   */
  gettimeofday(&TimeStart,NULL);
}

long EndTime()
{
  /*
   * Using previous value loaded in TimeStart, return the number of
   * microsecond (1/1000000 of a second) elapsed from the last call
   * of StartTime
   */
  gettimeofday(&TimeEnd,NULL);
  return((TimeEnd.tv_sec-TimeStart.tv_sec)*1000000+TimeEnd.tv_usec-TimeStart.tv_usec);
}

void init(int argc,char *argv[])
{
  /* 
   * Initialize variable
   */
  nrouter=1;
  nroute=0;
  naddress=0;
  strcpy(routertable[EMPTY].iprouter,"NO ROUTE TO HOST");

  if (argc<6) {
    action=A_NOTHING;
  } else {
    if (strcasecmp(argv[1],"SEARCH")==0) {
      printf("Executing Search...\n\n");
      action=A_SEARCH;
    }
    else if (strcasecmp(argv[1],"INSERT")==0) {
      printf("Executing Insert...\n\n");
      action=A_INSERT;
    }
    else if (strcasecmp(argv[1],"DELETE")==0) {
      printf("Executing Delete...\n\n");
      action=A_DELETE;
    }
    else
      action=A_NOTHING;
  }
  if (action!=A_NOTHING) {
    fileroutename=argv[2];
    fileaddressname=argv[3];
    filevarroutename=argv[6];
    howmany=atoi(argv[4]);
    howtimes=atoi(argv[5]);
  }
}

void openfile()
{
  /*
   * Open all the files needed for the program
   */
  if ((fileroute=fopen(fileroutename,"r"))==NULL) {
    perror("Error during opening route file");
    exit(-1);
  }
  if ((fileaddress=fopen(fileaddressname,"r"))==NULL) {
    perror("Error during opening address file");
    exit(-1);
  }
  if ((fileoutput=fopen("output.txt","w"))==NULL) {
    perror("Error during creating output file");
    exit(-1);
  }
  if (action!=A_SEARCH) {
    if ((filevarroute=fopen(filevarroutename,"r"))==NULL) {
      perror("Error during opening update file");
      exit(-1);
    }
  }
}

CELL searchrouter(char *iprouter)
{
  /*
   * Perform a sequential search in the table of router returning
   * the number of the router searched and adding it if it is not
   * present
   */
  CELL i;

  for (i=0;i<nrouter;i++)
    if (strcmp(iprouter,routertable[i].iprouter)==0) {
      return(i);
    }
  strcpy(routertable[nrouter].iprouter,iprouter);
  nrouter++;
  return(i);
}

int readoneaddress(t_address *address) 
{
  /* Read one address from the address file and store it in variable
   * pointed by "address", return 0 if successfull, 1 on error
   */
  char lineread[MAXLINEREAD]; 
  int byte_a,byte_b,byte_c,byte_d;

  if (fgets(lineread,MAXLINEREAD-1,fileaddress)==NULL)
    return(-1);
  else {
    sscanf(lineread,"%d.%d.%d.%d",&byte_a,&byte_b,&byte_c,&byte_d);
    address->firstpair=byte_a*256+byte_b;
    address->lastpair=byte_c*256+byte_d;
    return(0);
  }
}

int readoneroute(t_route *route,FILE *file) 
{
  /* Read one route from the file specified and store it in variable
   * pointed by "route", return 0 if successfull, 1 on route error,
   * -1 on file read error
   */ 
  char lineread[MAXLINEREAD]; 
  int byte_a,byte_b,byte_c,byte_d;
  int netbit,len;
  CELL nexthop;
  char routerread[LENIPROUTER];

  if (fgets(lineread,MAXLINEREAD,file)==NULL)
    return(-1);
  else {
    sscanf(lineread,"%d.%d.%d.%d/%d %s",&byte_a,&byte_b,&byte_c,&byte_d,&netbit,routerread);
    nexthop=searchrouter(routerread);
    route->firstpair=byte_a*256+byte_b;
    route->lastpair=byte_c*256+byte_d;
    route->netbit=netbit;
    route->nexthop=nexthop;
    len=32-netbit;
    if (len>16)
      len=16;
    if (route->lastpair+exp2(len)>65536) {
      fprintf(stderr,"Unconsistent route %d.%d.%d.%d/%d %s in your table: corrected\n",byte_a,byte_b,byte_c,byte_d,netbit,routerread);
      return(1);
    }
    else
      return(0);
  }
}

void loadaddress()
{
  /* 
   * Load address in array
   */
  t_address oneaddress;

  printf("Loading address...\n");
  while (readoneaddress(&oneaddress)==0) {
    address[naddress++]=oneaddress;
  }
  printf("  %d address loaded\n\n",naddress); 
}

int pstrcmp(const void *i, const void *j)
{
  /* 
   * Compare function for sorting using standard qsort
   * ordering is done by 16bit prefix (firstpair) and netmask
   * (netbit) from shorter to longer
   */
  if (((t_route *)i)->firstpair<((t_route *)j)->firstpair)
    return(-1);
  else if (((t_route *)i)->firstpair>((t_route *)j)->firstpair)
    return(1);
  else if (((t_route *)i)->netbit<((t_route *)j)->netbit)
    return(-1);
  else if (((t_route *)i)->netbit>((t_route *)j)->netbit)
    return(1);
  else
    return(0);
}

void loadroute()
{
  /*
   * Load route in array
   */
  t_route route;
  int result;

  printf("Loading routing table...\n");
  do {
    result=readoneroute(&route,fileroute);
    if (result==0)
      routetable[nroute++]=route;
  } while (result!=-1);
  routetable[nroute].firstpair=0;
  printf("  %d route loaded with %d different router\n\n",nroute,nrouter);
}

void loadvarroute()
{
  /*
   * Load update to the route table in array
   */
  t_route varroute;
  int result;

  printf("Loading update...\n");
  do {
    result=readoneroute(&varroute,filevarroute);
    if (result==0)
      varroutetable[varnroute++]=varroute;
  } while (result!=-1);
  varroutetable[varnroute].firstpair=0;
  printf("  %d update loaded for %d different router\n",varnroute,nrouter);
}

unsigned int exploderoute(t_route table[],unsigned int ntable)
{
  /*
   * Perform route expansion to 16 bit and return the number of
   * all new route
   */
  unsigned int newnroute;
  unsigned int i,j;
  unsigned int from,to;

  /* printf("Route expansion to 16 bit...\n"); */
  newnroute=ntable;
  for (i=0;i<ntable;i++) {
    if (table[i].netbit<16) {
      from=table[i].firstpair;
      to=table[i].firstpair+exp2(16-table[i].netbit);
      table[i].lastpair=0;
      for (j=from+1;j<to;j++) {
	table[newnroute]=table[i];
	table[newnroute++].firstpair=j;
      }
    }
  }
  table[newnroute].firstpair=0;
  /* printf("  %d route obtained\n",newnroute); */
  return(newnroute);
}

int intcmp(const void *a, const void *b)
{
  /* 
   * Compare function for sorting using standard qsort
   */
  if (*(unsigned int *)a<*(unsigned int *)b)
    return(-1);
  else if (*(unsigned int *)a>*(unsigned int *)b)
    return(1);
  else
    return(0);
}

unsigned int insertbreakpoint(unsigned int from, unsigned int to, CELL nexthop, t_routeline *routeline,unsigned int nrline)
{
  /*
   * Insert a breakpoint in breakpoint vector if it is not present
   */
  unsigned int inserted;
  unsigned int j,k;
  t_routeline newrouteline;
  unsigned int newnrline;

  inserted=0;
  k=0;
  for (j=0;j<nrline;j++) {
    if (routeline->address[j]<from) {
      newrouteline.address[k]=routeline->address[j];
      newrouteline.dest[k]=routeline->dest[j];
      k++;
    } else if (routeline->address[j]==from) {
      newrouteline.address[k]=from;
      newrouteline.dest[k]=nexthop;
      k++;
      inserted=1;
      if (j+1<nrline) {
	if (routeline->address[j+1]>to) {
	  newrouteline.address[k]=to;
	  newrouteline.dest[k]=routeline->dest[j];
	  k++;
	  inserted=2;
	} else {
	  newrouteline.address[k]=to;
	  newrouteline.dest[k]=routeline->dest[j+1];
	  k++;
	  inserted=2;
	  j++;
	}
      } else {
	newrouteline.address[k]=to;
	newrouteline.dest[k]=0;
	k++;
	inserted=2;
      }
    } else if (inserted==0) {
      newrouteline.address[k]=from;
      newrouteline.dest[k]=nexthop;
      k++;
      inserted=1;
      if (routeline->address[j]>to) {
	newrouteline.address[k]=to;
	if (j==0)
	  newrouteline.dest[k]=0;
	else
	  newrouteline.dest[k]=routeline->dest[j-1];
	k++;
	inserted=2;
	newrouteline.address[k]=routeline->address[j];
	newrouteline.dest[k]=routeline->dest[j];
	k++;
      } else if (routeline->address[j]==to) {
	newrouteline.address[k]=to;
	newrouteline.dest[k]=routeline->dest[j];
	k++;
	inserted=2;
      }
    } else {
      newrouteline.address[k]=routeline->address[j];
      newrouteline.dest[k]=routeline->dest[j];
      k++;
    }
  }
  if (inserted==0) {
    newrouteline.address[k]=from;
    newrouteline.dest[k]=nexthop;
    k++;
    inserted=1;
  }
  if (inserted==1) {
    newrouteline.address[k]=to;
    newrouteline.dest[k]=0;
    k++;
  }
  newnrline=k;
  for (j=0;j<newnrline;j++) {
    routeline->address[j]=newrouteline.address[j];
    routeline->dest[j]=newrouteline.dest[j];
  }
  return(newnrline);
}

unsigned int compactbreakpoint(t_routeline *routeline, unsigned int nrline)
{
  /*
   * Remove from breakpoint vector not necessary breakpoint
   */
  t_routeline newrouteline;
  unsigned int newnrline;
  unsigned int j,k;

  newrouteline.address[0]=routeline->address[0];
  newrouteline.dest[0]=routeline->dest[0];
  k=1;
  for (j=1;j<nrline;j++) {
    if (routeline->dest[j]!=routeline->dest[j-1]) {
      newrouteline.address[k]=routeline->address[j];
      newrouteline.dest[k]=routeline->dest[j];
      k++;
    }
  }
  newnrline=k;
  for (j=0;j<newnrline;j++) {
    routeline->address[j]=newrouteline.address[j];
    routeline->dest[j]=newrouteline.dest[j];
  }
  return(newnrline);
}

unsigned int updatecut(t_routeline routeline,unsigned int nrline)
{
  /*
   * Update cut vector with breakpoint from breakpoint vector
   */
  unsigned int newncut;
  unsigned int j,k;

  newncut=ncut;
  for (j=0;j<nrline;j++) {
    k=0;
    while ((k<newncut) && (cut[k]!=routeline.address[j]))
      k++;
    if (k==newncut)
      cut[newncut++]=routeline.address[j];
  }
  return(newncut);
}


void buildcolindextable()
{
  /* 
   * Create the lastpair table
   */
  unsigned int from,to;
  unsigned int nrline;
  char len;
  unsigned int i,j;
  unsigned int max=0;
  unsigned char nexthop;
  unsigned short firstpair;
  t_routeline routeline;

  /* printf("Building lastpair table...\n"); */
  printf("  Building breakpoint vector\n");
  ncut=0;
  i=0;
  do {
    firstpair=routetable[i].firstpair;
    nrline=0;
    while (firstpair==routetable[i].firstpair) {
      len=32-routetable[i].netbit;
      if (len>16)
	len=16;
      from=routetable[i].lastpair;
      to=routetable[i].lastpair+exp2(len);
      nexthop=routetable[i].nexthop;
      nrline=insertbreakpoint(from,to,nexthop,&routeline,nrline);
      i++;
    }
    nrline=compactbreakpoint(&routeline,nrline);
    ncut=updatecut(routeline,nrline);
    if (max<nrline)
      max=nrline;
  } while (i<nroute);
  /* printf("  %d breakpoint found\n",ncut); */
  printf("  %d max breakpoint found for one 16 bit route\n",max);
  printf("  Sorting breakpoint\n");
  qsort((void *)cut,(size_t)ncut,(size_t)sizeof(int),intcmp);
  printf("  Storing value in table\n");
  for (i=0;i<ncut-1;i++)
    for (j=cut[i];j<cut[i+1];j++)
      col_index[j]=i;
  nitem=ncut-1;
}

int compare(CELL item1[MAXCUT],CELL item2[MAXCUT],int ncell)
{
  /*
   * Compare two vector of ncell elements, element by element, return
   * 0 if equals, 1 if item1>item2, -1 if item1<item2 
   */
  int i;

  for (i=0;i<ncell;i++)
    if (item1[i]<item2[i])
      return(-1);
    else if (item1[i]>item2[i])
      return(1);
  return(0);
}

CELL *searchitem(CELL item[],unsigned short firstpair) 
{
  /*
   * Perform a search in a tree, inserting the element if it is not present
   * and return a pointer to the element
   */
  t_tree **node;
  CELL *itemvalue;
  unsigned int okvalue;
  unsigned int j;

  okvalue=0;
  node=&root;
  while ((*node!=NULL) && (okvalue==0)) {
    switch (compare(item,(*node)->number,nitem)) {
    case 0:
      okvalue=1;
      itemvalue=(*node)->number;
      break;
    case -1:
      node=&(*node)->left;
      break;
    case 1:
      node=&(*node)->right;
      break;
    }
  }
  if (*node==NULL) {
    itemvalue=(CELL*)malloc(nitem*sizeof(CELL));
    maxitemvalue++;
    *node=malloc(sizeof(t_tree));
    (*node)->number=itemvalue;
    for (j=0;j<nitem;j++)
      *((*node)->number+j)=item[j];
    (*node)->left=NULL;
    (*node)->right=NULL;
  }
  if (action!=A_SEARCH)
    treetable[firstpair]=*node;
  return(itemvalue);
}

void buildrowindextable()
{
  /*
   * Create the row_index table
   */
  CELL item[MAXCUT];
  unsigned int from;
  unsigned int to;
  char len;
  unsigned int i,j;
  unsigned short oldfirstpair;
  unsigned int different=0;

  root=NULL;
  printf("Building row index table...\n");
  printf("  Building the table\n");
  maxitemvalue=0;
  for (j=0;j<nitem;j++)
    item[j]=EMPTY;
  row_index[0]=searchitem(item,0);
  for (i=1;i<65536;i++) {
    row_index[i]=row_index[0];
    if (action!=A_SEARCH)
      treetable[i]=treetable[0];
  }
  i=0;
  do {
    different++;
    oldfirstpair=routetable[i].firstpair;
    for (j=0;j<nitem;j++)
      item[j]=EMPTY;
    do {
      len=32-routetable[i].netbit;
      if (len>16)
	len=16;
      from=routetable[i].lastpair;
      to=from+exp2(len);
      for (j=0;j<nitem;j++)
	if ((cut[j]>=from) && (cut[j+1]<=to))
	    item[j]=routetable[i].nexthop;
      i++;
    } while (routetable[i].firstpair==oldfirstpair);
    row_index[oldfirstpair]=searchitem(item,oldfirstpair);
  } while (i<nroute);
  printf("  %d different 16 bit route\n",different);
  printf("  Number of element to allocate %d x %d\n",maxitemvalue,nitem);
}

void createrow_indexlisttable()
{
  /*
   * Create the row_indexlist table
   */
  unsigned int i;
  unsigned short oldfirstpair;
  t_list *node;

  printf("Creating the row_indexlist table\n"); 
  oldfirstpair=routetable[0].firstpair-1;
  for (i=0;i<=nroute;i++) {
    if (routetable[i].firstpair==oldfirstpair) {
      node->next=(t_list*)malloc(sizeof(t_list));
      node=node->next;
    } else {
      oldfirstpair=routetable[i].firstpair;
      node=(t_list*)malloc(sizeof(t_list));
      row_indexlist[oldfirstpair]=node;
    }
    node->route=routetable[i];
    node->next=NULL;
  }
}

void expandtree(t_tree *node,unsigned int value[])
{
  /*
   * Modify tree structure to allow newly inserted breakpoint
   */
  unsigned int i;
  CELL *newnumber;

  if (node!=NULL) {
    newnumber=(CELL*)malloc(nitem*sizeof(CELL));
    if (newnumber==NULL) {
      fprintf(stderr,"Memory allocation error\n");
      exit(-1);
    }
    for (i=0;i<nitem;i++)
      *(newnumber+i)=*(node->number+value[i]);
    node->number=newnumber;
    expandtree(node->left,value);
    expandtree(node->right,value);
  }
}

void additem()
{
  /*
   * Perform expansion of the structure and update 
   * lastpair allowing a new breakpoint
   * in lastpair table
   */

  unsigned int i,j;
  int value[MAXCUT];

  nitem=ncut-1;  
  printf("Expansion to %d breakpoints\n",nitem);
  for (i=0;i<nitem;i++)
    value[i]=col_index[cut[i]];
  for (i=0;i<nitem;i++)
    for (j=cut[i];j<cut[i+1];j++)
      col_index[j]=i;
  expandtree(root,value);
}

void updaterow_index()
{
  /*
   * Reassign correct value to row_index table after an update
   */
  unsigned int i;

  for (i=0;i<65536;i++)
    row_index[i]=treetable[i]->number;
}

void updateroute(unsigned short firstvalue) 
{
  /*
   * Using values in firsttable compute breakpoint for a
   * 16 bit prefix
   */
  t_list *nodo;
  unsigned int from,to;
  /* unsigned int newcut[MAXCUT]; */
  unsigned int nthis;
  unsigned int nnewcut;
  char len;
  unsigned int i,j,k;
  CELL item[MAXCUT];
  unsigned int nrline;
  t_routeline routeline;
  CELL nexthop;

  nthis=0;
  nnewcut=0;
  nrline=0;
  nodo=row_indexlist[firstvalue];
  if (nodo==NULL) {
    for (j=0;j<nitem;j++)
      item[j]=EMPTY;    
  }
  else
    {
      while (nodo!=NULL) {
	len=32-nodo->route.netbit;
	if (len>16)
	  len=16;
	from=nodo->route.lastpair;
	to=nodo->route.lastpair+exp2(len);
	nexthop=nodo->route.nexthop;
	nrline=insertbreakpoint(from,to,nexthop,&routeline,nrline);
	nodo=nodo->next;
      }
      nrline=compactbreakpoint(&routeline,nrline);
      for (j=0;j<nrline;j++) {
	k=0;
	while ((k<ncut) && (cut[k]!=routeline.address[j]))
	  k++;
	if (k==ncut) {
	  cut[ncut++]=routeline.address[j];
	  nnewcut++;
	}
      }
      if (nnewcut!=0) {
	qsort((void *)cut,(size_t)ncut,(size_t)sizeof(int),intcmp);    
	additem();
      }
      for (j=0;j<nitem;j++)
	item[j]=EMPTY;
      for (i=0;i<nrline-1;i++)
	for (j=routeline.address[i];j<routeline.address[i+1];j++)
          item[col_index[j]]=routeline.dest[i];
    }
  row_index[firstvalue]=searchitem(item,firstvalue);  
}

void insertroute()
{
  /*
   * Insert a new route in structure
   */
  unsigned int i;
  t_list *newnodo,**nodo;
  unsigned int oldncut;

  oldncut=ncut;
  printf("Inserting new route\n");
  for (i=0;i<varnroute;i++) {
    newnodo=(t_list*)malloc(sizeof(t_list));
    newnodo->route=varroutetable[i];
    nodo=&row_indexlist[varroutetable[i].firstpair];
    while ((*nodo!=NULL) && (*nodo)->route.netbit<newnodo->route.netbit) {
      nodo=&(*nodo)->next;
    }
    newnodo->next=*nodo;
    *nodo=newnodo;
    if ((i==varnroute-1)||(varroutetable[i].firstpair!=varroutetable[i+1].firstpair))
      updateroute(varroutetable[i].firstpair);
  }
  if (ncut!=oldncut)
    updaterow_index();
}

void deleteroute()
{
  /*
   * Delete a route from structure
   */
  unsigned int i;
  t_list **nodo;

  printf("Deleting route...\n");
  for (i=0;i<varnroute;i++) {
    nodo=&row_indexlist[varroutetable[i].firstpair];
    while ((*nodo!=NULL) && ( (*nodo)->route.netbit!=varroutetable[i].netbit 
	  || (*nodo)->route.firstpair!=varroutetable[i].firstpair 
	  || (*nodo)->route.lastpair!=varroutetable[i].lastpair
	  || (*nodo)->route.nexthop!=varroutetable[i].nexthop ))
      {
	nodo=&(*nodo)->next;
      }
    if (*nodo!=NULL) {
      *nodo=(*nodo)->next;
      if ((i==varnroute-1)||(varroutetable[i].firstpair!=varroutetable[i+1].firstpair))
	updateroute(varroutetable[i].firstpair);
    }
    else
      printf("Route not found\n"); 
  }
}

void tryrouting()
{
  /*
   * Execute search
   */
  unsigned int i,j;
  unsigned long end1, end2;
  double IPtime;

  printf("Executing %d search %d times...\n",howmany,howtimes);
  for (i=0;i<howmany;i++)
    result[i]=*(row_index[address[i].firstpair]+col_index[address[i].lastpair]);
  StartTime();
  for (j=0;j<howtimes;j++)
    for (i=0;i<howmany;i++)
      result[i]=i;
  end1=EndTime();
  printf("  Wallclock time to perform null search: %ld microsecond\n",end1);  
  StartTime();
  for (j=0;j<howtimes;j++)
    for (i=0;i<howmany;i++)
      /* fast IP lookup: that's it! */
      result[i]=*(row_index[address[i].firstpair]+col_index[address[i].lastpair]);
  end2=EndTime();
  printf("  Wallclock time to search: %ld microsecond\n",end2);  
  IPtime =  ((double)(end2 - end1)) * (1000  / (double)(howtimes * howmany)) ;
  printf("  Wallclock time for the single IP lookup operation: %lf nanoseconds (i.e., %ld lookups per second)\n", IPtime, (long) (1000000000/IPtime));  
 
 
}

int main(int argc,char* argv[]) 
{
  unsigned int i;

  init(argc,argv);
  if (action==A_NOTHING) {
    printf("Usage:\n");
    printf("iplookup search <route table> <IP address file>\n\t\t<number of searches> <repeat times per search>\n");
    printf("iplookup insert <route table> <IP address file>\n\t\t<number of searches> <repeat times per search>\n\t\t<new routes' update table>\n");
  } else { 
    openfile();
    loadroute();
    StartTime();
    nroute=exploderoute(routetable,nroute);
    printf("%ld microseconds\n\n",EndTime());
    printf("Sorting routing table...\n");
    StartTime();
    qsort((void *)routetable,(size_t)nroute,(size_t)sizeof(t_route),pstrcmp);
    printf("%ld microseconds\n\n",EndTime());
    StartTime();
    buildcolindextable();
    printf("%ld microseconds\n\n",EndTime());
    StartTime();
    buildrowindextable();
    printf("%ld microseconds\n\n",EndTime());
    if (action!=A_SEARCH) {
      loadvarroute();
      StartTime();
      varnroute=exploderoute(varroutetable,varnroute);
      printf("%ld microseconds\n\n",EndTime());
      printf("Sorting (insert/delete) routing table\n");
      StartTime();
      qsort((void *)varroutetable,(size_t)varnroute,(size_t)sizeof(t_route),pstrcmp);
      printf("%ld microseconds\n\n",EndTime());
      StartTime();
      createrow_indexlisttable();
      printf("%ld microseconds\n\n",EndTime());
      if (action==A_INSERT) {
	StartTime();
	insertroute();
	printf("%ld microseconds\n\n",EndTime());
      } else if (action==A_DELETE) {
	StartTime();
	deleteroute();
	printf("%ld microseconds\n",EndTime());
      }
    }    
    loadaddress();
    tryrouting();
    for (i=0;i<howmany;i++)
      fprintf(fileoutput,"%s\n",routertable[result[i]].iprouter);
  }
  return(0);
}

