#include "iplookup.h"

struct net_table *create_alttab (trie_t *trie, ip_addr_t *tot_alternative);
struct net_table *create_subset (trie_t *root, struct net_table *alttab, 
				 ip_addr_t ip, ip_t netmask);

int main (int argc, char *argv[]) {
  ip_addr_t memory = 0;
#ifdef USE_TRIE
  trie_t *root;
#else
  ip_addr_t tot_bas;
  basic_t *basic;
#endif
#ifdef USE_N_PARTITION
  N_part_t *npart;
#endif
#ifdef USE_K_PARTITION
  K_part_t *kpart;
#endif
  addr_table_t **addresses;
  trie_t *trie, *update;
  ip_addr_t tot_leaves = 0;
  char flag;  
  col_index_t row_len;
  ip_addr_t tot_rows;
  ip_addr_t RLE_len;  
  rle_t **RLE_seq;  
  cell_t **row_index; 
  ip_addr_t b_size = 0, b_tmp = 0;
  struct net_table *in_table, *subtable;
  ip_addr_t ntimes = 1;
  char *next_hops[MAX_NEXT_HOPS];
  long tempo;
  long look_time[2];
  ip_addr_t tot_route, tot_alternative;
#if FINE_TIME
  double cpu_speed;
  my_time_t start_time ,stop_time;
  double total_time;
  printf ("Computing CPU speed for time measurement...");
  fflush (stdout);
  cpu_speed = detect_cpu_clock ();
  printf ("done\n");
#endif
  flag = eval_params (argc, argv, &ntimes);
  /*  if ntimes is 0  there isn't an argoument for address file*/
  if (ntimes == 0) 
    in_table = load_table (argv[argc - 1], next_hops, &tot_route, flag & 1);
  else
    in_table = load_table (argv[argc - 2], next_hops, &tot_route, flag & 1);
  if (in_table == NULL) return -1;
  update = create_mini (in_table);
  /*  Layer 1  */
#if FINE_TIME
  rdtscll (start_time);
#else
  Mark_Time ();  
#endif
  /*  Create trie for update operation */
  trie = create_trie (in_table, &tot_leaves);
  /*  In_table was freed and now set pointer to NULL  */
  in_table = NULL;  
  RLE_len = tot_leaves;
  RLE_seq = update_RLE_sequence (trie, &RLE_len, 0); /*  sup  */
  if (RLE_seq == NULL) return -1;

  /*  create table back-end  */
  row_index = rebuild_table (RLE_seq, flag, 
			      RLE_len, &tot_rows, &row_len);
  if (row_index == NULL) return -1;
  /*  RLE_seq was freed and now set pointer to NULL  */
  RLE_seq = NULL;

#if FINE_TIME
  rdtscll (stop_time);
  tempo = stop_time - start_time;
  total_time = (1000000 / cpu_speed) * tempo;
  printf ("Elapsed time %f\n", total_time);
#else
  tempo = Mark_Time ();
  printf ("Elapsed time %ld\n", tempo);
#endif

  /*  Layer 2  */
  subtable = create_alttab (update, &tot_alternative);
  //  subtable = create_alttab (trie, &tot_alternative);
  if (subtable == NULL) return -1;
  printf ("%u entries in second level table\n", tot_alternative);
#ifdef USE_TRIE
  root = create_mini (subtable);
  if (!root) {
    printf ("Unable to create trie\n");
    return -1;
  }
  memory = trie_mem (root);
#else
  basic = create_basic_intervcals (subtable, &tot_bas);
  printf ("Total basic intervals %u\n",tot_bas);
  if (basic == NULL) {
    printf ("Error creating endpoints...\n");
    return -1;
  }
  /*  Basic interval memory occupation in bytes */
  memory = tot_bas * sizeof (basic_t);  
  /*  Free rule's memory  */
#endif
#ifdef USE_N_PARTITION
  printf ("Creating N partition..."); fflush (stdout);
  npart = create_N_partition (basic, tot_bas, &memory);
  if (npart == NULL) {
    printf ("fail\n");
    return -1;
  }
  printf ("done\n");
#endif
#ifdef USE_K_PARTITION
  printf ("Creating K partition..."); fflush (stdout);
  kpart = create_K_partition (basic, tot_bas, &memory);
  if (kpart == NULL) {
    printf ("fail\n");
    return -1;
  }
  printf ("done\n");
#endif
  printf ("Used memory %u bytes\n",memory);

  printf ("Done\n");
  /*  Write cdg.log file if requested */
  if (flag & 1) {
    write_log_file (tot_rows, b_size, b_tmp, flag, memory, tempo,
		    tot_alternative);
  }
  /*  -------------Lookup---------------*/

  /*  If iteration is unspecified was set to 1 from eval_params ()
      if iteration is 0 skip lookup  */
  if (ntimes != 0) {
    addresses = load_address (argv[argc - 1], &tot_route);
    if (addresses == NULL) return -1;
    look_time[0] = no_lookup_time (addresses, tot_route, ntimes);

#ifdef USE_N_PARTITION
    look_time[1] = make_lookups (row_index, addresses, tot_route, ntimes,
				 npart, tot_bas, basic);
#endif
#ifdef USE_K_PARTITION
    look_time[1] = make_lookups (row_index, addresses, tot_route, ntimes,
				 kpart, basic);
#endif
#ifdef USE_BIN_SEARCH
    look_time[1] = make_lookups (row_index, addresses, tot_route, ntimes,
				 tot_bas, basic);
#endif
#ifdef USE_TRIE
    look_time[1] = make_lookups (row_index, addresses, tot_route, ntimes, 
				 root);
#endif
    printf ("-- Lookup time --\n");
#if FINE_TIME
    total_time = (1000000 / cpu_speed) * look_time[1];
    printf ("Entire lookup procedure %f\n", total_time);
    total_time = (1000000 / cpu_speed) * look_time[0];
    printf ("No lookup operations %f\n", total_time);
    total_time = (1000000 / cpu_speed) * (look_time[1] -look_time[0]);
    printf ("Real lookup time %f\n", total_time);
#else
    printf ("Entire lookup procedure %ld\n",look_time[1]);
    printf ("No lookup operations %ld\n",look_time[0]);
    printf ("Real lookup time %ld\n",look_time[1] - look_time[0]);
#endif
    /*  Add lookup information to cdg.log file if requested */
    if (flag & 1) {
#if FINE_TIME
      write_log_lookup (look_time, cpu_speed);
#else
      write_log_lookup (look_time);
#endif
    }
    write_results (addresses, tot_route, next_hops);
    unload_address (addresses, tot_route);
    printf ("Done\n");
  }

  unload_next_hops (next_hops);
  return 0;
}

struct net_table *create_alttab (trie_t *trie, ip_addr_t *tot_alternative) {
  struct net_table *subtable, *tmp;
  ip_addr_t i;
  subtable = (struct net_table *) malloc (sizeof (struct net_table));
  if (subtable == NULL) {
    printf ("Error creating alternative table...can`t go on\n");
    return NULL;
  }
  if (create_subset (trie, subtable, 0, 0) == NULL) {
    printf ("Error creating alternative table...can`t go on\n");
    return NULL;
  }
  tmp = subtable;
  subtable = subtable->next;
  free (tmp);
  tmp = subtable;
  for (i = 0; tmp != NULL; i ++) tmp = tmp->next;
  *tot_alternative = i;
  return subtable;
}

struct net_table *create_subset (trie_t *root, struct net_table *alttab, 
				 ip_addr_t ip, ip_t netmask) {
  if (alttab == NULL) return NULL;
  if (netmask > 24 && root->mark == 1) {
    alttab->next = (struct net_table *) malloc (sizeof (struct net_table));
    if (alttab->next == NULL) return NULL;
    alttab = alttab->next;
    alttab->ip = (ip << (32 - netmask));
    alttab->netmask = netmask;
    alttab->next_hop = root->next_hop;
    alttab->next = NULL;
  }
  if (netmask == 24 && (root->child[0] != NULL || root->child[1] != NULL)) {
    alttab->next = (struct net_table *) malloc (sizeof (struct net_table));
    if (alttab->next == NULL) return NULL;
    alttab = alttab->next;
    alttab->ip = (ip << (32 - netmask));
    alttab->netmask = netmask;
    alttab->next_hop = root->next_hop;
    alttab->next = NULL;
  }
  if (root->child[0] != NULL)
    alttab = create_subset (root->child[0], alttab, (ip << 1), netmask + 1);
  if (alttab == NULL) return NULL;
  if (root->child[1] != NULL)
    alttab = create_subset (root->child[1], alttab, (ip << 1)+1, netmask+1);
  return alttab;
}



struct net_table *create_subset2 (trie_t *root, struct net_table *alttab, 
				 ip_addr_t ip, ip_t netmask) {
  if (alttab == NULL) return NULL;
  /*  if is a leaf  */
  if (root->child[0] == NULL && root->child[1] == NULL && netmask > 24) {
    ip = (ip << (32 - netmask));
    alttab->next = (struct net_table *) malloc (sizeof (struct net_table));
    if (alttab->next == NULL) return NULL;
    alttab = alttab->next;
    alttab->ip = ip;
    alttab->netmask = netmask;
    alttab->next_hop = root->next_hop;
    alttab->next = NULL;
    return alttab;
  }
  /*  Only for initial case  */  
  if (root->child[0] == NULL && root->child[1] == NULL && ip == 0) {
    ip = (ip << (32 - netmask));
    alttab->next = (struct net_table *) malloc (sizeof (struct net_table));
    if (alttab->next == NULL) return NULL;
    alttab = alttab->next;
    alttab->ip = ip;
    alttab->netmask = netmask;
    alttab->next_hop = root->next_hop;
    alttab->next = NULL;
    return alttab;
  }
  if (root->child[0] != NULL)
    alttab = create_subset (root->child[0], alttab, (ip << 1), netmask + 1);
  if (alttab == NULL) return NULL;
  if (root->child[1] != NULL)
    alttab = create_subset (root->child[1], alttab, (ip << 1)+1, netmask+1);
  return alttab;
}
