#include "iplookup.h"

cluster_t **create_cluster (rle_t **rle_seq, ip_addr_t *size, ip_addr_t *sort, ip_addr_t *tab_size) {
  cluster_t **RLE;
  cluster_t ***RLE_sort;
  ip_addr_t max, cl_len, cluster, clust_pos;
  ip_addr_t i, j, vlen;
  ip_addr_t exp_val = exp (32 - K_SIZE);
  ip_addr_t *run_len;
  usi_t *subs;
  cell_t *next_hop;
  RLE = (cluster_t **) malloc (exp (K_SIZE) * sizeof (cluster_t *));
  if (RLE == NULL) {
    printf ("Malloc failure creating RLE main structure\n");
    return NULL;
  }
  /*  calcolate max RLE size for cluster  */
  max = max_RLE_cluster_length (rle_seq, *size);
  /*  Allocate max len buffers for len and next hop  */
  run_len = (ip_addr_t *) malloc (max * sizeof (ip_addr_t));
  if (run_len == NULL) {
    printf ("Malloc failure creating RLE buffer\n");
    return NULL;
  }
  next_hop = (cell_t *) malloc (max * sizeof (cell_t));
  if (next_hop == NULL) {
    printf ("Malloc failure creating RLE buffer\n");
    return NULL;
  }
  cluster = 0;
  clust_pos = 0;
  cl_len = 0;
  vlen = 0;
  for (i = 0; i < *size; i ++) {
    if (rle_seq[i]->run_length < exp_val - cl_len) {
      cl_len += rle_seq[i]->run_length;
      /*  Buffer overflow  */
      if (vlen == max) {
	printf ("Buffer too small...\n");
	max += OVER_BUFFER;
	run_len = (ip_addr_t *) realloc ((ip_addr_t *) run_len, 
					 max * sizeof (ip_addr_t));
	if (run_len == NULL) {
	  printf ("Malloc failure creating RLE buffer\n");
	  return NULL;
	}
	next_hop = (cell_t *) realloc ((cell_t *) next_hop,
				       max * sizeof (cell_t));
	if (next_hop == NULL) {
	  printf ("Malloc failure creating RLE buffer\n");
	  return NULL;
	}
	printf ("corrected\n");
      }
      run_len[vlen] = rle_seq[i]->run_length;
      next_hop[vlen] = rle_seq[i]->next_hop;
      vlen ++;
      free (rle_seq[i]);
    }
    else {
      RLE[cluster] = (cluster_t *) malloc (sizeof (cluster_t));
      if (RLE[cluster] == NULL) {
	printf ("Malloc failure creating RLE cluster\n");
	return NULL;
      }
      RLE[cluster]->num = cluster;
      rle_seq[i]-> run_length -= (exp_val - cl_len);
      /*  Buffer overflow  */
      if (vlen == max) {
	printf ("Buffer too small...");
	max += OVER_BUFFER;
	run_len = (ip_addr_t *) realloc ((ip_addr_t *) run_len, 
					 max * sizeof (ip_addr_t));
	if (run_len == NULL) {
	  printf ("Malloc failure creating RLE buffer\n");
	  return NULL;
	}
	next_hop = (cell_t *) realloc ((cell_t *) next_hop,
				       max * sizeof (cell_t));
	if (next_hop == NULL) {
	  printf ("Malloc failure creating RLE buffer\n");
	  return NULL;
	}
	printf ("corrected\n");
      }
      run_len[vlen] = exp_val - cl_len;
      next_hop[vlen] = rle_seq[i]->next_hop;
      vlen ++;
      RLE[cluster]->vlen = vlen;
      RLE[cluster]->next_hop = (cell_t *) malloc (vlen * sizeof (cell_t));
      if (RLE[cluster]->next_hop == NULL) {
	printf ("Malloc failure creating RLE cluster\n");
	return NULL;
      }
      RLE[cluster]->run_length=(ip_addr_t*) malloc (vlen * sizeof (ip_addr_t));
      if (RLE[cluster]->run_length == NULL) {
	printf ("Malloc failure creating RLE cluster\n");
	return NULL;
      }
      for (j = 0; j < vlen; j ++) {
	RLE[cluster]->next_hop[j] = next_hop[j];
	RLE[cluster]->run_length[j] = run_len[j];
      }
      /*  inizialize counters and free sequence  */
      cl_len = 0;
      vlen = 0;
      /*  verify is not first cluster  */
      if (cluster > 0) {
	/*  Compare with previus cluster, if equal it is destroyed  */
	if (is_equal (RLE[cluster], RLE[cluster - 1]) == 0) {
	  free (RLE[cluster]->next_hop);
	  free (RLE[cluster]->run_length);
	  free (RLE[cluster]);
	  RLE[cluster] = NULL;
	  cluster --;
	}
      }
      /*  take note of order  */
      sort[clust_pos] = cluster;
      clust_pos ++;
      cluster ++;
      /*  If element is not fully used will be processed again  */
      if (rle_seq[i]->run_length != 0) i --;
      else free (rle_seq[i]);  /*  Free used element  */
    }
  }
  free (rle_seq);
  printf ("Total clusters %u\n",cluster);
  printf ("Sorting...");
  fflush (stdout);
  /* Create triple pointer vector to make sort untouching original structure */
  RLE_sort = (cluster_t ***) malloc (cluster * sizeof (cluster_t **));
  for (i = 0; i < cluster; i ++) RLE_sort[i] = &RLE[i];
  qsort ((void *) RLE_sort, cluster, sizeof (cluster_t **), comp_clust3); 
  printf ("done\n");
  printf ("Erasing useles clusters...");
  /*  Create subst vector and initialize this  */
  subs = (usi_t *) malloc (cluster * sizeof (usi_t));
  for (i = 0; i < cluster; i ++) subs[i] = i;  
  j = 0;
  cl_len = 0;
  /*  compare from second to last cluster  */
  for (i = 1; i < cluster; i ++) {
    /*  Cluster are equals  */
    if (is_equal (*RLE_sort[j], *RLE_sort[i]) == 0) {
      ip_addr_t ni = (*RLE_sort[i])->num;
      ip_addr_t nj = (*RLE_sort[j])->num;
      free ((*RLE_sort[i])->next_hop);
      free ((*RLE_sort[i])->run_length);
      free ((*RLE_sort[i]));
      *RLE_sort[i] = NULL;
      cl_len ++;
      /*  Take note of change in sort vector  */
      /*
         assert(ni==RLE_sort[i]-RLE);
         assert(nj==RLE_sort[j]-RLE);
      */
      subs[ni] = nj;
    }
    else j = i;  /*  go to next element  */
  }
  /*  apply changes to sort vector  */
  applay_changes (sort, subs);
  /*  free resources for updating sort vector  */
  free (RLE_sort);
  free (subs);
  printf ("done\n");
  /*  size will contain the len of RLE structure  */
  *size = cluster;
  /*  tab_size will contain the effective different clusters */
  *tab_size = cluster - cl_len;
  return RLE;
}

void applay_changes (ip_addr_t *sort, usi_t *subs) {
  ip_addr_t oldval, i, sost;
  ip_addr_t exp_val = exp (32 - K_SIZE);
  oldval = sort[0];
  sost = subs[sort[0]];
  while (sost != sort[0]) {
    sort[0] = sost;
    sost = subs[sort[0]];
  }
  for (i = 1; i < exp_val; i ++) {
    if (sort[i] == oldval) sort[i] = sort[i - 1];
    else {
      oldval = sort[i];
      sost = subs[sort[i]];
      while (sost != sort[i]) {
	sort[i] = sost;
	sost = subs[sort[i]];
      }
    }
  }
}

char is_equal (cluster_t *el1, cluster_t *el2) {
  ip_addr_t i;
  if (el1->vlen != el2->vlen) return -1;
  for (i = 0; i < el1->vlen; i ++) {
    if (el1->next_hop[i] != el2->next_hop[i]) return -1;
    if (el1->run_length[i] != el2->run_length[i]) return -1;    
  }
  return 0;
}

int comp_clust2 (const void *elem1, const void *elem2) {
  ip_addr_t i;
  const cluster_t **el1, **el2;
  el1 = (const cluster_t **) elem1;
  el2 = (const cluster_t **) elem2;  
  if (el1[0]->vlen < el2[0]->vlen) return 1;
  if (el1[0]->vlen > el2[0]->vlen) return -1;
  for (i = 0; i < el1[0]->vlen; i ++) {
    if (el1[0]->next_hop[i] < el2[0]->next_hop[i]) return 1;
    if (el1[0]->next_hop[i] > el2[0]->next_hop[i]) return -1;
    if (el1[0]->run_length[i] < el2[0]->run_length[i]) return 1;    
    if (el1[0]->run_length[i] > el2[0]->run_length[i]) return -1;    
  }
  return 0;
}

int comp_clust3 (const void *elem1, const void *elem2) {
  const cluster_t ***el1, ***el2;
  el1 = (const cluster_t ***) elem1;
  el2 = (const cluster_t ***) elem2;  
  return comp_clust2(*el1,*el2);
}

/*  
    Return the maximum number of different next hop in each cluster,
    This is only an euristic, let you trust me !!
*/
ip_addr_t max_RLE_cluster_length (rle_t **rle_seq, ip_addr_t size) {
  ip_addr_t exp_val = exp (32 - K_SIZE);
  ip_addr_t i, len, max, last, tmp;
  max = len = last = tmp = 0;
  for (i = 0; i < size; i ++) {
    len += rle_seq[i]->run_length;
    tmp ++;
    if (len % exp_val < last) {
       tmp = 0;
    }
    last = len % exp_val;
    if (tmp > max) max = tmp;
  }
  return (max == 0) ? 1 : max + OVER_BUFFER;
}

