#include "iplookup.h"

void dump (basic_t *vect, ip_addr_t tot_rules) {
  ip_addr_t i;
  printf ("%u\n",tot_rules); fflush (stdout);
  for (i = 0; i < tot_rules; i ++)
    printf ("%u) %u %u\n",i, vect[i].endpoint, vect[i].rule);
  printf ("--"); fflush (stdout);
}

basic_t *create_basic_intervcals (struct net_table *rules, ip_addr_t *tot_rules) {
  trie_t *trie;
  ip_addr_t tot_nodes;
  ip_addr_t i, j;
  basic_t **vect, *ret;
  trie = create_trie (rules, &tot_nodes);
  if (!trie) return NULL;
  vect = (basic_t **) malloc (tot_nodes * sizeof (basic_t *));
  if (vect == NULL) return NULL;
  for (i = 0; i < tot_nodes; i ++) {
    vect[i] = (basic_t *) malloc (sizeof (basic_t));
    if (vect[i] == NULL) return NULL;
  }
  *tot_rules = 0;
  visit_tree (trie, vect, tot_rules, 0, 0);
  j = 0;
  for (i = 1; i < *tot_rules; i ++)
    if (vect[j]->rule != vect[i]->rule) {
      j ++;
      vect[j]->rule = vect[i]->rule;
      vect[j]->endpoint = vect[i]->endpoint;
    }
  *tot_rules = j + 1;
  ret = (basic_t *) malloc (*tot_rules * sizeof (basic_t));
  for (j = 0; j < *tot_rules; j ++) {
    ret[j].rule = vect[j]->rule;
    ret[j].endpoint = vect[j]->endpoint;
  }
  for (j = 0; j < tot_nodes; j ++) {
    free (vect[j]);
  }
  free (vect);
  free_trie (trie);
  return ret;
}

void free_trie (trie_t *root) {
  if (root->child[0] != NULL) free_trie (root->child[0]);
  if (root->child[1] != NULL) free_trie (root->child[1]);
  free (root);
}

void visit_tree (trie_t *root, basic_t **vect, ip_addr_t *RLE_ind, ip_addr_t ip, ip_t netmask) {
  /*  if is a leaf  */
  if (root->child[0] == NULL && root->child[1] == NULL) {
    ip = (ip << (32 - netmask));
    vect[*RLE_ind]->rule = root->next_hop;
    vect[*RLE_ind]->endpoint = ip;
    *RLE_ind = *RLE_ind + 1;
    return;
  }
  if (root->child[0] != NULL)
    visit_tree (root->child[0], vect, RLE_ind, (ip << 1), netmask + 1);
  if (root->child[1] != NULL)
    visit_tree (root->child[1], vect, RLE_ind, (ip << 1)+1, netmask+1);
}
