#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/time.h>
#include <unistd.h>

typedef unsigned char cell_t;     /*  next hop  */
typedef unsigned char ip_t;       /*  IP element from 0 to 255  */
typedef unsigned int ip_addr_t;   /*  IPv4 entire address  */
typedef unsigned short int col_index_t; /*  A type for colunn index  */
typedef unsigned char col_short_t;  /*  Short data value for colun index  */
typedef short int bucket_t;       /*  for bucket colun index  */

typedef unsigned short int usi_t; /*  Short unsigned int  */

#if FINE_TIME
typedef long long int my_time_t;
#endif


#define TRUE 0
#define FALSE 1

#define MAX_IP_LUNG_AS_STRING 16  /*  strlen ("XXX.XXX.XXX.XXX");  */
#define MAX_RIGA_READ 256
#define MAX_NEXT_HOPS 256   /*  Maximum number of next hops in table  */
#define MAX_SHORT_COL 256   /*  Maxinum value storable in col_short_t   */
#define K_SIZE 16
#define	OVER_BUFFER 1   /*  Usually this is a good value!  */
#define FOLLOW_UP MAX_NEXT_HOPS - 1  /*  match everything */
#define K_PART  8           /*  K in K partition  */

#define first_k(x) (x >> (32 - K_SIZE))
#define exp(x) (1 << (x))
#define bit_k(x,k) ((x & (1 << (k))) >> (k))
#define Exp(x) (1 << (x))
#define Ph_1(x,k) ((x >> (32 - k)) + 1) << (32 - k)

#if FINE_TIME
#define rdtscll(val) \
     __asm__ __volatile__("rdtsc" : "=A" (val))
#endif

typedef struct {
  unsigned int len;
  ip_addr_t endpoints[2];
} N_elems_t;

typedef N_elems_t K_part_t;

typedef struct {
  ip_addr_t len;
  N_elems_t *kids;
} N_part_t;

typedef struct {
  ip_addr_t endpoint;
  cell_t rule;
} basic_t;

struct net_table {
  ip_addr_t ip;  
  ip_t netmask;
  cell_t next_hop;
  struct net_table *next;
};

typedef struct {
  cell_t next_hop;       /*  Next hop  */
  ip_addr_t run_length;  /*  length */
} rle_t;

typedef struct {
  ip_addr_t num;         /*  cluster number (rebuild right order after sort) */
  ip_addr_t vlen;
  cell_t *next_hop;
  ip_addr_t *run_length;
} cluster_t;

/*  The size of IP halves is unknown because depend from K_SIZE costant  */
typedef struct {
  usi_t first_pair;      /*   first part of ip address  */
  usi_t last_pair;       /*  second part of ip address  */
  cell_t next_hop;
  void *next;
} addr_list_t;

/*  this type havn't the pointer next. That need for optimize cache use in
    lookup operation  */
typedef struct {
  usi_t first_pair;      /*   first part of ip address  */
  usi_t last_pair;       /*  second part of ip address  */
  cell_t next_hop;
} addr_table_t;


typedef struct {
  cell_t next_hop;
  ip_t mark; 
  void *child[2];
} trie_t;

/*  createRLE.c  */
cluster_t **create_cluster (rle_t **rle_seq, ip_addr_t *size, ip_addr_t *sort, ip_addr_t *tab_size);
void applay_changes (ip_addr_t *sort, usi_t *subs);
char is_equal (cluster_t *el1, cluster_t *el2);
int comp_clust2 (const void *elem1, const void *elem2);
int comp_clust3 (const void *elem1, const void *elem2);
ip_addr_t max_RLE_cluster_length (rle_t **rle_seq, ip_addr_t size);

/*  createipt.c  */
cell_t **rebuild_table (rle_t **RLE_seq, char flag, ip_addr_t tot_route, ip_addr_t *tot_rows, col_index_t *row_len);
void create_colun_index (col_index_t *col_index, cluster_t **RLE, ip_addr_t RLE_size, ip_addr_t exp_col);
char createrow (cell_t *row, ip_addr_t *metric, col_index_t size, cluster_t *RLE, col_index_t row_len);
cell_t **create_ip_table (cluster_t **cluster, ip_addr_t row_len, ip_addr_t *sort, ip_addr_t *metric, ip_addr_t met_len, ip_addr_t tab_len);
ip_addr_t *create_metrics (col_index_t *col_index);

/*  loadroute.c  */
int ins_in_next_hop_array (char **next_hops, char *next_hop);
ip_t used_bits (ip_t elem);
ip_t calcola_netmask (ip_t ip1, ip_t ip2, ip_t ip3, ip_t ip4);
ip_addr_t correct_nerwork (ip_addr_t ip, ip_t netmask);
void unload_next_hops (char **next_hops);
struct net_table *load_table (char *nome, char **next_hops,ip_addr_t *rou_c, char flag);

/*  supportfn.c  */
#if FINE_TIME
double detect_cpu_clock (void);
#endif
long int Mark_Time (void);
char eval_params (int argc, char *argv[], ip_addr_t *ntimes);
void echo_use (void);
void help (void);
void write_log_file (ip_addr_t tot_rows, ip_addr_t b_size, ip_addr_t b_tmp, char flag, ip_addr_t memory, long tempo, ip_addr_t layer2);
#if FINE_TIME
void write_log_lookup (long *look_time, double cpu_speed);
#else
void write_log_lookup (long *look_time);
#endif
void write_results (addr_table_t **array, ip_addr_t len, char **next_hop);
long no_lookup_time (addr_table_t **array, ip_addr_t len, ip_addr_t ntimes);

/*  t_backend.c  */

/*  cdglookup.c  */
int main (int argc, char *argv[]);

/*  load_addr.c  */
addr_table_t **load_address (char *nome, ip_addr_t *len);
void unload_address (addr_table_t **array, ip_addr_t len);
int load_data (addr_list_t *self, char *riga);

/*  cdgupdate.c  */
trie_t *create_trie (struct net_table *in_table, ip_addr_t *tot_nodes);
trie_t *create_root (void);
int trie_insert_network (struct net_table *elem, trie_t *root);
rle_t **build_RLE_sequence (rle_t **RLE_seq, ip_addr_t *RLE_len);
rle_t **build_empty_RLE_sequence (ip_addr_t len);
int trie_remove_network (ip_addr_t ip, ip_t netmask, trie_t *root);
ip_addr_t merge_leaves (trie_t *root, ip_addr_t delete);
void cover_holes (trie_t *root);
rle_t **update_RLE_sequence (trie_t *root, ip_addr_t *RLE_len, char flag);
void visit_b_tree_sup (trie_t *root, rle_t **RLE, ip_addr_t *RLE_ind, ip_addr_t ip, ip_t netmask);
void visit_b_tree_inf (trie_t *root, rle_t **RLE, ip_addr_t *RLE_ind, ip_addr_t ip, ip_t netmask);
void visit_b_tree (trie_t *root, rle_t **RLE, ip_addr_t *RLE_ind, ip_addr_t ip, ip_t netmask);

/*  mk_lookup.c  */
#ifdef USE_N_PARTITION
long make_lookups (cell_t **row, addr_table_t **array, ip_addr_t len,
                   ip_addr_t ntimes, N_part_t *npart, ip_addr_t tot_bas,
                   basic_t *basic);
#endif
#ifdef USE_K_PARTITION
long make_lookups (cell_t **row, addr_table_t **array, ip_addr_t len,
		   ip_addr_t ntimes, K_part_t *kpart, basic_t *basic);
#endif
#ifdef USE_BIN_SEARCH
long make_lookups (cell_t **row, addr_table_t **array, ip_addr_t len,
		   ip_addr_t ntimes, ip_addr_t tot_bas, basic_t *basic);
#endif
#ifdef USE_TRIE
long make_lookups (cell_t **row, addr_table_t **array, ip_addr_t len,
		   ip_addr_t ntimes, trie_t *root);
#endif


/*  endpoint.c  */
void dump (basic_t *vect, ip_addr_t tot_rules);
basic_t*create_basic_intervcals(struct net_table *rules, ip_addr_t *tot_rules);
void free_trie (trie_t *root);
void visit_tree (trie_t *root, basic_t **vect, ip_addr_t *RLE_ind,
                   ip_addr_t ip, ip_t netmask);

/*  support.c  */
ip_addr_t binary_search (basic_t *basic, ip_addr_t len, ip_addr_t query);
void write_log (addr_table_t **addresses, ip_addr_t tot_addr, char **hp);
cell_t trie_lookup (trie_t *trie, ip_addr_t lookup);
ip_addr_t trie_mem (trie_t *root);

/*  Npart.c  */
ip_addr_t query_N_partition (N_part_t *root, ip_addr_t query, basic_t *points,
                             ip_addr_t len);
void dump_N_partition (N_part_t *root);
N_part_t *create_N_partition (basic_t *points, ip_addr_t len, ip_addr_t *mem);

/*  Kpart.c  */
ip_addr_t query_K_partition (K_part_t *part, ip_addr_t query, basic_t *points);
void dump_K_partition (K_part_t *part);
K_part_t *create_K_partition (basic_t *points, ip_addr_t len, ip_addr_t *mem);

/*  minitrie.c  */
trie_t *create_mini (struct net_table *in_table);
int mini_insert (struct net_table *elem, trie_t *root);
