#include "iplookup.h"

#if FINE_TIME
double detect_cpu_clock (void) {
  struct timeval tm_begin, tm_end;
  my_time_t tsc_begin, tsc_end;
  gettimeofday (&tm_begin, 0);
  rdtscll (tsc_begin);
  gettimeofday (&tm_begin, 0);
  usleep (1000000);
  rdtscll (tsc_end);
  gettimeofday (&tm_end, 0);
  return (tsc_end - tsc_begin) / (tm_end.tv_sec - tm_begin.tv_sec +
                                  (tm_end.tv_usec - tm_begin.tv_usec) / 1e6);
}
#else
long int Mark_Time (void) {
  static struct timeval TimeStart;
  struct timeval TimeEnd;
  static char set = 0;
  if (set == 0) {
    set ++;
    gettimeofday (&TimeStart, NULL);
    return 0;
  }
  else {
    set = 0;
    gettimeofday (&TimeEnd, NULL);
    return ((TimeEnd.tv_sec - TimeStart.tv_sec) * 1000000) +
      (TimeEnd.tv_usec - TimeStart.tv_usec);
  }
}
#endif
char eval_params (int argc, char *argv[], ip_addr_t *ntimes) {
  short int i;  
  char flag = 0;
  if ((argc < 2) || (argc > 5)) echo_use ();
  switch (argc) {
  case 2:
    if (!strcmp (argv[1] ,"-H") || !strcmp (argv[1] ,"-h")) {
      help ();
      exit (1);
    }
    else echo_use ();
    break;
  case 3:
    for (i = 0; i < strlen (argv[1]); i ++)
      if (argv[1][i] < '0' || argv[1][i] > '9') {
	flag = 1;
	break;
      }
    if (flag == 0)
      *ntimes = atol (argv[1]);
    flag = 0;
    break;
  case 4:
    if (argv[1][0] == '-') {
      for (i = 0; i < strlen (argv[2]); i ++)
	if (argv[2][i] < '0' || argv[2][i] > '9') {
	  flag = 1;
	  break;
	}
      if (flag == 0) {
	*ntimes = atol (argv[2]);
	if (*ntimes != 0) echo_use ();
      }
      flag = 0;
      /*  ignore - at start of string  */
      for (i = 1; i < strlen (argv[1]); i ++) {
        if (toupper (argv[1][i]) == 'L') flag ++;
        if (toupper (argv[1][i]) == 'H') {
	  help ();
	  exit (1);
	}
      }
    }
    else {
      for (i = 0; i < strlen (argv[1]); i ++)
	if (argv[1][i] < '0' || argv[1][i] > '9') {
	  flag = 1;
	  break;
	}
      if (flag == 0)
	*ntimes = atol (argv[1]);
      else 
	echo_use ();
      flag = 0;
      if (*ntimes == 0) {
	if (argv[2][0] == '-') {
	  /*  ignore - at start of string  */
	  for (i = 1; i < strlen (argv[2]); i ++) {
	    if (toupper (argv[2][i]) == 'L') flag ++;
	    if (toupper (argv[2][i]) == 'H') {
	      help ();
	      exit (1);
	    }
	  }
	}
	else echo_use ();
      }
    }
    break;
  case 5:
    if (argv[1][0] == '-') {
      *ntimes = atol (argv[2]);
      /*  ignore - at start of string  */
      for (i = 1; i < strlen (argv[1]); i ++) {
        if (toupper (argv[1][i]) == 'L') flag ++;
        if (toupper (argv[1][i]) == 'H') {
	  help ();
	  exit (1);
	}
      }
    }
    else {
      *ntimes = atol (argv[1]);
      /*  ignore - at start of string  */
      for (i = 1; i < strlen (argv[2]); i ++) {
        if (toupper (argv[2][i]) == 'L') flag ++;
        if (toupper (argv[2][i]) == 'H') {
	  help ();
	  exit (1);
	}
      }
    }
    break;
  }
  return flag;
}

void echo_use (void) {
  printf ("Use: cdglookup [iteration] [-LH] <table file> [addr. file]\n");
  printf ("Try 'cdglookup -H' for more information.\n");
  exit (-2);
}

void help (void) {
  printf ("cdglookup [iteration] [-LH] <table file> [addr. file]\n\n");
  printf ("iteration is the number of times that, lookups of each address\n");
  printf ("in address file, will be performed\n");
  printf ("if iteration is 0 no lookups will be performed and address file\n");
  printf ("parameter is unspecified\n");
  printf ("Options:\n");
  printf ("-H to obtain this help\n");
  printf ("-L to append to cdg.log basic information about table\n");
  printf ("Returned value\n");
  printf (" 0 if success\n");
  printf ("-1 if error\n");
  printf ("-2 if parameter error\n");
  printf (" 1 If help was requested\n");
}

void write_log_file (ip_addr_t tot_rows, ip_addr_t b_size, ip_addr_t b_tmp, char flag, ip_addr_t memory, long tempo, ip_addr_t layer2) {
  FILE *log;
  ip_addr_t i;
  if ((log = fopen ("cdg.log","a")) == NULL) {
    printf ("Unable to create log file\n");
    return;
  }
  fprintf(log,"Layer 1 size (rows x cols) %u x 256 ",tot_rows);
  /*  Compute layer 1 size  */
  i = ((tot_rows * 256) * sizeof (cell_t)) + 
    (sizeof (cell_t *) * exp (K_SIZE));
  fprintf(log,"bytes %u\n",i);
  fprintf(log,"Layer 2 occuped bytes %u\n",memory);
  fprintf (log,"%u entries in second level table\n", layer2);
  i += memory;
  fprintf (log,"Bytes = %u   Kb = %.2f   Mb = %.2f\n",
	   i, (double) i / 1024, (double) i / 1048576); 
  fprintf (log,"Elapsed time %ld\n", tempo);
  fclose (log);
}

#if FINE_TIME
void write_log_lookup (long *look_time, double cpu_speed) {
  double total_time;
#else
void write_log_lookup (long *look_time) {
#endif
  FILE *log;
  if ((log = fopen ("cdg.log","a")) == NULL) {
    printf ("Unable to create log file\n");
    return;
  }
  fprintf (log, "-- Lookup time --\n");
#if FINE_TIME
  total_time = (1000000 / cpu_speed) * look_time[1];
  fprintf (log,"Entire lookup procedure %f\n", total_time);
  total_time = (1000000 / cpu_speed) * look_time[0];
  fprintf (log,"No lookup operations %f\n", total_time);
  total_time = (1000000 / cpu_speed) * (look_time[1] -look_time[0]);
  fprintf (log,"Real lookup time %f\n", total_time);
#else
  fprintf (log, "Entire lookup procedure %ld\n",look_time[1]);
  fprintf (log, "No lookup operations %ld\n",look_time[0]);
  fprintf (log, "Real lookup time %ld\n",look_time[1] - look_time[0]);
#endif
  fclose (log);
}

void write_results (addr_table_t **array, ip_addr_t len, char **next_hop) {
  ip_addr_t i;
  FILE *pf;
  /*  Open log file  */
  if ((pf = fopen ("cdglookup.log","w")) == NULL) {
    printf ("Unable to write log file\n");
    return;
  }
  /*  Store lookups results in log file  */
  for (i = 0; i < len; i ++) {
    /*  write the address looked up  */
    fprintf (pf,"%u.%u.%u.%u -> ",
	     array[i]->first_pair >> 8,
	     array[i]->first_pair % 256,
	     array[i]->last_pair >> 8,
	     array[i]->last_pair % 256); 
    /*  array[i]->next_hop contain the index of next_hop array, 
	by costruction in ins_in_next_hop_array function         */
    fprintf (pf, "%s\n",next_hop[array[i]->next_hop]);
  }
  /*  Close */
  fclose (pf);
}

long no_lookup_time (addr_table_t **array, ip_addr_t len, ip_addr_t ntimes) {
  register ip_addr_t i,j;
#if FINE_TIME
  my_time_t start_time ,stop_time;
  rdtscll (start_time);
#else
  Mark_Time ();
#endif
  /*  repeat the same operation n times  */
  for (i = 0; i < ntimes; i ++) {
    /*  Each loop is a lookup  */
    for (j = 0; j < len; j ++) {
      /*  empty assigment  */
      array[j]->next_hop = 0;
    }
  }
#if FINE_TIME
  rdtscll (stop_time);
  return stop_time - start_time;
#else
  return Mark_Time ();
#endif
}
