#include "iplookup.h"

quadr_t **create_quads_S (struct net_table *in_table, ip_addr_t *tot_route) {
  quadr_t **vect;       /*  The triple vector  */
  quadr_t *list, *temp;
  ip_addr_t tot_triple = 0;  /*  Inserted triple  */
  ip_addr_t i, useless_count;
  ip_addr_t multi_hop = 0;   /*  count multi hop  */
  struct net_table *tmp;
  if (*tot_route > exp (31)) {
    printf ("Routing lable too large\n");
    return NULL;
  }
  printf ("Creating quadruple vector...");
  /*  I will crreate 2 quadruple for each in_table entry
      This malloc potentially allocate mutch memory and may fail
      In real cases this will never happen
      for 100.000 entry allocate 781Kb
  */
  vect = (quadr_t **) malloc ((*tot_route * 2) * sizeof (quadr_t *));
  if (vect == NULL) {
    printf ("fail\nMalloc failure creating quadruple vector... can't go on\n");
    return NULL;
  }
  /*  Create S vector  */
  while (in_table != NULL) {
    vect[tot_triple] = (quadr_t *) malloc (sizeof (quadr_t));
    if (vect[tot_triple] == NULL) {
      printf ("fail\nMalloc failure creating triple entry... can't go on\n");
      return NULL;
    }
    vect[tot_triple]->ip_0 = in_table->ip;
    /*  netmask equal to 32 is an exception  */
    if (in_table->netmask != 32) 
      vect[tot_triple]->netmask = in_table->netmask;
    else
      vect[tot_triple]->netmask = 0;
    vect[tot_triple]->next_hop = in_table->next_hop;
    vect[tot_triple]->tipo = 0;
    tot_triple ++;
    vect[tot_triple] = (quadr_t *) malloc (sizeof (quadr_t));
    if (vect[tot_triple] == NULL) {
      printf ("fail\nMalloc failure creating triple entry... can't go on\n");
      return NULL;
    }
    vect[tot_triple]->ip_0 = ip_1 (in_table->ip, in_table->netmask);
    /*  netmask equal to 32 is an exception  */
    if (in_table->netmask != 32) 
      vect[tot_triple]->netmask = 32 - in_table->netmask;
    else
      vect[tot_triple]->netmask = 32;
    vect[tot_triple]->next_hop = in_table->next_hop;
    vect[tot_triple]->tipo = 1;
    tot_triple ++;
    /*  free net table entry now become useless  */
    tmp = in_table;
    in_table = in_table->next;
    free (tmp);
  }
  printf ("done\n");
  /*  Sort lexicografically vect  */
  printf ("Sorting..."); 
  qsort ((void *) vect, tot_triple, sizeof (quadr_t **), compar); 
  printf ("done\n");
  /*  Create ordered list from vector  */
  for (i = 0; i < tot_triple - 1; i ++) {
    vect[i]->next = vect[i + 1];
  }
  /*  Insert epsilon case all 1, will always the first element  */  
  vect[tot_triple - 1]->next = (quadr_t *) malloc (sizeof (quadr_t));
  if (vect[tot_triple - 1]->next == NULL) {
    printf ("fail\nMalloc failure creating triple entry... can't go on\n");
    return NULL;
  }
  list = vect[tot_triple - 1]->next;
  list->ip_0 = ~0;  /*  11111....   */
  list->netmask = 32;
  list->tipo = 1;
  list->next_hop = 0;  /*  NO ROUTE TO HOST  */
  list->next = NULL;   /*  this is the last element   */
  /*  illegal value, it will never read, but I prefere to inizialize it  */
  list->l_val = LAST_L_VALUE;  
  tot_triple ++;
  /*  Go to first element of list  */
  list = vect[0];
  /*  Free vect, from now, array become only list  */
  free (vect);
  /*  Insert epsilon case all 0, will always the first element  */
  temp = (quadr_t *) malloc (sizeof (quadr_t));
  if (temp == NULL) {
    printf ("fail\nMalloc failure creating triple entry... can't go on\n");
    return NULL;
  }
  temp->ip_0 = 0;
  temp->netmask = 0;
  temp->next_hop = 0;  /*  NO ROUTE TO HOST  */
  temp->tipo = 0;
  temp->next = list;
  tot_triple ++;
  /* Set new start point for list  */
  list = temp;
  /*  Cover holes between networks  */
  tot_triple += ins_holes_quads (list, &i);
  if (i != 0) {
    printf ("Malloc failure creating quadruple entry... can't go on\n");
    return NULL;
  }
  /*  rebuild vect after inserting NO ROUTE TO HOST cases  */
  vect = (quadr_t **) malloc (tot_triple * sizeof (quadr_t *));
  if (vect == NULL) {
    printf ("Malloc failure creating quadruple vector... can't go on\n");
    return NULL;
  }
  /*  Create vector from list  */
  temp = list;
  for (i = 0; i < tot_triple; i ++) {
    vect[i] = temp;
    temp = temp->next;
  }
  /*  for each quadr calcolate value of l and free useless  */   
  useless_count = 0;
  for  (i = 0; i < tot_triple - 1; i ++) {
    /*  distance from a to a'  */
    vect[i]->l_val = vect[i + 1]->ip_0 - vect[i]->ip_0 + 1; /* is inclusive  */
    if (vect[i]->tipo == vect[i + 1]->tipo) vect[i]->l_val --;
    if (vect[i]->tipo == 1 && vect[i + 1]->tipo == 0) {
      /*  Because I use unsigned value  */
      if (vect[i]->l_val >= 2) vect[i]->l_val -= 2;
      else { 
	/*  i can never be 0 because first quadr is <0,0,0,0>  */ 
	if (vect[i - 1] != NULL) {
	/*  This will always happen, but control is performed to avoid
	    segmentation fault  */
	  vect[i - 1]->l_val -= vect[i]->l_val;
	}
	vect[i]->l_val = 0;
      }
    }
    /*  If useless destroy  */
    if (vect[i]->l_val == 0) {
      if (vect[i]->ip_0 == vect[i + 1]->ip_0 && 
	  vect[i]->tipo == vect[i + 1]->tipo && 
	  vect[i]->netmask == vect[i + 1]->netmask)
	multi_hop ++;
      useless_count ++;
      free (vect[i]);
      vect[i] = NULL;
    }
  }
  if (multi_hop)
    printf ("Warning %u multi hop route found\n", multi_hop);
  printf ("%u quadruple (%u useless)\n",tot_triple, useless_count);
  *tot_route = tot_triple;
  return vect;
}

int ins_quad (quadr_t *el, ip_addr_t ip, ip_t net, ip_t t, cell_t nh) {
  quadr_t *val;
  val = (quadr_t *) malloc (sizeof (quadr_t));  
  if (val == NULL) return 1;
  val->ip_0 = ip;
  val->netmask = net;
  val->tipo = t;
  val->next_hop = nh;
  val->next = el->next;
  el->next = val;
  return 0;
}

ip_addr_t ins_holes_quads (quadr_t *list, ip_addr_t *err) {
  quadr_t *self, *temp;
  ip_addr_t tot = 0;
  cell_t stack[256];
  int depth;
  stack[0] = 0;  /* base is NO ROUTE TO HOST  */  
  self = list;
  temp = list->next;
  tot = depth = 0;
  while (temp != NULL) {
    if ((self->tipo == 1) && (temp->ip_0 - self->ip_0 > 1)) {
      *err = ins_quad (self, self->ip_0 + 1, 0, 0, stack[depth]);
      if (*err != 0) return 0;
      self = self->next;
      *err = ins_quad (self, temp->ip_0 - 1, 32, 1, stack[depth]);
      if (*err != 0) return 0;
      self = self->next;
      tot += 2;
    }
    if (temp->tipo == 0) {  /* Warning, remember I use temp, not self !!  */
      depth ++;
      stack[depth] = temp->next_hop;
    }
    else depth --;
    self = self->next;
    temp = temp->next;
  }
  *err = 0;
  return tot;
}

int compar (const void *elem1, const void *elem2) {
  quadr_t **el1, **el2;
  /*  Cast need for use structure  */
  el1 = (quadr_t **) elem1;
  el2 = (quadr_t **) elem2;
  if (el1[0]->ip_0 < el2[0]->ip_0) return -1;
  if (el1[0]->ip_0 > el2[0]->ip_0) return 1;
  if (el1[0]->tipo < el2[0]->tipo) return -1;
  if (el1[0]->tipo > el2[0]->tipo) return 1;
  if (el1[0]->netmask < el2[0]->netmask) return -1;
  if (el1[0]->netmask > el2[0]->netmask) return 1;
  /* If arrive here mean I'm comparing 2 occurrences of the same network
     and I choose arbitrarily one of these  */
  if (el1[0]->next_hop < el2[0]->next_hop) return -1;
  if (el1[0]->next_hop > el2[0]->next_hop) return 1;
  return 0;
}

ip_addr_t ip_1 (ip_addr_t address, ip_t netmask) {
  if (netmask == 32) return address;
  return address | ((1 << (32 - netmask)) - 1);
}
 
