#include "iplookup.h"

cell_t **rebuild_table (rle_t **RLE_seq, char flag, col_index_t *col_index, ip_addr_t tot_route, ip_addr_t *tot_rows, col_index_t *row_len) {
  cluster_t **cluster = NULL;
  cluster_t ***clusters = NULL;
  ip_addr_t *sort = NULL;
  ip_addr_t *metric;
  ip_addr_t exp_col = exp (K_SIZE);
  cell_t **row_index;
  cell_t **block, **match;
  col_index_t **graph, *subs;
  ip_addr_t i;
  if (flag & 8) 
    sort = (ip_addr_t *) malloc (exp (K_SIZE + 1) * sizeof (ip_addr_t));
  else
    sort = (ip_addr_t *) malloc (exp (K_SIZE) * sizeof (ip_addr_t));
  if (sort == NULL) {
    printf ("Malloc failure allocating sort vector... can't go on\n");
    return NULL;
  }
  if (flag & 8) {
    clusters = create_split_cluster (RLE_seq, &tot_route, sort, tot_rows);  
    if (clusters == NULL) return NULL;
  }
  else {
    cluster = create_cluster (RLE_seq, &tot_route, sort, &tot_rows[0]);  
    if (cluster == NULL) return NULL;
  }
  i = exp_col >> 1;
  if (flag & 8) {
    create_colun_index (col_index, clusters[0], tot_route, i); 
    create_colun_index (col_index + i, clusters[1], tot_route, i);
  }
  else {
    create_colun_index (col_index, cluster, tot_route, exp (32 - K_SIZE)); 
  }
  metric = create_metrics (col_index);
  if (metric == NULL) return NULL;
  if (flag & 8) {   /*  splitted cdg table  */
    row_len[0] = col_index[i - 1] + 1;
    row_len[1] = col_index[exp_col - 1] + 1;
    block = create_block (clusters, tot_route, metric, row_len);
    /*  adjust sort values  */
    for (i = exp (32 - K_SIZE); i < exp (32 - K_SIZE + 1); i ++)
      sort[i] += tot_route;
    printf ("tables size (rows x cols) %u x %u | %u x %u\n",
	    tot_rows[0], row_len[0], tot_rows[1], row_len[1]);
#ifdef NO_GRAPH
    match = NULL; 
#else
    match = create_match_set (block, tot_rows, row_len, tot_route);  
#endif
    if (!match) {
#ifndef NO_GRAPH
      printf ("Match error...defoult will be used\n");
#endif
    }
    else {
      /*  graph size row_len[0] x row_len[1]  */
      graph = create_bipartite_graph (match, tot_rows, row_len);
      destroy_match_set (match);
      if (!graph) 
	printf ("Match error...defoult will be used\n");
      else {
	subs = find_perfect_matching (graph, row_len);
	destroy_graph (graph, row_len[0]);
	if (!subs)
	  printf ("Match error...defoult will be used\n");
	else {
	  if (row_len[0] > row_len[1]) {
	    make_permutations (block + tot_route, 
			       tot_rows[1], row_len[1], row_len[0], 
			       col_index + (exp_col >> 1), subs);
	  }
	  else {
	    make_permutations (block, tot_rows[0], row_len[0], 
			       row_len[1], col_index, subs);
	  }
	}
	/*  free serource  */
	free (subs);
      }
    }
    /*  Erase holes in data strucure  */
    if (compress_structure (block, sort, tot_route << 1)) return NULL;
    row_len[0] = (row_len[0] < row_len[1]) ? row_len[1] : row_len[0];
    i = erase_rows (block, sort, tot_rows[0] + tot_rows[1], row_len[0]);
    if (i == ~0) return NULL;  /* ~0 is the error return value */
    tot_rows[0] = tot_rows[0] + tot_rows[1] - i;
    /*  Create the row index structure  */
    row_index = shuffle (block, sort);
    if (row_index == NULL) return NULL;
  }
  else {  /*  Normal cdg table (not splitted)  */
    /*  Create table */
    row_len[0] = col_index[exp (32 - K_SIZE) - 1] + 1;  /*  by costruction  */
    row_index = create_ip_table (cluster, tot_route, sort, metric, 
				 row_len[0], tot_rows[0]);
    if (row_index == NULL) return NULL;
  }
  printf ("table size (rows x cols) %u x %u\n",tot_rows[0], row_len[0]);
  return row_index;
}


void *build_col_index (col_index_t *col_index, char flag, col_index_t *row_len, ip_addr_t *b_size, ip_addr_t *b_tmp, char *comp_col) {
  bucket_t *col_bucket = NULL;
  col_short_t *col_short = NULL;
  ip_addr_t i, tmp;
  /*  Verify if colunn index is compressable in short type  */
  if (flag & 8) {   /*  splitted cdg table  */
    if (row_len[0] < MAX_SHORT_COL && row_len[1] < MAX_SHORT_COL) *comp_col =1;
  }
  else {   /*  Normal cdg table (not splitted)  */
    if (row_len[0] < MAX_SHORT_COL) *comp_col = 1;
  }
  /* Force use normal col_index or comressed col_index NO bucket index  */
  if (flag & 32) {
  /*  Compress colunn index in shorter type if possible and necessary */
    if (*comp_col == 1) {
      col_short = compress_col_index_type (col_index);
      if (col_short == NULL) 
	*comp_col = 0;  /* col_index is not compressed */
      else
	return (void *) col_short;
    }
    return (void *) col_index;
  }
  /*  compute the short possible colun index type  */
  *b_size = compute_bucket_size (col_index);
  *b_tmp = tot_bucket_fault (col_index, *b_size);
  /*  compute size of col index or, if possible, short index  */
  i = (exp (K_SIZE) * sizeof (col_index_t)) >> *comp_col;
  /*  compute bucket index size  */
  tmp = ((exp (K_SIZE) / *b_size) + (*b_size * *b_tmp)) * sizeof (bucket_t);
  /*  choose the shortest  */
  if (i > tmp) {
    /*  create bucket colun index  */
    col_bucket = create_bucket_col_index (col_index);
    if (col_bucket != NULL) {
      *comp_col = 2;
      return (void *) col_bucket;
    }
  }
  /*  Compress colunn index in shorter type if possible and necessary */
  if (*comp_col == 1) {
    col_short = compress_col_index_type (col_index);
    if (col_short == NULL)
      *comp_col = 0;  /* col_index is not compressed */
    else
      return (void *) col_short;
  }
  return (void *) col_index;
}


void *rebuild_col_index (col_index_t *col_index, char flag, col_index_t *row_len, ip_addr_t *b_size, ip_addr_t *b_tmp, char comp_col) {
  bucket_t *col_bucket = NULL;
  col_short_t *col_short = NULL;
  ip_addr_t tmp;
  if (comp_col == 1) {
    /*  Verify if colunn index is compressable in short type  */
    if (flag & 8) {   /*  splitted cdg table  */
      if (row_len[0] > MAX_SHORT_COL || row_len[1] > MAX_SHORT_COL) {
	printf ("Colun index too large for this lookup procedure\n");
	printf ("Try cdglookup -g\nProgramm aborted\n");
	exit (-1);
      }
    }
    else {   /*  Normal cdg table (not splitted)  */
      if (row_len[0] > MAX_SHORT_COL) {
	printf ("Colun index too large for this lookup procedure\n");
	printf ("Try cdglookup -g\nProgramm aborted\n");
	exit (-1);
      }
    }
    col_short = compress_col_index_type (col_index);
    if (col_short == NULL) {
      printf ("Error compressing colunn index\nProgramm aborted");
      exit (-1);
    }
    else
      return (void *) col_short;
  }
  /*  compute the short possible colun index type  */
  *b_size = compute_bucket_size (col_index);
  *b_tmp = tot_bucket_fault (col_index, *b_size);
  /*  compute bucket index size  */
  tmp = ((exp (K_SIZE) / *b_size) + (*b_size * *b_tmp)) * sizeof (bucket_t);
  /*  create bucket colun index  */
  col_bucket = create_bucket_col_index (col_index);
  if (col_bucket != NULL) {
    return (void *) col_bucket;
  }
  printf ("Error creating bucket index\nProgramm aborted");
  exit (-1);
}


