#include "iplookup.h"
#include <assert.h>

ip_addr_t min16 = 0;
ip_addr_t p1624 = 0;
ip_addr_t max24 = 0;
ip_addr_t pref24 = 0;
u_int64_t modc2 = 0;

void print_stat (void) {
  printf ("Min 16 %u Tra 16 e 24 %u Mag 24 %u\n",min16, p1624, max24);
  printf ("Pref /24 %u Media caso b %llu\n",pref24, modc2 / p1624);
}

void update (trie_t *root, update_t *request) {
  if (request->netmask < 16) min16 ++;
  if ((request->netmask >= 16) && (request->netmask <= 24)) p1624 ++;
  if (request->netmask == 24) pref24 ++;
  if (request->netmask > 24) max24 ++;
  if (request->op == 0) 
    update_announce (root, request);
  else    
    update_withdraw (root, request);
}

void update_withdraw (trie_t *root, update_t *request) {
  ip_t i;
  int kid;
  trie_t *node, *father = NULL, *root16 = NULL;
  ip_addr_t rows = 0, j, t = 0;
  cell_t riga1[256], riga2[256];
  printf ("Withdraw %u.%u.%u.%u/%u -> %u\n",request->ip >> 24, 
	  (request->ip >> 16) % 256, (request->ip >> 8) % 256,
	  request->ip % 256, request->netmask, request->next_hop);
  node = root;
  for (i = 0; i < request->netmask; i ++) {
    kid = bit_k (request->ip, 31 - i);
    if (node->child[kid] == NULL) {   /*  Haven't chlid from kid side  */
      printf ("Route not exist...delete ignored\n");
      return;
    }
    /*  Second case at level 16  */
    if ((request->netmask >= 16) && (request->netmask <= 24) && i == 16) {
      root16 = node;
      /*  For testing only  */
      for (j = 0; j < 256; j ++) riga1[j] = 255;
      get_row_elems (riga1, node, 0, 0);
      /*  For testing only  */
      for (j = 0; j < 256; j ++) if (riga1[j] == 255) abort ();
    }
    father = node;
    node = node->child[kid];
  }
  if (node == NULL) {
    printf ("Route not exist...delete ignored\n");
    return;
  }
  if (node->mark == 0) {
    printf ("Route not exist...delete ignored\n");
    return;
  }
  node->mark = 0;
  node->next_hop = father->next_hop;
  /*  Second case at level 16  */
  if (request->netmask == 16) {
    //if ((request->netmask >= 16) && (request->netmask <= 24) && i == 16) {
    root16 = node;
    /*  For testing only  */
    for (j = 0; j < 256; j ++) riga1[j] = 255;
    get_row_elems (riga1, node, 0, 0);
    /*  For testing only  */
    for (j = 0; j < 256; j ++) if (riga1[j] == 255) abort ();
  }
  node->next_hop = request->next_hop;
  node->mark = 1;
  if (request->netmask < 16) {
    rows = cover_holes_min16 (node, request->netmask, request->netmask, 0, &t);
    printf ("Prefix %u modifica %u rows %u elems\n",request->netmask, rows, t);
  }
  if ((request->netmask >= 16) && (request->netmask <= 24)) {
    cover_holes_16_24 (node, request->netmask, request->netmask);
    assert (root16 != NULL);
    get_row_elems (riga2, root16, 0, 0);
    modc2 += differ (riga1, riga2);
    printf ("Differ %u\n",differ (riga1, riga2));
  }
  if (request->netmask > 24) 
    cover_holes_max24 (node, request->netmask, request->netmask);
}

void update_announce (trie_t *root, update_t *request) {
  ip_t i;
  int kid;
  trie_t *node, *tmp, *root16 = NULL;
  ip_addr_t rows = 0, j, t = 0;
  cell_t riga1[256], riga2[256];
  printf ("Announce %u.%u.%u.%u/%u -> %u\n",request->ip >> 24, 
	  (request->ip >> 16) % 256, (request->ip >> 8) % 256,
	  request->ip % 256, request->netmask, request->next_hop);
  node = root;
  for (i = 0; i < request->netmask; i ++) {
    kid = bit_k (request->ip, 31 - i);
    if (node->child[kid] == NULL) {   /*  Haven't chlid from kid side  */
      node->child[kid] = malloc (sizeof (trie_t));
      if (!node->child[kid]) return;
      tmp = (trie_t *) node->child[kid];
      tmp->next_hop = node->next_hop;
      tmp->child[0] = NULL;
      tmp->child[1] = NULL;
      tmp->mark = 0;
    }
    /*  Second case at level 16  */
    if ((request->netmask >= 16) && (request->netmask <= 24) && i == 16) {
      root16 = node;
      /*  For testing only  */
      for (j = 0; j < 256; j ++) riga1[j] = 255;
      get_row_elems (riga1, node, 0, 0);
      /*  For testing only  */
      for (j = 0; j < 256; j ++) if (riga1[j] == 255) abort ();
    }
    node = node->child[kid];
  }
  /*  Second case at level 16  */
  if ((request->netmask >= 16) && (request->netmask <= 24) && i == 16) {
    root16 = node;
    /*  For testing only  */
    for (j = 0; j < 256; j ++) riga1[j] = 255;
    get_row_elems (riga1, node, 0, 0);
    /*  For testing only  */
    for (j = 0; j < 256; j ++) if (riga1[j] == 255) abort ();
  }
  node->next_hop = request->next_hop;
  node->mark = 1;
  if (request->netmask < 16) {
    rows = cover_holes_min16 (node, request->netmask, request->netmask, 0, &t);
    printf ("Prefix %u modifica %u rows %u elems\n",request->netmask, rows, t);
  }
  if ((request->netmask >= 16) && (request->netmask <= 24)) {
    cover_holes_16_24 (node, request->netmask, request->netmask);
    assert (root16 != NULL);
    get_row_elems (riga2, root16, 0, 0);
    modc2 += differ (riga1, riga2);
    printf ("Differ %u\n",differ (riga1, riga2));
  }
  if (request->netmask > 24) 
    cover_holes_max24 (node, request->netmask, request->netmask);
}

ip_addr_t cover_holes_min16 (trie_t *root, ip_t level, 
			     ip_t base, ip_addr_t rows, ip_addr_t *tot) {
  cell_t r1[256], r2[256];
  if (level == 16) {
    get_row_elems (r1, root, 0, 0);
    rows ++;
  }
  if (root->child[0] != NULL) {
    if (((trie_t*) root->child[0])->mark == 0) {
      ((trie_t*) root->child[0])->next_hop = root->next_hop;
      rows = cover_holes_min16 (root->child[0], level + 1, base, rows, tot);
    }
  }
  if (root->child[1] != NULL) {
    if (((trie_t*) root->child[1])->mark == 0) {
      ((trie_t*) root->child[1])->next_hop = root->next_hop;
      rows = cover_holes_min16 (root->child[1], level + 1, base, rows, tot);
    }
  }
  if (level == 16) {
    get_row_elems (r2, root, 0, 0);
    *tot = *tot + differ (r1, r2);
    /*printf ("modifica %u ",differ (r1, r2));*/
  }
  return rows;
}

void cover_holes_16_24 (trie_t *root, ip_t level, ip_t base) {
  if (root->child[0] != NULL) {
    if (((trie_t*) root->child[0])->mark == 0) {
      ((trie_t*) root->child[0])->next_hop = root->next_hop;
      cover_holes_16_24 (root->child[0], level + 1, base);
    }
  }
  if (root->child[1] != NULL) {
    if (((trie_t*) root->child[1])->mark == 0) {
      ((trie_t*) root->child[1])->next_hop = root->next_hop;
      cover_holes_16_24 (root->child[1], level + 1, base);
    }
  }
}

/*  Layer 2  */
void cover_holes_max24 (trie_t *root, ip_t level, ip_t base) {
  if (root->child[0] != NULL) {
    if (((trie_t*) root->child[0])->mark == 0) {
      ((trie_t*) root->child[0])->next_hop = root->next_hop;
      cover_holes_max24 (root->child[0], level + 1, base);
    }
  }
  if (root->child[1] != NULL) {
    if (((trie_t*) root->child[1])->mark == 0) {
      ((trie_t*) root->child[1])->next_hop = root->next_hop;
      cover_holes_max24 (root->child[1], level + 1, base);
    }
  }
}


void get_row_elems (cell_t *riga, trie_t *root, ip_addr_t ip,ip_t mask) {
  /*  If I'm a leaf the following 2 else are execuded  */
  ip_addr_t i;
  if (mask == 8) {
    riga[ip] = root->next_hop;
    return;
  }
  if (root->child[0] != NULL) 
    get_row_elems (riga, root->child[0], ip, mask + 1);
  else {
    for (i = 0; i < exp (8 - (mask + 1)); i ++) {
      riga[ip + i] = root->next_hop;
    }
  }
  if (root->child[1] != NULL) 
    get_row_elems (riga, root->child[1], (ip | exp (8 - (mask + 1))), 
		   mask + 1);
  else {
    for (i = 0; i < exp (8 - (mask + 1)); i ++) {
      riga[(ip | exp (8 - (mask + 1))) + i] = root->next_hop;
    }
  }
}

ip_addr_t differ (cell_t *s1, cell_t *s2) {
  ip_addr_t i, diff = 0;
  for (i = 0; i < 256; i ++) 
    if (s1[i] != s2[i]) diff ++;
  return diff;
}
