#include "iplookup.h"

void make_updates (char *nome, char **next_hops, trie_t *root) {
  FILE *pf;
  char op;
  int error;  /*  Error inside input routing table  */
  int ip[4], netmask, i;
  char riga[MAX_RIGA_READ], next_hop[MAX_IP_LUNG_AS_STRING];
  ip_addr_t route_count, err_route_count;
  ip_addr_t wit=0,ann=0;
  update_t *elem;
  if ((pf = fopen (nome, "r")) == NULL) {
    printf ("Unable to open file\n");
    abort ();
  }
  /*  header (will be destroyed at end)  */
  elem = (update_t *) malloc (sizeof (update_t));
  route_count = 0;
  err_route_count = 0;
  /*  Load entry form file  */
  while (fgets (riga, MAX_RIGA_READ, pf) != NULL) {
    error = 0;  /*  Not errors found in this entry of input table  */
    route_count ++;
    switch (riga[0]) {
    case 'A':
    case 'a':
      sscanf (riga,"%c %d.%d.%d.%d/%d %s",&op, &ip[0],&ip[1],&ip[2],&ip[3],
	      &netmask,next_hop);
      elem->op = 0;   /*  ins  */
      ann ++;
      break;
    case 'w':
    case 'W':
      sscanf (riga,"%c %d.%d.%d.%d/%d",&op, &ip[0],&ip[1],&ip[2],&ip[3],
	      &netmask);
      elem->op = 1;  /*  del  */
      wit ++;
      break;
    default:
      route_count --;
      continue;
    }
    /*  Verify data integrity and copy into list  */
    elem->ip = 0;
    for (i = 0; i < 4; i ++) {
      if (ip[i] < 0 || ip[i] > 255) {
	printf ("Address error %d.%d.%d.%d/%d, route ignored\n",ip[0],ip[1],ip[2],ip[3],netmask);
	i = 4;  /*  exit from loop  */
	error = 1;
      }
      else elem->ip = (elem->ip * 256) + ip[i];
    }
    /*  Useless netmask test if entry will be ignored  */
    if (error != 1) {
      /*  Veryfy netmassk integrity  */
      if (netmask < 1 || netmask > 32) {
	printf ("Netmask error %d.%d.%d.%d/%d, route ignored\n",ip[0],ip[1],ip[2],ip[3],netmask);
	error = 1;
      }
      else {
	elem->netmask = netmask;
	/*  Compute minimal netmask and correct if not enough  */
	netmask = calcola_netmask (ip[0], ip[1], ip[2], ip[3]);
	if (netmask > elem->netmask) {
	  printf ("Network error %d.%d.%d.%d/%d...",ip[0],ip[1],ip[2],ip[3],elem->netmask);
	  elem->ip = correct_nerwork (elem->ip, elem->netmask);
	  ip[0] = elem->ip >> 24;
	  ip[1] = (elem->ip << 8) >> 24;
	  ip[2] = (elem->ip << 16) >> 24;
	  ip[3] = elem->ip % 256;
	  printf ("corrected in %d.%d.%d.%d/%d\n",ip[0],ip[1],ip[2],ip[3],elem->netmask);
	}
      }
    } 
    /*  Insert next hop in his array and code in cell_t  */
    if (elem->op == 0) {  /*  Only for announcements  */
      i = ins_in_next_hop_array (next_hops, next_hop);
      /*  overflow if malloc fail, but programm however abort  */
      elem->next_hop = (cell_t) i;  
      /*  Verify allocation faliure  */
      if (i == -1) {
	printf ("Error insetring in next hop array");
	abort ();
      }
    }
    if (error == 0) {
      /*  Make update  */
      update (root, elem);
    }
    else {
      err_route_count ++;
    }
  }
  free (elem);
  printf ("%u update loaded %d with errors\n",route_count, err_route_count);
  printf ("%u Annuncement, %u Withdrawals\n",ann,wit);
  fclose (pf);
}
