#include "iplookup.h"

int ins_in_next_hop_array (char **next_hops, char *next_hop) {
  char trovato;
  cell_t i;
  trovato = 1;   /* mean not found  */
  i = 0;
  while (next_hops[i] != NULL && trovato != 0) {
    trovato = strcmp (next_hops[i], next_hop);
    i ++;
  }
  /*  Next hop was found in the previus posizion of array  */
  if (trovato == 0) return i - 1;
  /*  Next hop not found, will be inserted now  */
  next_hops[i] = (char *) malloc (strlen (next_hop) + 1); /* + '\0'  */
  if (next_hops[i] == NULL) {
    printf ("Malloc faliure creating next hop array... can't go on\n");
    return -1;
  }
  strcpy (next_hops[i], next_hop);
  /*  Is the position if last inserted  */
  return i;
}

ip_t used_bits (ip_t elem) {
  if (elem & 1) return 8;
  if (elem & 2) return 7;
  if (elem & 4) return 6;
  if (elem & 8) return 5;
  if (elem & 16) return 4;
  if (elem & 32) return 3;
  if (elem & 64) return 2;
  if (elem & 128) return 1;
  return 0;
}

ip_t calcola_netmask (ip_t ip1, ip_t ip2, ip_t ip3, ip_t ip4) {
  if (ip4 != 0) return 24 + used_bits (ip4);
  if (ip3 != 0) return 16 + used_bits (ip3);
  if (ip2 != 0) return 8 + used_bits (ip2);
  return used_bits (ip1);
}

ip_addr_t correct_nerwork (ip_addr_t ip, ip_t netmask) {
  ip >>= (32 - netmask);
  return ip << (32 - netmask);
}

void unload_next_hops (char **next_hops) {
  ip_addr_t i;
  /*  Free next_hops array  */
  for (i = 0; next_hops[i] != NULL; i ++) 
    free (next_hops[i]);
}

struct net_table *load_table (char *nome, char **next_hops,ip_addr_t *rou_c) {
  FILE *pf;
  int error;  /*  Error inside input routing table  */
  int ip[4], netmask, i;
  char riga[MAX_RIGA_READ], next_hop[MAX_IP_LUNG_AS_STRING];
  ip_addr_t route_count, err_route_count;
  struct net_table *init, *self, *prec;
  if ((pf = fopen (nome, "r")) == NULL) {
    printf ("Unable to open file\n");
    return NULL;
  }
  /*  header (will be destroyed at end)  */
  init = (struct net_table *) malloc (sizeof (struct net_table));
  prec = init;
  route_count = 0;
  err_route_count = 0;
  /* Inizialize next hop vector */
  for (i = 0; i < MAX_NEXT_HOPS; i ++) next_hops[i] = NULL;
  next_hops[0] = (char *) malloc (17); /* NO ROUTE TO HOST + `\0` */
  if (next_hops[0] == NULL) {
    printf ("Malloc faliure creating next hop array... can't go on\n");
    return NULL;
  }
  strcpy (next_hops[0],"NO ROUTE TO HOST");
  /*  Load entry form file  */
  while (fgets (riga, MAX_RIGA_READ, pf) != NULL) {
    error = 0;  /*  Not errors found in this entry of input table  */
    route_count ++;
    self = (struct net_table *) malloc (sizeof (struct net_table));
    if (self == NULL) {
      printf ("Malloc faliure... can't go on\n");
      /*  Free already allocated resources  */
      prec->next = NULL;
      while (init != NULL) {
	prec = init;
	init = init->next;
	free (init);
      }
      /*  Free next_hops array  */
      for (i = 0; next_hops[i] != NULL; i ++) free (next_hops[i]);
      return NULL;
    }
    sscanf (riga,"%d.%d.%d.%d/%d %s",&ip[0],&ip[1],&ip[2],&ip[3],&netmask,next_hop);
    /*  Verify data integrity and copy into list  */
    self->ip = 0;
    for (i = 0; i < 4; i ++) {
      if (ip[i] < 0 || ip[i] > 255) {
	printf ("Address error %d.%d.%d.%d/%d, route ignored\n",ip[0],ip[1],ip[2],ip[3],netmask);
	i = 4;  /*  exit from loop  */
	error = 1;
      }
      else self->ip = (self->ip * 256) + ip[i];
    }
    /*  Useless netmask test if entry will be ignored  */
    if (error != 1) {
      /*  Veryfy netmassk integrity  */
      if (netmask < 1 || netmask > 32) {
	printf ("Netmask error %d.%d.%d.%d/%d, route ignored\n",ip[0],ip[1],ip[2],ip[3],netmask);
	error = 1;
      }
      else {
	self->netmask = netmask;
	/*  Compute minimal netmask and correct if not enough  */
	netmask = calcola_netmask (ip[0], ip[1], ip[2], ip[3]);
	if (netmask > self->netmask) {
	  printf ("Network error %d.%d.%d.%d/%d...",ip[0],ip[1],ip[2],ip[3],self->netmask);
	  self->ip = correct_nerwork (self->ip, self->netmask);
	  ip[0] = self->ip >> 24;
	  ip[1] = (self->ip << 8) >> 24;
	  ip[2] = (self->ip << 16) >> 24;
	  ip[3] = self->ip % 256;
	  printf ("corrected in %d.%d.%d.%d/%d\n",ip[0],ip[1],ip[2],ip[3],self->netmask);
	}
      }
    } 
    /*  Insert next hop in his array and code in cell_t  */
    i = ins_in_next_hop_array (next_hops, next_hop);
    /*  overflow if malloc fail, but programm however abort  */
    self->next_hop = (cell_t) i;  
    /*  Verify allocation faliure  */
    if (i == -1) {
      /*  Free already allocated resources  */
      free (self);
      prec->next = NULL;
      while (init != NULL) {
	prec = init;
	init = init->next;
	free (init);
      }
      /*  Free next_hops array  */
      for (i = 0; next_hops[i] != NULL; i ++) free (next_hops[i]);
      return NULL;
    }
    if (error == 0) {
      prec->next = self;
      prec = prec->next;
    }
    else {
      free (self);
      err_route_count ++;
    }
  }
  /*  close list and destroy useless header  */
  prec->next = NULL;
  prec = init;
  init = init->next;
  free (prec);
  /*  set the number of corrected elements loaded  */
  *rou_c = route_count - err_route_count;
  printf ("%u route loaded %d with errors\n",route_count, err_route_count);
  /*  It count the number of different next hop  */
  for (i = 0; next_hops[i] != NULL; i ++);
  /*  Ignore NO ROUTE TO HOST  */
  printf ("%d different next hop loaded\n", i - 1);
  fclose (pf);
  return init;
}
